/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXSSHFS, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Callbacks.h"
#include "Cache.h"

//
// Imported variables.
//

extern SftpClient Sftp;

//
// Global variables.
//

Int64 CachedHandlesEntry::localHandlesCount_ = 1;

HANDLE CachedHandlesMutex = NULL;

HANDLE CacheGarbageCollectorThread = NULL;

Int IdleCount = 0;

//
// <Local handle> |-> <CachedHandlesEntry>.
//

static map<Int64, CachedHandlesEntry> CachedHandles;

//
// Filenames map
//

//
// 'LOCAL NON-CASE SENSITIVE NAME' |-> 'Remote Case Sensitive Name'.
//

static map<wstring, CachedPathsEntry> CachedPaths;

//
// Internal functions.
//

//
// Cached handle constructor.
//
// localHandle - local side handle (IN).
// localPath   - local path seen on client side (IN).
// accessMode  - handle's access mode (see CreateFile on MSDN) (IN).
// create      - handle's create disposition (see MSDN) (IN).
//

CachedHandlesEntry::CachedHandlesEntry(Int64 localHandle, 
                                           const wchar_t *localPath,
                                               DWORD accessMode, DWORD create,
                                                   Int isDir, Int *error)
{
  localHandle_   = localHandle;
  remoteHandle_  = -1;
  localPath_     = localPath;
  accessMode_    = accessMode;
  create_        = create;
  evaluated_     = 0;
  isDir_         = isDir;
  error_         = 0;
  markedToClose_ = 0;
  lastAccess_    = GetTimeMs();
  childsCount_   = -1;
  childsList_    = NULL;

  //
  // If create disposition is set to create new file
  // evaluate handle immediately.
  //
  
/*
  FIXME: Some applications use alone CreateFile call to check
         is file exist to decide what to do next. In this future
         lazy evaluation causes wrong application's behavior.
  
  if (create & CREATE_ALWAYS || create & CREATE_NEW)
*/  

  if (Not(isDir))
  {
    DBG_MSG("Immediatelly evaluating [%lld]...", localHandle);
  
    if ((create & CREATE_ALWAYS) || (create & CREATE_NEW))
    {
      evaluated_ = 0;
    }
    
    evaluate();
  }
  
  if (error)
  {
    *error = error_;
  }
};

CachedHandlesEntry::CachedHandlesEntry()
{
  localHandle_   = -1;
  remoteHandle_  = -1;
  accessMode_    = 0;
  create_        = 0;
  evaluated_     = 0;
  isDir_         = 0;
  error_         = 0;
  markedToClose_ = 0;
  childsCount_   = -1;
  childsList_    = NULL;  
}

//
// Generate unique local handle.
//
// WARNING: Function is NOT thread safe.
//

Int64 CachedHandlesEntry::generateHandle()
{
  Int64 ret = localHandlesCount_;
  
  localHandlesCount_ ++;
  
  return ret;
}

//
// Lazy evaluate of cached handle.
// Until this call, handle exists only locally i.e. it's NOT
// really opened on server side.
//
// RETURNS: 0 if OK.
//

Int CachedHandlesEntry::evaluate()
{
  DBG_ENTER("CachedHandlesEntry::evaluate");

  Int exitCode = -1;
  Int mode     = 0;
  Int exist    = 0;
  Int isDir    = 0;
  
  Char remotePath[MAX_PATH] = {0};

  //
  // Skip if already evaluated.
  //
  
  if (evaluated_ == 1)
  {
    DBG_MSG("CachedHandlesEntry::evaluate :"
                " [%ls] already evaluated.", localPath_.c_str());
  }
  
  //
  // If not evaluated yet, open remote file on server side.
  //
  
  else
  {
    //
    // Create full remote path.
    //

    GetFullRemotePath(remotePath, (wchar_t *) localPath_.c_str(), MAX_PATH);

    DBG_MSG("CachedHandlesEntry::evaluate{path = [%s],"
                " accessMode = [%d], create = [%d]}.",
                    remotePath, accessMode_, create_);

    //
    // Stat file first.
    //

/*
    FIXME: Cache directory content if possible.
    
    exist = (CacheLocalStat(localPath_.c_str()) == 0);
*/    
    
    exist = (Sftp.stat(remotePath) == 0);
    
    DBG_MSG("Exist : [%s][%d]", remotePath, exist);
    
    if (isDir_ == 0)
    {
      DBG_MSG("Opening file...");
/*
      //
      // Native file systems returns ACCESS_DENIED on attemp to open
      // directory via createfile.
      // Windows expolorer still use CreateFile() to open directories.
      //
      
      if (fileInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      {
        DBG_MSG("CreateFile : Cannot open directories in this way.");
        
        SetLastError(ERROR_ACCESS_DENIED);
        
        goto fail;
      }      
*/      
      //
      // Translate WINAPI access mode flags to SFTP open mode.
      //

      if (accessMode_ & FILE_READ_DATA || accessMode_ & GENERIC_READ)
      {
        DBG_MSG("CreateFile : FILE_READ_DATA");

        mode |= SSH2_FXF_READ;
      }

      if (accessMode_ & FILE_WRITE_DATA || accessMode_ & GENERIC_WRITE)
      {
        DBG_MSG("CreateFile : FILE_WRITE_DATA");

        mode |= SSH2_FXF_WRITE;
      }

      if (accessMode_ & FILE_APPEND_DATA)
      {
        DBG_MSG("CreateFile : FILE_APPEND_DATA");

        mode |= SSH2_FXF_APPEND;
      }

      //
      // Process create disposition arg.
      //

      switch(create_)
      {
        //
        // Create always. If exists open, if not exist open for overwrite.
        //

        case CREATE_ALWAYS:
        {
          DBG_MSG("CreateFile : CREATE_ALWAYS");

          if (exist)
          {
            mode |= SSH2_FXF_TRUNC;
          }
          else
          {
            mode |= SSH2_FXF_CREAT;
            
/*            
  FIXME: Cache directory content if possible.
            
            CacheClearChildsList(localPath_.c_str());
*/            
          }

          break;
        }

        //
        // Create new. If exists fail, if not create new.
        //

        case CREATE_NEW:
        {
          DBG_MSG("CreateFile : CREATE_NEW");

          if (exist)
          {
            SetLastError(ERROR_FILE_EXISTS);

            goto fail;
          }
          else
          {
            mode |= SSH2_FXF_CREAT;

/*
  FIXME: Cache directory content if possible.
  
            CacheClearChildsList(localPath_.c_str());
*/            
          }

          break;
        }

        //
        // Open always. If exists open, if not craete new file.
        //

        case OPEN_ALWAYS:
        {
          DBG_MSG("CreateFile : OPEN_ALWAYS");

          if (Not(exist))
          {
            mode |= SSH2_FXF_CREAT;
          }

          break;
        }

        //
        // Open existing. If exists open, if not fail.
        //

        case OPEN_EXISTING:
        {
          DBG_MSG("CreateFile : OPEN_EXISTING");

          if (Not(exist))
          {
            SetLastError(ERROR_FILE_NOT_FOUND);

            goto fail;
          }

          break;
        }

        //
        // Truncate existing. If exists open and zero, if not fail.
        //

        case TRUNCATE_EXISTING:
        {
          DBG_MSG("CreateFile : TRUNCATE_EXISTING");

          if (exist)
          {
            mode |= SSH2_FXF_TRUNC;
          }

          break;
        }
      }
    }  

    //
    // Retrieve handle from server.
    //

    remoteHandle_ = Sftp.open(remotePath, mode, localHandle_, &isDir);

    FAIL(remoteHandle_ < 0);

    DBG_MSG("CachedHandlesEntry::evaluate : Assigned handle [%ls] |-> [%d].",
                localPath_.c_str(), remoteHandle_);

    evaluated_ = 1;
    
    CacheAddPath(localPath_.c_str(), remotePath);
  }
  
  //
  // Directories always dead, because after readDir() reached EOF
  // one time it become unusable. We cache result of FindFIles()
  // until handle dead.
  // 
  
  markedToClose_ = 0;
  lastAccess_    = GetTimeMs();
  
  //
  // Error handler.
  //
  
  error_   = 0;
  exitCode = 0;
  
  fail:

  if (exitCode)
  {
    error_ = GetLastError();
    
    if (error_ == 0)
    {
      error_ = ERROR_ACCESS_DENIED;
    }
    
    DBG_MSG("Cannot evaluate cached handle for [%s].\n"
                 "Error code is : %d.", remotePath, error_);
  
/*
  FIXME: Cache directory content if possible.
    
    if (error_ == ERROR_FILE_NOT_FOUND)
    {
      CacheRemoveLocalPath(localPath_.c_str());
    }
*/    
  }
  
  DBG_LEAVE("CachedHandlesEntry::evaluate");
  
  return exitCode;
}

//
// Exported API functions.
//

//
// Insert new local handle to cache.
// WARNING: Handle is NOT opened on server side until
// call to CacheEvaluate() functions.
//
// localPath   - locally seen path (IN).
// accessMode  - handle's access mode (see CreateFile on msdn) (IN).
// create      - handle's create disposition (see msdn) (IN).
// isDir       - 1 if target path is directory (IN).
//
// RETURNS: Assigned local side handle.
//

Int64 CacheInsertHandle(const wchar_t *localPath, DWORD accessMode, 
                            DWORD create, Int isDir, Int *error)
{
  CacheLock();
  
  Int64 localHandle = -1;
  
  map<Int64, CachedHandlesEntry>::iterator it;
  
  //
  // Try to live up one of dead handle first.
  //
  
  for (it = CachedHandles.begin(); 
           localHandle == -1 && it != CachedHandles.end(); it++)
  {
    CachedHandlesEntry &e = it -> second;
    
    if (e.localPath_ == localPath)
    {
      //
      // Dead handle with the same path found.
      // Live up it again.
      //
      
      if (e.markedToClose_ && e.accessMode_ == accessMode)
      {
        localHandle = e.localHandle_;

        e.markedToClose_ = 0;
        e.lastAccess_    = GetTimeMs();
      
        if (error)
        {
          *error = 0;
        }

        DBG_MSG("Cleared close flag from [%ls][%lld].", localPath, localHandle);
      }

      //
      // The target path is the same, but handle is in use
      // or has different accessMode.
      // Use it only to determine real target type (file/dir) 
      // without calling to server.
      //
      
/*
  FIXME: More test needed.
  
      else
      {
        isDir = e.isDir_;
        
        DBG_MSG("CacheInsertHandle : Remained [%d] type for [%ls].", 
                    e.isDir_, localPath);
      }
*/
    }
  }
    
  // 
  // If not found, generate new handle.
  //
  
  if (localHandle == -1)
  {
    localHandle = CachedHandlesEntry::generateHandle();
  

    DBG_MSG("Caching [%ls]...", localPath);

    DBG_MSG("Caching handle [%ls][%lld][%x][%x]...",
                localPath, localHandle, accessMode, create);
  
    CachedHandlesEntry e(localHandle, localPath,
                             accessMode, create, isDir, error);

    if (e.isDir_)
    {
      localHandle |= DIR_HANDLE_MASK;
    }
    
    CachedHandles[localHandle] = e;
  }
  
  CacheUnlock();
  
  return localHandle;
}

//
// Close cached handle.
//
// localHandle - local handle number managed by caller (IN).
// immediately - force to close on server without delay (IN).
//

void CacheCloseHandle(Int64 localHandle, Int immediately)
{
  DBG_MSG("Closing cached handle [%lld]...", localHandle);
  
  map<Int64, CachedHandlesEntry>::iterator it;
  
  CacheLock();  
  
  //
  // Find handle in cache.
  //
    
  it = CachedHandles.find(localHandle);
  
  FAILEX(it == CachedHandles.end(), 
             "Handle [%lld] not found in cache.\n", localHandle)

  //
  // Immediately close request (e.g. before delete).
  //
  
  if (immediately || it -> second.accessMode_ & FILE_WRITE_DATA
          || it -> second.accessMode_ & GENERIC_WRITE)
  {
    DBG_MSG("Immediately closing [%lld][%ls]...", 
                localHandle, it -> second.localPath_.c_str());
  
    //
    // If handle is evaluated close it on server side.
    //
  
    if (it -> second.evaluated_)
    {
      Sftp.close(it -> second.remoteHandle_);
    }
    
    //
    // Free cached childs list if exist.
    //
    
/*
  FIXME: Cache directory content if possible.
  
    CacheClearChildsList(it -> second.localPath_.c_str());
*/    

    if (it -> second.childsList_ != NULL)
    {
      free(it -> second.childsList_);
    }

    //
    // Remove handle from cache.
    // 

    CachedHandles.erase(it);
  }
  
  //
  // Default lazy close. Only mark to close.
  //

  else
  {
    it -> second.markedToClose_ = 1;
    it -> second.lastAccess_    = GetTimeMs();
  
    DBG_MSG("[%lld][%ls] marked to close.", 
                localHandle, it -> second.localPath_.c_str());
  }  

  fail:;
  
  CacheUnlock();
}

//
// Lazy evaluation of cached handle. File is NOT opened on server
// until this function called.
//
// RETURNS: Remote handle of -1 if error.
//

Int64 CacheEvaluateHandle(Int64 localHandle)
{
  DBG_MSG("Evaluating handle [%lld]...", localHandle);
  
  Int exitCode = -1;
  
  map<Int64, CachedHandlesEntry>::iterator it;
  
  CacheLock();
  
  //
  // Find handle in cache.
  //
    
  it = CachedHandles.find(localHandle);
  
  FAILEX(it == CachedHandles.end(),
             "ERROR: Handle [%lld] not found in cache.\n", localHandle)

  //
  // Evaluate it on server side.
  //
  
  FAIL(it -> second.evaluate());
  
  exitCode = it -> second.remoteHandle_;

  fail:
  
  CacheUnlock();
  
  return exitCode;
}

//
//
//

Int CacheGarbageCollector(void *unused)
{
  DBG_ENTER("CacheGarbageCollector");
  
  map<Int64, CachedHandlesEntry>::iterator it;
  map<Int64, CachedHandlesEntry>::iterator next;
  
  double current = 0;
  
  while(1)
  {
    current = GetTimeMs();
    
    Sleep(CLOSE_DELAY);
    
    //
    // Lock cached handles map.
    //
    
    CacheLock();
  
    for (it = CachedHandles.begin(); it != CachedHandles.end();)
    {
      CachedHandlesEntry &e = it -> second;
 
      //DBG_MSG("[%ls][%lf]\n", e.localPath_.c_str(), current - e.lastAccess_);
      
      next = it;
      
      next ++;
      
      if (e.markedToClose_ && current - e.lastAccess_ > CLOSE_DELAY)
      {
        DBG_MSG("GarbageCollector : Closing handle [%lld][%lld]...", 
                    e.localHandle_, e.remoteHandle_);
        
        //
        // If handle is evaluated close it on server side.
        //
  
        if (e.evaluated_)
        {
          IdleCountReset();
          
          Sftp.close(e.remoteHandle_);
        }

        //
        // Free cached childs list if exist.
        //
    
        if (e.childsList_ != NULL)
        {
          free(e.childsList_);
        }

        //
        // Remove handle from cache.
        // 

        CachedHandles.erase(it);
      }
      
      it = next;
    }
    
    //
    // Release handle map.
    //
    
    CacheUnlock();
 
    //
    // Stat remote root as keep alive packet.
    // It allow us to know, when server dead when
    // user is idle on client side.
    //
    
    IdleCount ++;
    
    if (IdleCount >= MAX_IDLE_COUNT)
    {
      DBG_MSG("[Keep alive].");
      
      Char fullPath[MAX_PATH];
      
      GetFullRemotePath(fullPath, L"/", MAX_PATH);

      Sftp.stat(fullPath);
    }
    
    //
    // Removes fake 0 bytes files from recycler bin.
    //
    
    RecyclerFlush();
  }
  
  DBG_LEAVE("CacheGarbageCollector");
  
  return 0;
}

//
// Init mutex and garbage collector threads to manage 
// handles cache.
//
// RETURNS: 0 if OK.
//

Int InitCache()
{
  DBG_ENTER("InitCache");
  
  Int exitCode = -1;
  
  //
  // Create mutex to lock cached handles.
  //

  if (CachedHandlesMutex == NULL)
  {
    CachedHandlesMutex = CreateMutex(NULL, FALSE, NULL);
  
    FAILEX(CachedHandlesMutex == NULL, "ERROR: Cannot create mutex.\n");
  }
  
  //
  // Create garbage collector thread to avoid often
  // open/close/open sequence. Real close is delayed.
  //
  
  if (CacheGarbageCollectorThread == NULL)
  {
    CacheGarbageCollectorThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) CacheGarbageCollector,
                                                   NULL, 0, NULL);
  
    FAILEX(CacheGarbageCollectorThread == NULL,
             "ERROR: Cannot create garbage collector thread.\n");
  }             
  
  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:

  if (exitCode)
  {
    Error("ERROR: Cannot init cache.\n"
              "Error code is : %d.\n", (Int) GetLastError());
  }
  
  DBG_LEAVE("InitCache");
  
  return exitCode;  
}

//
// Cache list of childs to before opened parent directory.
// WARNING: childsList MUST be dynamic allocated by caller, but
//          its freed by cache garbage colector when parent's 
//          handle dead.
//

Int CacheSetChildsList(Int64 localParentHandle, 
                           Int childsCount, WIN32_FIND_DATAW *childsList)
{
  DBG_ENTER("CacheSetChildsList");
  
  Int exitCode = -1;
  
  map<Int64, CachedHandlesEntry>::iterator it;
  
  CacheLock();
  
  //
  // Find parent's handle in cache. 
  //
  
  it = CachedHandles.find(localParentHandle);
  
  FAILEX(it == CachedHandles.end(),
             "ERROR: Cannot set childs for not cached handle.\n");
  
  //
  // Set child list in found handle.
  //

  it -> second.childsCount_ = childsCount;
  it -> second.childsList_  = childsList;
  it -> second.isDir_       = 1;         
  
  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:

  if (exitCode)
  {
    Error("ERROR: Cannot set cached childs list.\n");
  }
  
  CacheUnlock();
  
  DBG_LEAVE("CacheSetChildsList");
  
  return exitCode;
}

//
// Clear cached childs related to given target path.
//

Int CacheClearChildsList(const wchar_t *localPath)
{
  DBG_ENTER("CacheClearChildsList");

  wchar_t parent[MAX_PATH];
  
  Int exitCode = -1;
  
  map<Int64, CachedHandlesEntry>::iterator it;
  
  CacheLock();
  
  //
  // Get parent path for given target. 
  // Eg. '/root' for '/root/xyz'.
  //
  
  FAIL(localPath[0] == L'\\' && localPath[1] == 0);
  FAIL(GetParentPath(parent, sizeof(parent), localPath));
  
  //
  // Find parent's handle in cache. 
  //
  
  for (it = CachedHandles.begin(); it != CachedHandles.end(); it++)
  {
    if (it -> second.localPath_ == parent)
    {
      if (it -> second.childsList_ != NULL)
      {
        DBG_MSG("Clearing childs for [%ls]...", parent);

        free(it -> second.childsList_);
        
        it -> second.childsCount_ = -1;

        //
        // Force directory reopen to avoid read content after EOF.
        //
        
        if (it -> second.evaluated_)
        {
          Sftp.close(it -> second.remoteHandle_);
        }
        
        it -> second.evaluated_ = 0;
      }
    }
  }  
  
  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:

  if (exitCode)
  {
    Error("ERROR: Cannot clear cached childs list.\n");
  }
  
  CacheUnlock();
  
  DBG_LEAVE("CacheClearChildsList");
  
  return exitCode;
}

//
//
//

Int CacheGetChildsList(Int64 localParentHandle, 
                           Int *childsCount, WIN32_FIND_DATAW **childsList)
{
  DBG_ENTER("CacheGetChildsList");
  
  Int exitCode = -1;
  
  map<Int64, CachedHandlesEntry>::iterator it;
  
  CacheLock();
  
  //
  // Find parent's handle in cache. 
  //
  
  it = CachedHandles.find(localParentHandle);
  
  FAILEX(it == CachedHandles.end(),
             "ERROR: Cannot get childs for not cached handle.\n");
  
  //
  // Fail if not cached yet.
  //

  FAILEX(it -> second.childsCount_ == -1,
             "Childs for [%lld] not cached yet.\n", localParentHandle);
  
  //
  // Set child list in found handle.
  //

  *childsCount = it -> second.childsCount_;
  *childsList  = it -> second.childsList_;
  
  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:

  CacheUnlock();
  
  DBG_LEAVE("CacheGetChildsList");
  
  return exitCode;
}

//
// Sftp stat working on cached data only.
//
// remotePath   - target REMOTE path on server side (IN).
// extendedInfo - extendet info data if needed (OUT/OPT).
//
// RETURNS: 0 if OK.
// 

Int CacheRemoteStat(const Char *remotePath, BY_HANDLE_FILE_INFORMATION *extInfo)
{
  Int exitCode = -1;
  
  Int found    = 0;
  
  map<wstring, CachedPathsEntry>::iterator it;
  
  CacheLock();
  
  for (it = CachedPaths.begin(); Not(found) && it != CachedPaths.end(); it++)
  {
    if (it -> second.remotePath_ == remotePath)
    {
      found = 1;
      
      if (extInfo != NULL)
      {
        FAILEX(Not(it -> second.extInfoSet_),
                   "CacheRemoteStat : Extended data not cached for [%s].\n", 
                       remotePath);
      
        *extInfo = it -> second.extInfo_;
      }
    }
  }
  
  FAILEX(Not(found), "CacheRemoteStat : [%s] not cached yet.", remotePath);
  
  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:

  CacheUnlock();
 
  return exitCode;
}

//
// Add new path entry to cache.
//
// localPath - relative local path on client side (IN).
// entry     - cache entry to assign (IN).
//
// RETURNS: 0 if OK.
//

Int CacheLocalStat(const wchar_t *localPath, CachedPathsEntry *entry)
{
  Int exitCode = -1;
  
  map<wstring, CachedPathsEntry>::iterator it;
  
  CacheLock();
  
  it = CachedPaths.find(localPath);
  
  FAIL(it == CachedPaths.end());
  
  if (entry)
  {
    *entry = it -> second;
  }

  exitCode = 0;
  
  fail:

  CacheUnlock();
  
  return exitCode;
}

//
//
//

Int CacheAddPath(wchar_t *localPath, 
                     const Char *remotePath,
                         WIN32_FIND_DATAW *findData, 
                             Int capitalize)
{
  Char cleanRemote[MAX_PATH] = {0};
  
  Char *dst = cleanRemote;
  
  Char lastChar = 0;
  
  CachedPathsEntry e;
  
  //
  // Normalize remote path. Remove duplicated slashes.
  //
  
  for (Int i = 0; remotePath[i]; i++)
  {
    switch(remotePath[i])
    {
      case '/':
      case '\\':
      {
        if (lastChar != '/' && lastChar != '\\')
        {
          *dst = '/';
        
          dst ++;
        }
        
        break;
      }

      default:
      {
        *dst = remotePath[i];
        
        dst ++;
        
        break;
      }
    }
    
    lastChar = remotePath[i];
  }

  DBG_MSG("CacheAddPath : Normalized remote from [%s] to [%s].\n", remotePath, cleanRemote);

  //
  // Format extended data if passed.
  // 

  e.remotePath_ = cleanRemote;
  
  if (findData != NULL)
  {
    e.extInfoSet_ = 1;
    
    e.extInfo_.dwFileAttributes     = findData -> dwFileAttributes;
    e.extInfo_.ftCreationTime       = findData -> ftCreationTime;
    e.extInfo_.ftLastAccessTime     = findData -> ftLastAccessTime;
    e.extInfo_.ftLastWriteTime      = findData -> ftLastWriteTime;
    e.extInfo_.dwVolumeSerialNumber = 0;
    e.extInfo_.nFileSizeHigh        = findData -> nFileSizeHigh;
    e.extInfo_.nFileSizeLow         = findData -> nFileSizeLow;
    e.extInfo_.nNumberOfLinks       = 0;
    e.extInfo_.nFileIndexHigh       = 0;
    e.extInfo_.nFileIndexLow        = 0;
  }
  else
  {
    e.extInfoSet_ = 0;
  }  
  
  //
  // Capitalize local path if needed.
  //

  if (capitalize && localPath[0] && localPath[1])
  {
    for (Int i = 0; localPath[i]; i++)
    {
      localPath[i] = towupper(localPath[i]);
    }
  }  
  
  //
  // Add new path into cache.
  //
  
  CacheLock();
  
  CachedPaths[localPath] = e;

  CacheUnlock();
      
  DBG_MSG("CacheAddPath : Assigned [%ls] |-> [%s].", localPath, cleanRemote);
  
  return 0;
}

Int CacheAddPath(const wchar_t *localPath, 
                     const Char *remotePath,
                         WIN32_FIND_DATAW *findData)
{
  return CacheAddPath((wchar_t *) localPath, remotePath, findData, 0);
}

//
// Unconditional remove patch from cache.
//
// localPath - relative local path on client side (IN).
//
// RETURNS: 0 if OK.
//

Int CacheRemoveLocalPath(const wchar_t *localPath)
{
  map<wstring, CachedPathsEntry>::iterator it;
  
  it = CachedPaths.find(localPath);
  
  if (it != CachedPaths.end())
  {
    CachedPaths.erase(it);
  }
  
  return 0;
}

//
//
//

void CacheCloseHandlesByPath(const wchar_t *localPath, Int immediately)
{
  DBG_MSG("Closing cached handles for path [%ls]...", localPath);
  
  map<Int64, CachedHandlesEntry>::iterator it;
  map<Int64, CachedHandlesEntry>::iterator next;
  
  CacheLock();  
  
  //
  // Find handles with ginev localPath in cache.
  //

  for (it = CachedHandles.begin(); it != CachedHandles.end();)
  {
    next = it;
    
    next ++;
    
    if (it -> second.localPath_ == localPath)
    {
      //
      // Default lazy close. Only mark to close.
      //

      it -> second.markedToClose_ = 1;
      it -> second.lastAccess_    = GetTimeMs();
  
      DBG_MSG("[%lld][%ls] marked to close.", 
                  it -> second.localHandle_, 
                      it -> second.localPath_.c_str());
    
      //
      // Immediately close request (e.g. before delete).
      //
  
      if (immediately || it -> second.accessMode_ & FILE_WRITE_DATA
              || it -> second.accessMode_ & GENERIC_WRITE || it -> second.isDir_)
      {
        DBG_MSG("Immediately closing [%lld][%ls]...", 
                    it -> second.localHandle_, 
                        it -> second.localPath_.c_str());
  
        //
        // If handle is evaluated close it on server side.
        //
  
        if (it -> second.evaluated_)
        {
          Sftp.close(it -> second.remoteHandle_);
        }
    
        //
        // Free cached childs list if exist.
        //

        if (it -> second.childsList_ != NULL)
        {
          free(it -> second.childsList_);
        }

        //
        // Remove handle from cache.
        // 

        CachedHandles.erase(it);
      }
    }

    it = next;
  }  
  
  CacheUnlock();
}

void IdleCountReset()
{
  IdleCount = 0;
}

void CacheCleanup() 
{
  DBG_ENTER("CacheCleanup");

  TerminateThread(CacheGarbageCollectorThread, 0);

  DBG_LEAVE("CacheCleanup");
}
