/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXSSHFS, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Cache_H
#define Cache_H

//
// Includes.
//

#include <map>
#include "Debug.h"
#include "SftpClient.h"

//
// Defines.
//

//
// After CloseHandle() callback handles are marked to close only.
// Marked handles are closed on server after CLOSE_DELAY ms or when
// it explicite set (e.g. before attemp delete file).
// 
// When the same path is opened until CLOSE_DELAY reached marked
// handles can be live up again (marked to close flag cleared).
//
// Delay closes reduce multiple queries to server on fast
// open/close/open/... sequence (e.g. explorer.exe on Win7).
// 

#define CLOSE_DELAY 5000

#define MAX_IDLE_COUNT 2

#define CacheLock() WaitForSingleObject(CachedHandlesMutex, INFINITE)

#define CacheUnlock() ReleaseMutex(CachedHandlesMutex)

//
// Structs.
//

struct CachedHandlesEntry
{
  static Int64 localHandlesCount_;

  static Int64 generateHandle();
  
  Int64 localHandle_;
  Int64 remoteHandle_;

  wstring localPath_;
  
  Int evaluated_;
  Int isDir_;
  Int error_;
  Int markedToClose_;

  double lastAccess_;  

  DWORD accessMode_;
  DWORD create_;
  
  Int childsCount_;

  WIN32_FIND_DATAW *childsList_;

  Int evaluate();

  CachedHandlesEntry(Int64 localHandle, const wchar_t *localPath, 
                         DWORD accessMode, DWORD create, Int isDir = 0, 
                             Int *error = NULL);

  CachedHandlesEntry();
};

struct CachedPathsEntry
{
  string remotePath_;
  
  Int extInfoSet_;
  
  BY_HANDLE_FILE_INFORMATION extInfo_;
};

//
// Exported functions.
//

//
// General init/uninit/setup functions.
//

Int InitCache();

//
// Handles cache. We implemented here:
//
// 1. lazy open (open on server if really needed)
//
// 2. delayed close (close on server after X ms to 
//    avoid oftet open/close/open series).
//
// 3. Reusing handled closed locally ('marketToClose'), but don't close
//    on server side yet.
//

Int64 CacheInsertHandle(const wchar_t *localPath, 
                            DWORD accessMode, DWORD create, Int isDir = 0, 
                                Int *error = NULL);

void CacheCloseHandle(Int64 localHandle, Int immediately = 0);

void CacheCloseHandlesByPath(const wchar_t *localPath, Int immediately = 0);

Int64 CacheEvaluateHandle(Int64 localHandle);

Int CacheGarbageCollector(void *unused);

//
// Cache direcotires content.
//

Int CacheChildsList(Int64 localParentHandle, 
                        Int childsCount, WIN32_FIND_DATAW *childsList);

Int CacheGetChildsList(Int64 localParentHandle,
                           Int *childsCount, WIN32_FIND_DATAW **childsList);

Int CacheSetChildsList(Int64 localParentHandle,
                           Int childsCount, WIN32_FIND_DATAW *childsList);

Int CacheClearChildsList(const wchar_t *localPath);

//
// Higher level function basing on one of above mechanism.
//

//
// Sftp stat function working on cached data only.
//
                           
Int CacheLocalStat(const wchar_t *localPath, CachedPathsEntry *entry = NULL);

Int CacheRemoteStat(const Char *remotePath, BY_HANDLE_FILE_INFORMATION *extInfo);

Int CacheAddPath(wchar_t *localPath, const Char *remotePath,
                     WIN32_FIND_DATAW *findData = NULL, Int capitalize = 1);

Int CacheAddPath(const wchar_t *localPath, const Char *remotePath,
                     WIN32_FIND_DATAW *findData = NULL);

Int CacheRemoveLocalPath(const wchar_t *localPath);

void IdleCountReset();

void CacheCleanup();

#endif /* Cache_H */
