/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXSSHFS, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Callbacks.h"
#include <map>
#include <set>
#include <Sddl.h>

#undef ProcessId

#define USE_CASE_SENSITIVE_MAP
#define SKIP_DESKTOP_INI

#define RECYCLER_BIN_CYCLES 6

//
// GUID to import IShellFolder interface.
// We use it to get access to virtual recycled bin folder.
//

BYTE IID_IShellFolderBlob[] = 
{
  0xe6, 0x14, 0x02, 0x00, 
  0x00, 0x00, 0x00, 0x00, 
  0xc0, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x46
};

const IID &_IID_IShellFolder = (IID &) IID_IShellFolderBlob;

//
// Global variables.
//

extern OptionsStruct Options;

extern SftpClient Sftp;

static map<Int, wstring> PidCurrentDir;

static set<string> Recycler;
static set<string> DeleteOnClose;

static Int RecyclerWakeUp = 0;

//
// Fake recycler to give false 0 byte size to explorer
// when he try to copy file to recycler.
//

//
// Add local path to fake recycler.
//

void RecyclerAdd(const wchar_t *local)
{
  char remote[MAX_PATH];
  
  GetFullRemotePath(remote, local, sizeof(remote));
  
  Recycler.insert(remote);
  
  RecyclerWakeUp = RECYCLER_BIN_CYCLES;
}

//
// Remove local path from fake recycler.
//

void RecyclerDel(const wchar_t *local)
{
  char remote[MAX_PATH];
  
  GetFullRemotePath(remote, local, sizeof(remote));
  
  Recycler.erase(remote);
}

//
// Check does given local path exist in recycler.
//

Int RecyclerExists(const wchar_t *local)
{
  char remote[MAX_PATH];
  
  GetFullRemotePath(remote, local, sizeof(remote));
  
  return Recycler.count(remote);
}

//
// Delete files from recycled bin for given root.
// We need it because reparse point (e.g. c:/mount)
// uses the same rules as parent drive (e.g. C:).
// We list whole recycled bin's content and checks
// manually original location. If it matches with
// given root we delete file physically.
//
// root - root path for original location (before delete) (IN/OPT).
//
// If no root specified current local mount point is used.
//
// RETURNS: 0 if OK.
//

Int RecyclerFlush(const wchar_t *rootIn)
{
  DBG_ENTER("RecyclerFlush");
  
  int exitCode = -1;

  HRESULT hr = S_OK;

  LPSHELLFOLDER desktopFolder = NULL;

  LPSHELLFOLDER2 recycleFolder = NULL;
  
  LPITEMIDLIST recycleList = NULL;
  
  LPITEMIDLIST pidl = NULL;
  
  LPENUMIDLIST enumFiles = NULL;
  
  wchar_t *root = (wchar_t *) rootIn;
  
  //
  // Dont go in if nothing deleted recently.
  //
  
  if (RecyclerWakeUp > 0)
  {
    if (root == NULL)
    {
      root = Options.localRoot_;
    }
  
    int rootLen = wcslen(root);

    //
    // Get desktop folder.
    //

    DBG_MSG("Retrieving virutal desktop folder...");
  
    HFAIL(SHGetDesktopFolder(&desktopFolder));
  
    //
    // Get recycler bin from desktop.
    //
  
    DBG_MSG("Retrieving virutal recycler bin folder...");
  
    HFAIL(SHGetSpecialFolderLocation(NULL, CSIDL_BITBUCKET, &recycleList));

    //
    // Cast recycler bin to ShellFolder interface.
    //

    DBG_MSG("Casting recycler bin to IShellFolder...");
   
    HFAIL(desktopFolder -> BindToObject(recycleList, NULL,
                                           _IID_IShellFolder, 
                                               (LPVOID *) &recycleFolder));

    //
    // Iterate through list.
    //

    DBG_MSG("Listing recycler's content...");
  
    HFAIL(recycleFolder -> EnumObjects(NULL, SHCONTF_FOLDERS | SHCONTF_NONFOLDERS |
                                           SHCONTF_INCLUDEHIDDEN, &enumFiles));

    //
    // Enumerate content of recycled bin.
    //
  
    while (enumFiles -> Next(1, &pidl, NULL) != S_FALSE)
    {
      SHELLDETAILS sd;

      wchar_t *original = NULL;
      wchar_t *current  = NULL;

      STRRET currentStrRet;

      //
      // Check original location.
      //

      recycleFolder -> GetDetailsOf(pidl, 1, &sd);

      StrRetToStrW(&sd.str, pidl, &original);

      if (original[0])
      {
        //
        // Check does original location matches with pattern.
        //

        NormalizePath(original, wcslen(original), L'/');

        if (original && wcsnicmp(original, root, rootLen) == 0)
        {
          //
          // Original location matches.
          // Get physical 'current' path.
          //

          recycleFolder -> GetDisplayNameOf(pidl, SHGDN_FORPARSING, &currentStrRet);

          //
          // And remove it from disk.
          //

          if (StrRetToStrW(&currentStrRet, pidl, &current) == 0)
          {
            DBG_MSG("RecyclerFlush : Removing [%ls]...", current);

            if (GetFileAttributesW(current) & FILE_ATTRIBUTE_DIRECTORY)
            {
              RemoveDirectoryW(current);
            }
            else
            {
              DeleteFileW(current);
            }

            CoTaskMemFree(current);
          }
        }
      }

      CoTaskMemFree(original);
      CoTaskMemFree(pidl);
    }

    RecyclerWakeUp --;
  }

  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    fprintf(stderr, "ERROR: Cannot flush recycled bin.\n"
                "Error code is : %d.\nHresult code is 0x%x.\n", 
                    (Int) GetLastError(), (Unsigned Int) hr);
  }

  if (desktopFolder != NULL)
  {
    desktopFolder -> Release();
  }
  
  if (recycleFolder != NULL)
  {
    recycleFolder -> Release();
  }
  
  if (recycleList != NULL)
  {
    CoTaskMemFree(recycleList);
  }
  
  if (enumFiles != NULL)
  {
    enumFiles -> Release();
  }
  
  DBG_LEAVE("RecyclerFlush");
  
  return exitCode;
}

//
// Matching wildcards
//

Int WINAPI DokanFillFileData(PWIN32_FIND_DATAW FindData, PDOKAN_FILE_INFO  FileInfo)
{
  /*
  FIXME: Wildcards matching not implemented
  */

  return 0;
}

//
// Translate relative (to mounted drive) path to absolute path on server.
//
// fullPath   - full, remote path on server side (OUT).
// path       - path on local mounted drive (IN).
// size       - size of fullBath buffer in bytes (IN).
// capitalize - capitalize local path before cache (IN/OPT).
// overwrite  - forve overwrite of cached name (IN/OPT).
//

void GetFullRemotePath(Char *fullPath, const wchar_t *path, Int size)
{
  GetFullRemotePath(fullPath, (wchar_t *) path, size);
}

void GetFullRemotePath(Char *remotePath, wchar_t *localPath, Int size, Int overwrite)
{
  CachedPathsEntry cachedEntry;
  
  IdleCountReset();

  #ifdef USE_CASE_SENSITIVE_MAP
  
  //
  // If real case sensitive name on server side
  // already known use it. This is case, when target path
  // already listed from server on created by client side.
  // Server side devices what is real name in this case.
  //
  
  if (Not(overwrite) && CacheLocalStat(localPath, &cachedEntry) == 0)
  {
    StringSizeCopy(remotePath, cachedEntry.remotePath_.c_str(), size);

    DBG_MSG("GetFullRemotePath : Reusing [%ls] |-> [%s]...", localPath, remotePath);
  }
  
  //
  // Real, case sensitive path on server side is not known
  // yet. It means it's a new path created only on client side.
  // Client side decides what is real name in this case.
  //

  else

  #endif /* USE_CASE_SENSITIVE_MAP */
  
  {
    if (remotePath && localPath)
    {
      //
      // Dokan uses UTF16, but we need to send UTF8 to server.
      //
      
      Char temp[MAX_PATH] = {0};

      WideCharToMultiByte(CP_UTF8, 0, localPath, -1, temp, sizeof(temp), NULL, NULL);

      WideCharToMultiByte(CP_UTF8, 0, Options.remoteRoot_, -1, remotePath,
                              size, NULL, NULL);

      //
      // remotePath := <remote root>/<path> in UTF8.
      //
      
      StringSizeCat(remotePath, temp, size);
      
      Int len = StringLength(remotePath);
  
      //
      // Avoid mixing '/' and '\\' style. 
      //
    
      for (Int i = 0; i < len; i++)
      {
        if (remotePath[i] == '\\')
        {
          remotePath[i] = '/';
        }
      }
    
      //
      // Remove last '/' from path if exist.
      //
    
      while(len > 1 && remotePath[len - 1] == '/')
      {
        remotePath[len - 1] = 0;
  
        len --;
      }
    
      //
      //
      //
    
      if (strlen(remotePath) == 2 && remotePath[1] == ':')
      {
        strcat(remotePath, "/");
      }

      #ifdef USE_CASE_SENSITIVE_MAP
      
      //
      // Save real case-sensitive name.
      //

      CacheAddPath(localPath, remotePath);

      #endif /* USE_CASE_SENSITIVE_MAP */
    }
  }
}

//
// I/O Callbacks. We redirect I/O operations to Sftp client here.
//

//
// Open remote file. Arguments are the same as WINAPI CreateFile.
// See msdn for more info.
//
// path       - path to remote file (IN).
// accesMode  - read/write flag - see CreateFile on msdn (IN).
// sharedMode - usunsed.
// create     - create disposition flag - see CreateFile on msdn (IN).
// flags      - unused.
// dokanInfo  - handle to opened file (OUT).
//
// RETURNS: 0 if OK.
//

int DOKAN_CALLBACK NxCreateFile(LPCWSTR path, DWORD accessMode, DWORD sharedMode,
                                    DWORD create, DWORD flags,
                                        PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("CreateFile");

  Int status = 0;
  
  Int isDir = dokanInfo -> IsDirectory;
  
  Int skip = 0;
  
  DBG_MSG("CreateFile([%ls][ac=0x%x][sh=0x%x][cr=0x%x][fl=0x%x][%d])",
              path, accessMode, sharedMode, create, flags, isDir);

  //
  // Workaround to skip often call for Desktop.ini file.
  //

  #ifdef SKIP_DESKTOP_INI
  {
    wchar_t *p = NULL;
    
    if ((p = StrStrIW(path, L"\\Desktop.ini")))
    {
      p += sizeof("\\Desktop.ini") - 1;
      
      if (p[0] == 0)
      {
        DBG_MSG("CreateFile : 'Desktop.ini' skipped.");
  
        skip = 1;

        status = ERROR_FILE_NOT_FOUND;
      }  
    }
  } 
  #endif 

  //
  // Workaround to avoid copying files to recycler bin.  
  //
  
  if (RecyclerExists(path))
  {
    dokanInfo -> Context = 0;
  
    skip = 1;    
  }
  
  //
  // Save DELETE_ON_CLOSE flag.
  // Windows 8 doesn't use DeleteFile() callback.
  // He uses DELETE_ON_CLOSE flag instead.
  //

/*
 * FIXME: Modify kernel driver instead.
 */
  
  if (flags & FILE_FLAG_DELETE_ON_CLOSE)
  {
    DBG_MSG("CreateFile : Setting DELETE_ON_CLOSE for [%ls]...", path);
    
    char remote[MAX_PATH];
    
    GetFullRemotePath(remote, path, sizeof(remote));
  
    DeleteOnClose.insert(remote);
  }   
  
  //
  // Filter spam for opened directory from explorer.exe.
  //
  
/*
 * FIXME: Limit to 'explorer.exe' process only. 
 *        Trick may fails for other applications.
 */
/*  
  if (wcscmp(path, L"\\") == 0 || 
          PidCurrentDir[dokanInfo -> ProcessId] == path)
  {
    DBG_MSG("CreateFile : Skipping [%ls] directory for pid [%d].",
                path, dokanInfo, dokanInfo -> ProcessId);

    dokanInfo -> Context = 0;
    
    status = ERROR_SUCCESS;

    skip = 1;
  }
*/  
  if (Not(skip))
  {
    dokanInfo -> Context = CacheInsertHandle(path, accessMode,
                                                   create, 0, &status);
  }
  
  if (dokanInfo -> Context & DIR_HANDLE_MASK)
  {
    dokanInfo -> IsDirectory = TRUE;
  }

  DBG_MSG("Callback [%d][CreateFile][%ls][%x][%x][%x][%x] finished with status : %d.", 
              dokanInfo -> ProcessId, path, accessMode, sharedMode, create, flags, -status);

  DBG_LEAVE("CreateFile");

  return -status;
}

//
// Create remote directory.
//
// path      - relative remote path (IN).
// dokanInfo - unused.
//
// RETURNS. 0 if OK.
//

int DOKAN_CALLBACK NxCreateDirectory(LPCWSTR path, PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("CreateDirectory");

  Char fullPath[MAX_PATH] = {0};
  
  GetFullRemotePath(fullPath, path, MAX_PATH);

  DBG_MSG("CreateDirectory : path = [%s]", fullPath);
  
/*
  FIXME: Cache directory content if possible.

  CacheClearChildsList(path);
*/

  Int ret = Sftp.mkdir(fullPath);
 
  DBG_MSG("Callback [%d][CreateDirectory][%ls] finished with code [%d].", 
              dokanInfo -> ProcessId, path, -ret);

  DBG_LEAVE("CreateDirectory");

  return -ret;
}

//
// Open remote directory.
//
// path      - relative remote path to open (IN).
// dokanInfo - handle to opened dir (OUT).
//
// RETURNS: 0 if OK.
//

int DOKAN_CALLBACK NxOpenDirectory(LPCWSTR path, PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("OpenDirectory");

  Int status = 0;
  
  dokanInfo -> Context = CacheInsertHandle(path, 0, 0, 1, &status);
  
  dokanInfo -> IsDirectory = TRUE;

  if (status == 0 && wcscmp(path, L"\\"))
  {
    PidCurrentDir[dokanInfo -> ProcessId] = path;
  }
  
  DBG_MSG("Callback [%d][OpenDirectory][%ls] finished with code [%d].", 
              dokanInfo -> ProcessId, path, -status);  

  DBG_LEAVE("OpenDirectory");

  return -status;
}

//
// Close handle to remote resource (received from OpenDirectory or CreateFile).
//
// path      - unused.
// dokanInfo - handle to close (IN);
//
// RETURNS: 0 if OK.
//

int DOKAN_CALLBACK NxCloseFile(LPCWSTR path, PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("CloseFile");
    
  Int exitCode = -1;
 
  Int64 handle = (Int64) dokanInfo -> Context;
  
  FAIL(dokanInfo -> Context == 0);

  DBG_MSG("CloseFile : Closing remote handle [%d]...", handle);

  IdleCountReset();
  
  CacheCloseHandle(handle);
  
  dokanInfo -> Context = 0;
  
  //
  // Clean up.
  //
  
  exitCode = 0;
    
  if (exitCode)
  {
    DBG_MSG("CloseFile : ERROR: Cannot close remote handle."
                " Error code is %u.", GetLastError());
  }
  
  fail:
  
  DBG_MSG("Callback [%d][CloseFile][%ls] finished with code [%d].", 
              dokanInfo -> ProcessId, path, 0);
  
  DBG_LEAVE("CloseFile");

  return 0;
}

//
// Read data from remote file.
//
// path      - unused.
// buffer    - destination buffer (OUT).
// size      - how much bytes to read (IN).
// readed    - how much bytes readed after call (OUT).
// offset    - file's offset of first byte to read (IN).
// dokanInfo - handle to remote file received from CreateFile (IN).
//

int DOKAN_CALLBACK NxReadFile(LPCWSTR path, LPVOID buffer, DWORD size,
                                  LPDWORD readed, LONGLONG offset,
                                      PDOKAN_FILE_INFO dokanInfo)
{                    
  DBG_ENTER("ReadFile");
  
  Int ret   = -1;
  Int isDir = dokanInfo -> IsDirectory;

  Int64 handle = (Int64) dokanInfo -> Context;
  
  DBG_MSG("ReadFile : path = [%ls], buffer = [%p], bufferSize = [%d],"
              "offset = [%lld], context = [%lld], isDir = [%d]",
                  path, buffer, size, offset, handle, isDir);
                  
  //
  // Workaround to avoid copying files to recycler bin.  
  //
  
  if (RecyclerExists(path))
  { 
    *readed = 0;
    
    ret = 0;
  }
    
  else
  {
    //
    // Filter spam for opened directory from explorer.exe.
    //
  
/*
 * FIXME: Limit to 'explorer.exe' process only. 
 *        Trick may fails for other applications.
 */
  
    if (wcscmp(path, L"\\") == 0 || PidCurrentDir[dokanInfo -> ProcessId] == path)
    {
      DBG_MSG("CreateFile : Path [%ls] is main directory for pid [%d].\n",
                  path, dokanInfo, dokanInfo -> ProcessId);

      goto fail;
    }

    FAIL(handle == 0);
    FAIL(buffer == NULL);
    FAIL(size == 0L);
    FAIL(isDir);
    FAIL(path[0] == L'\\' && path[1] == 0);
  
    IdleCountReset();
  
    CacheEvaluateHandle(dokanInfo -> Context);

    *readed = Sftp.read(dokanInfo -> Context, (Char *) buffer, offset, size);

    if (Int(*readed) >= 0)
    {
      ret = 0;
    }
    else
    {
      ret = -ERROR_ACCESS_DENIED;
    }
  }
  
  fail:
  
  DBG_MSG("Callback [%d][ReadFile][%ls][%d] finished with code [%d].", 
              dokanInfo -> ProcessId, path, size, ret);
  
  DBG_LEAVE("ReadFile");
  
  return ret;
}

//
// Write data to remote file.
//
// path      - unused.
// buffer    - buffer with data to write (IN).
// size      - how much bytes to write (IN).
// readed    - how much bytes written after call (OUT).
// offset    - file's offset of first byte to write (IN).
// dokanInfo - handle to remote file received from CreateFile (IN).
//

int DOKAN_CALLBACK NxWriteFile(LPCWSTR path, LPCVOID buffer, DWORD size,
                                   LPDWORD written, LONGLONG offset,
                                       PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("WriteFile");
  
  Int ret   = -1;
  Int isDir = dokanInfo -> IsDirectory;
  
  Int64 handle = (Int64) dokanInfo -> Context;
  
  DBG_MSG("WriteFile : path = [%ls], buffer = [%p], bufferSize = [%d],"
              " offset = [%lld], context = [%llx]", 
                  path, buffer, size, offset, handle);
  
  FAIL(handle == 0);
  FAIL(buffer == NULL);
  FAIL(size == 0);
  FAIL(isDir);
  FAIL(path[0] == L'\\' && path[1] == 0);
  
  IdleCountReset();

  CacheEvaluateHandle(dokanInfo -> Context);

  *written = Sftp.write(dokanInfo -> Context, (Char *) buffer, offset, size);

  if (Int(*written) > 0)
  {
    ret = 0;
  }
  else
  {
    ret = -ERROR_ACCESS_DENIED;
  }

  fail:
  
  DBG_MSG("Callback [%d][WriteFile][%ls][%d] finished with code [%d].", 
              dokanInfo -> ProcessId, path, size, ret);
  
  DBG_LEAVE("WriteFile");

  return ret;
}

//
// Returns info about remote file.
//
// path      - path to remote file. (IN)
// fileInfo  - info about file. (OUT)
// dokanInfo - unused.
//
// RETURNS: 0 if OK.
//

Int DOKAN_CALLBACK NxGetFileInformation(LPCWSTR path, 
                                            LPBY_HANDLE_FILE_INFORMATION fileInfo,
                                                PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("GetFileInformation");

  Int exitCode = -2;
  
  Char fullPath[MAX_PATH] = {0};

  //
  // Workaround to avoid copying files to recycler bin.
  // If file already market to delete (from explorer.exe)
  // show fake 0 byte size.
  //
  
  if (RecyclerExists(path))
  {
    ZeroMemory(fileInfo, sizeof(BY_HANDLE_FILE_INFORMATION));
    
    exitCode = ERROR_SUCCESS;
  }
  
  //
  // In general case stat server to get info.
  //
  
  else
  {
    //
    // Create full remote path.
    //
  
    GetFullRemotePath(fullPath, path, MAX_PATH);

    DBG_MSG("GetFileInformation : {path = [%s]}.", fullPath);

    exitCode = Sftp.stat(fullPath, fileInfo);

    //
    // Set hidden attribute
    //

    if (fileInfo != NULL && path[0] == '.')
    {
      fileInfo -> dwFileAttributes |= FILE_ATTRIBUTE_HIDDEN;
    }
  }
  
  DBG_MSG("Callback [%d][GetFileInformation][%ls] finished with code [%d].", 
              dokanInfo -> ProcessId, path, exitCode);
  
  DBG_LEAVE("GetFileInformation");
  
  return exitCode;
}

//
// List files in remote directory.
//
// path        - local client side root directory to list (IN).
// fillFinData - list with files found in remote path (OUT).
// dokanInfo   - handle to remote dir opened via OpenDirectory(). (IN).
//
// RETURNS: 0 if OK.
//

int DOKAN_CALLBACK NxFindFiles(LPCWSTR path, PFillFindData fillFindData,
                                   PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("FindFiles");
  
  Int exitCode = -1;
  
  Char serverRoot[MAX_PATH] = {0};
  Char serverElem[MAX_PATH] = {0};
  Char serverPath[MAX_PATH] = {0};
    
  wchar_t clientRoot[MAX_PATH] = {0};
  wchar_t clientPath[MAX_PATH] = {0};
  
  WIN32_FIND_DATAW *list = NULL;
  
  Int count = 0;

  Int64 handle = dokanInfo -> Context;

  FAIL(handle == 0);  
  
  CacheEvaluateHandle(handle);
  
  //
  // Retrieve from server only if childs not cached yet.
  //

/*  
  FIXME: Use cache if possible.
  
  if (CacheGetChildsList(handle, &count, &list))
*/

  {
    GetFullRemotePath(serverRoot, path, MAX_PATH);

    DBG_MSG("FindFiles : Remote path = [%s], handle = [%d].", serverRoot, handle);

    //                                                
    // Retrieve list of remote files.
    //
  
    DBG_MSG("FindFiles : Retreaving remote list...");
  
    FAIL(Sftp.readDir(&list, &count, handle));
  
    //
    // Cache result.
    //

/*  
  FIXME: Use cache if possible.
    
    CacheSetChildsList(handle, count, list);
*/    
  }

  //
  // Prepare serverRoot and clientRoot with correct
  // '/' or '\\' suffix.
  //
  
  strncat(serverRoot, "/", MAX_PATH);

  wcsncpy(clientRoot, path, MAX_PATH);
  
  if (clientRoot[0] && clientRoot[1])
  {
    wcsncat(clientRoot, L"\\", MAX_PATH);
  }  
  
  //
  // Set locale to correct processing of
  // UTF16 strings with localized characters.
  //
  
  setlocale(LC_ALL, "");

  //
  // Put whole list to output.
  //
  
  if (count > 0)
  {
    for (Int i = 0; i < count; i++)
    {
      //
      // Create UTF8 case sensitive element name.
      //
      
      WideCharToMultiByte(CP_UTF8, 0, list[i].cFileName, -1, 
                              serverElem, MAX_PATH, NULL, NULL);

      //
      // Create full case sensitive server path.
      // serverRoot/serverElem.
      //
      
      strncpy(serverPath, serverRoot, MAX_PATH);
      strncat(serverPath, serverElem, MAX_PATH);

      //
      // Create full client case non-sensitive path.
      //

      wcsncpy(clientPath, clientRoot, MAX_PATH);
      wcsncat(clientPath, list[i].cFileName, MAX_PATH);
      
      #ifdef USE_CASE_SENSITIVE_MAP
      
      //
      // Save real case sensitive name related to non-case 
      // sensitive client name.
      // 'CLIENT NON CASE SENSITIVE' |-> 'Server case sensitive'.
      //

      CacheAddPath(clientPath, serverPath);
      
      #endif /* USE_CASE_SENSITIVE_MAP */

      //
      // Set hidden attribute.
      //

      if (list[i].cFileName[0] == '.')
      {
        list[i].dwFileAttributes |= FILE_ATTRIBUTE_HIDDEN;
      }

      //
      // Pass element to dokan.
      //

      if (fillFindData != NULL)
      {
        fillFindData(&list[i], dokanInfo);
      }  
    }
  }
  
  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    DBG_MSG("FindFiles : ERROR. Cannot find remote files."
                " Error code is %u.", GetLastError());
  }
  
  DBG_MSG("Callback [%d][FindFiles][%ls] finished with code [%d].", 
              dokanInfo -> ProcessId, path, exitCode);
  
  if (list != NULL)
  {
    free(list);
  }
  
  DBG_LEAVE("FindFiles");
  
  return exitCode;
}

//
// Delete remote file.
//
// path      - relative remote path to remove (IN).
// dokanInfo - unused.
//
// RETURNS: 0 if OK.
//

int DOKAN_CALLBACK NxDeleteFile(LPCWSTR path, PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("DeleteFile");

  Char remote[MAX_PATH] = {0};
  
  DBG_MSG("DeleteDirectory : pathname = [%ls]", path);

  DBG_MSG("Callback [%d][DeleteFile][%ls] finished with code [%d].", 
              dokanInfo -> ProcessId, path, 0);
              
  GetFullRemotePath(remote, path, MAX_PATH);
  
  RecyclerAdd(path);
  
  Sftp.remove(remote);
    
  DBG_LEAVE("DeleteFile");

  return 0;
}

//
// Delete remote directory.
//
// path      - relative remote path to remove (IN).
// dokanInfo - unused.
//
// RETURNS: 0 if OK.
//

int DOKAN_CALLBACK NxDeleteDirectory(LPCWSTR path, PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("DeleteDirectory");

  WIN32_FIND_DATAW *list = NULL;
  
  Int ret   = -1;
  Int count = 0;

  Char fullPath[MAX_PATH] = {0};
  
  GetFullRemotePath(fullPath, path, MAX_PATH);

  DBG_MSG("DeleteDirectory : pathname = [%ls]", path);
  
  //
  // Get directory handle.
  //
  
  Int64 handle = dokanInfo -> Context;

  FAIL(handle == 0);
  
  CacheEvaluateHandle(handle);
  
/*
  FIXME: Cache directory content if possible.

  CacheClearChildsList(path);
*/  
  
  //
  // Count content of directory if exist.
  //
  
  if (Sftp.readDir(&list, &count, handle) == 0)
  {
    //
    // If directory empty we can remove it immediately, 
    // pass success code to caller.
    //

    if (count <= 2)
    {
      ret = ERROR_SUCCESS;
    
      PidCurrentDir[dokanInfo -> ProcessId] = L"";
      
      RecyclerAdd(path);
      
      Sftp.remove(fullPath);
    }
    
    //
    // Directory not empty, we can't remove it immediately,
    // pass ERROR_DIR_NOT_EMPTY code to caller.
    //
    
    else
    {
      ret = ERROR_DIR_NOT_EMPTY;
    }
  }
  
  //
  // Dir not exist.
  //
  
  else
  {
    ret = ERROR_FILE_NOT_FOUND;
  }
  
  CacheCloseHandlesByPath(path, 1);
  
  //
  // Clean up.
  //
  
  fail:
  
  if (list != NULL)
  {
    free(list);
  }
  
  DBG_MSG("Callback [%d][DeleteDirectory][%ls] finished with code [%d].", 
              dokanInfo -> ProcessId, path, ret);
  
  DBG_LEAVE("DeleteDirectory");

  return -ret;
}

//
// Rename file1 to file2.
//

int DOKAN_CALLBACK NxMoveFile(LPCWSTR path1, LPCWSTR path2,
                                  BOOL replaceExisting, PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("MoveFile");

  Int goOn = 1;
  
  Char fullPath1[MAX_PATH] = {0};
  Char fullPath2[MAX_PATH] = {0};
  
/*
  wchar_t parentOld[MAX_PATH] = L"\\";
  
  Int lastSlash = 0;

  for (Int i = 0; path1[i]; i++)
  {
    if (path1[i] == L'/' || path1[i] == L'\\')
    {
      lastSlash = i;
    }
  }
  
  if (lastSlash > 0)
  {
    wcsncmp(parentOld, path1, lastSlash);
  
    DBG_MSG("Caching old parent's content [%ls]...\n", parentOld); 
  }
*/

  GetFullRemotePath(fullPath1, path1, MAX_PATH);
  GetFullRemotePath(fullPath2, path2, MAX_PATH);
  
  //
  // Try retrieve known part of new path from old one.
  // It fixed scenario when caller wants to rename:
  //
  // 'Some/Already/Cached/OldPath' to 'SOME/ALREADY/CACHED/NewPath'.
  //   ^^^   ^^^  ^^^                   ^^^ ^^^ ^^^ ^^^
  //  Old path already cached          This part of new   
  //  in case sensitive map.           path is already cached,
  //                                   resuse it.
  //
  //
  
  for (Int i = 0; goOn && fullPath1[i] && fullPath2[i]; i++)
  {
    switch(fullPath1[i])
    {
      case 'a'...'z':
      case 'A'...'Z':
      {
        if ((fullPath1[i] | 0x20) == (fullPath2[i] | 0x20))
        {
          fullPath2[i] = fullPath1[i];
        }
        else
        {
          goOn = 0;
        }
      
        break;
      }
/*      
      case '/':
      case '\\':
      {
        if (fullPath2[i] != '/' && fullPath2[i] != '\\')
        {
          goOn = 0;
        }

        break;
      }
*/      
      
      default:
      {
        if (fullPath1[i] != fullPath2[i])
        {
          goOn = 0;
        }
      }
    }
  }
  
/*
  FIXME: Cache directory content if possible.
  
  CacheClearChildsList(path1);
  CacheClearChildsList(path2);
*/

  CacheCloseHandlesByPath(path1, 1);
  
  Int ret = Sftp.rename(fullPath1, fullPath2);

  DBG_MSG("MoveFile :: [path1 = [%s], path2 = [%s], ctx = [%lld]", 
              fullPath1, fullPath2, dokanInfo -> DokanContext);
  
  DBG_MSG("Callback [%d][MoveFile][%ls][%ls] finished with code [%d].", 
              dokanInfo -> ProcessId, path1, path2, ret);
  
  DBG_LEAVE("MoveFile");
  
  return ret;
}

//
// CleanUp callback.
//

int DOKAN_CALLBACK NxCleanup(LPCWSTR path, PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("Cleanup");
  
  Int exitCode = -1;

  Char remote[MAX_PATH];
  
  //
  // Clean up resource.
  //
  
  if (dokanInfo -> Context)
  {
    //
    // Check for own DELTE_ON_CLOSE flag.
    // We save it due to explorer on Windows 8.
    //

/*
 * FIXME: Modify kernel driver instead.
 */
    
    GetFullRemotePath(remote, path, sizeof(remote));
  
    if (DeleteOnClose.count(remote))
    {
      DeleteOnClose.erase(remote);
    
      dokanInfo -> DeleteOnClose = TRUE;
    }   

    //
    // Delete object on server if needed.
    //
    
    if (dokanInfo -> DeleteOnClose)
    {
      //
      // Remove file from our own virtual recycler.
      // It breaks giving fake 0 bytes len to explorer.
      //
      
      RecyclerDel(path);
      
      //
      // When we get here, explorer.exe already copied fake
      // 0 byte file to recycler bin. Delete this file from
      // disk.
      //
      
      RecyclerFlush(Options.localRoot_);
      
      //
      // Get full server side path.
      //
  
      Char fullPath[MAX_PATH] = {0};
    
      GetFullRemotePath(fullPath, path, MAX_PATH);

      DBG_MSG("Cleanup : pathname = [%s]", fullPath);
    
      //
      // Free object on server side.
      // 
      
      CacheCloseHandlesByPath(path, 1);
      
      //
      // Remove object on server.
      //
      
      FAIL(Sftp.remove(fullPath));
          
      dokanInfo -> Context = 0;
    }
  }
  
  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:

  DBG_MSG("Callback [%d][Cleanup][%ls][%d] finished with code [%d].", 
              dokanInfo -> ProcessId, path, dokanInfo -> DeleteOnClose, exitCode);
  
  DBG_LEAVE("Cleanup");
  
  return exitCode;
}


//
// FIXME. Implementation needed.
//

int DOKAN_CALLBACK NxSetFileTime(LPCWSTR path, CONST FILETIME *creatTime,
                                     CONST FILETIME *lastAccessTime,
                                         CONST FILETIME *lastWriteTime,
                                             PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("SetFileTime");

  DBG_MSG("Callback [%d][SetFileTime][%ls] not implemented.", 
              dokanInfo -> ProcessId, path);
  
  DBG_LEAVE("SetFileTime");
  
  return 0;
}

//
// Below functions are unimplemented. For compatibility only.
//

int DOKAN_CALLBACK NxUnlockFile(LPCWSTR path, LONGLONG offset,
                                    LONGLONG length, PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("UnlockFile");

  DBG_MSG("Callback [%d][UnlockFile][%ls] not implemented.", 
              dokanInfo -> ProcessId, path);
  
  DBG_LEAVE("UnlockFile");
  
  return 0;
}

//
// FIXME. Should we need this? We can try to map unix like attributes
// to basic access list from winapi.
//

int DOKAN_CALLBACK NxGetFileSecurity(LPCWSTR path, PSECURITY_INFORMATION si,
                                         PSECURITY_DESCRIPTOR sd,
                                             ULONG bufferSize, PULONG sizeNeeded,
                                                 PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("GetFileSecurity");

  LPTSTR userSidString  = NULL;
  LPTSTR groupSidString = NULL;
  LPTSTR pStringBuffer  = NULL;

  Char buffer[1024];
  Char finalBuffer[2048];

  Int ret = 0;

  Char *errorMessage = NULL;

  HANDLE tokenHandle = INVALID_HANDLE_VALUE;

  PSECURITY_DESCRIPTOR secDescrTmp = NULL;

  if (GetFileSecurityW(Options.localRoot_, *si, sd, bufferSize, sizeNeeded) == FALSE)
  {
    if (GetLastError() == ERROR_INSUFFICIENT_BUFFER && *sizeNeeded > bufferSize)
    {
      ret = - ERROR_INSUFFICIENT_BUFFER;

      FAIL_MSG("GetFileSecurityW");
    }

    if (!OpenProcessToken(GetCurrentProcess(), TOKEN_READ, &tokenHandle))
    {
      FAIL_MSG("OpenProcess");
    }

    DWORD returnLength;

    if (!GetTokenInformation(tokenHandle, TokenUser, buffer, sizeof(buffer),
            &returnLength))
    {
      FAIL_MSG("GetTokenInformation(TokenUser)");
    }

    PTOKEN_USER userToken = (PTOKEN_USER) buffer;

    if (!ConvertSidToStringSid(userToken -> User.Sid, &userSidString))
    {
      FAIL_MSG("ConvertSidToStringSid(User sid)");
    }

    if (!GetTokenInformation(tokenHandle, TokenGroups, buffer, sizeof(buffer),
            &returnLength))
    {
      FAIL_MSG("GetTokenInformation(TokenGroups)");
    }

    PTOKEN_GROUPS groupsToken = (PTOKEN_GROUPS) buffer;

    if (groupsToken->GroupCount > 0)
    {
      if (!ConvertSidToStringSid(groupsToken->Groups[0].Sid, &groupSidString))
      {
        FAIL_MSG("ConvertSidToStringSid(Group sid)");
      }

      StringSizePrint(buffer, 1024, "O:%sG:%s", userSidString, groupSidString);
    }
    else
    {
      StringSizePrint(buffer, 1024, "O:%s", userSidString);
    }

    //
    // Initialize basic access control list.
    //
    
    if (dokanInfo -> IsDirectory)
    {
      StringSizePrint(finalBuffer, 2048, "%sD:PAI(A;OICI;FA;;;AU)", buffer);
    }
    else
    {
      StringSizePrint(finalBuffer, 2048, "%sD:AI(A;ID;FA;;;AU)", buffer);
    }

    ULONG size = 0;

    if (!ConvertStringSecurityDescriptorToSecurityDescriptor(finalBuffer,
            SDDL_REVISION_1, &secDescrTmp, &size))
    {
      FAIL_MSG("ConvertSecurityDescriptorToStringSecurityDescriptor(finalBuffer)");
    }

    if (!ConvertSecurityDescriptorToStringSecurityDescriptor(secDescrTmp,
            SDDL_REVISION_1, *si, &pStringBuffer, NULL))
    {
      FAIL_MSG("ConvertSecurityDescriptorToStringSecurityDescriptor");
    }

    LocalFree(secDescrTmp);

    secDescrTmp = NULL;

    size = 0;

    if (!ConvertStringSecurityDescriptorToSecurityDescriptor(pStringBuffer,
            SDDL_REVISION_1, &secDescrTmp, &size))
    {
      FAIL_MSG("ConvertStringSecurityDescriptorToSecurityDescriptor(pStringBuffer)");
    }

    *sizeNeeded = size;

    if (size > bufferSize)
    {
      ret = - ERROR_INSUFFICIENT_BUFFER;

      FAIL_MSG("copying security descriptor to sd");
    }

    MemoryCopy(sd, secDescrTmp, size);
  }

  fail:

  if (errorMessage == NULL)
  {
    DBG_MSG("Callback [%d][GetFileSecurity][%ls] finished with code [%d].", 
                dokanInfo -> ProcessId, path, ret);
  }
  else
  {
    if (ret == 0)
    {
      ret = - ERROR_CALL_NOT_IMPLEMENTED;
    }

    DBG_MSG("Callback [%d][GetFileSecurity][%ls] - [%ls] failed with code [%d]",
                dokanInfo -> ProcessId, path, errorMessage, GetLastError());

    StringReset(errorMessage);
  }

  if (userSidString != NULL)
  {
    LocalFree(userSidString);    
  }

  if (groupSidString != NULL)
  {
    LocalFree(groupSidString);    
  }

  if (tokenHandle != INVALID_HANDLE_VALUE)
  {
    CloseHandle(tokenHandle);
  }

  if (pStringBuffer != NULL)
  {
    LocalFree(pStringBuffer);
  }

  if (secDescrTmp != NULL)
  {
    LocalFree(secDescrTmp);
  }

  DBG_LEAVE("GetFileSecurity");

  return ret;
}


int DOKAN_CALLBACK NxSetFileSecurity(LPCWSTR path, PSECURITY_INFORMATION si,
                                         PSECURITY_DESCRIPTOR sd,
                                             ULONG sdLength,
                                                 PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("SetFileSecurity");

  DBG_MSG("Callback [%d][SetFileSecurity][%ls] not implemented.",
              dokanInfo -> ProcessId, path);
  
  DBG_LEAVE("SetFileSecurity");
  
  return 0;
}


int DOKAN_CALLBACK NxGetDiskFreeSpace(PULONGLONG freeBytesAvailable,
                                          PULONGLONG totalNumberOfBytes,
                                              PULONGLONG totalNumberOfFreeBytes,
                                                  PDOKAN_FILE_INFO dokanInfo)
{
  SftpStatvfsReplyMsg stvfs;
  
  Char remotePath[MAX_PATH] = {0};
  
  GetFullRemotePath(remotePath, L"/", MAX_PATH);

  
  if (Sftp.statvfs(&stvfs, remotePath) == 0)
  {

    *freeBytesAvailable     = stvfs.bsize_ * stvfs.bavail_;

    *totalNumberOfBytes     = stvfs.bsize_ * stvfs.blocks_;

    *totalNumberOfFreeBytes = stvfs.bsize_ * stvfs.bfree_;
  }
  else
  {
    //
    // 10GB free of 10GB total
    //

    *freeBytesAvailable     = 10 * 1024 * 1024 * 1024UL;

    *totalNumberOfBytes     = 10 * 1024 * 1024 * 1024UL;

    *totalNumberOfFreeBytes = 10 * 1024 * 1024 * 1024UL;
  } 

  return 0;
}


int DOKAN_CALLBACK NxGetVolumeInformation(LPWSTR name, DWORD nameLen,
                                              LPDWORD serial,
                                                  LPDWORD maxComponentLen,
                                                      LPDWORD fileSystemFlag,
                                                          LPWSTR fileSystemNameBuffer,
                                                              DWORD fileSystemNameSize,
                                                                  PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("GetVolumeInformation");

  /*
  FIXME: Use secure funtion
  */

  if (nameLen / sizeof(WCHAR) > 10)
  {
    wcscpy(name, L"NoMachine");
  }

  *serial = 0x19831116;

  *maxComponentLen = 256;

  *fileSystemFlag = FILE_CASE_SENSITIVE_SEARCH |
                    FILE_CASE_PRESERVED_NAMES |
                    FILE_SUPPORTS_REMOTE_STORAGE |
                    FILE_UNICODE_ON_DISK;
//                    FILE_PERSISTENT_ACLS;

  if (fileSystemNameSize / sizeof(WCHAR) > 10)
  {
    wcscpy(fileSystemNameBuffer, L"NoMachine");
  }
  
  DBG_MSG("Callback [%d][GetVolumeInformation][%ls] finished with code 0.", 
              dokanInfo -> ProcessId, name);
  
  DBG_LEAVE("GetVolumeInformation");
  
  return 0;
}


int DOKAN_CALLBACK NxUnmount(PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("Unmount");

  DBG_MSG("Callback [%d][Unmount] not implemented.", dokanInfo -> ProcessId);
  
  DBG_LEAVE("Unmount");

  return 0;
}


int DOKAN_CALLBACK NxSetEndOfFile(LPCWSTR path, LONGLONG offset, 
                                      PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("SetEndOfFile");

  DBG_MSG("Callback [%d][SetEndOfFile][%ls] not implemented.", 
              dokanInfo -> ProcessId, path);

  DBG_LEAVE("SetEndOfFile");
  
  return 0;
}


int DOKAN_CALLBACK NxSetAllocationSize(LPCWSTR path, LONGLONG allocSize,
                                           PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("AllocationSize");

  DBG_MSG("Callback [%d][SetAllocationSize][%ls] not implemented.", 
              dokanInfo -> ProcessId, path);

  DBG_LEAVE("AllocationSize");
  
  return 0;
}


int DOKAN_CALLBACK NxLockFile(LPCWSTR path, LONGLONG offset,
                                  LONGLONG length, PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("LockFile");

  DBG_MSG("Callback [%d][LockFile][%ls] not implemented.", 
              dokanInfo -> ProcessId, path);

  DBG_LEAVE("LockFile");
  
  return 0;
}  

//
// FIXME. Should we implement this? 
//

int DOKAN_CALLBACK NxSetFileAttributes(LPCWSTR path, DWORD attrib,
                                           PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("SetFileAttributes");

  DBG_MSG("Callback [%d][SetFileAttributes][%ls] not implemented.", 
              dokanInfo -> ProcessId, path);

  DBG_LEAVE("SetFileAttributes");
  
  return 0;
}  


int DOKAN_CALLBACK NxFlushFileBuffers(LPCWSTR path, PDOKAN_FILE_INFO dokanInfo)
{
  DBG_ENTER("FlushFileBuffers");
  
  DBG_MSG("Callback [%d][FlushFileBuffers][%ls] not implemented.", 
              dokanInfo -> ProcessId, path);

  DBG_LEAVE("FlushFileBuffers");
  
  return 0;
}
