/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXSSHFS, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Debug.h"

extern HANDLE ErrorHandle;

static Int DebugLevel = 0;

//
// Print error message.
//

void DbgError(const Char *fmt, ...)
{
  va_list ap;

  const Int BUFFER_SIZE = 1024 * 16;
  
  Char msg[BUFFER_SIZE];
  
  DWORD written = 0;
  
  va_start(ap, fmt);
  
  vsnprintf(msg, BUFFER_SIZE, fmt, ap);
  
  va_end(ap);
  
  WriteFile(ErrorHandle, msg, strlen(msg), &written, NULL);
  
  FlushFileBuffers(ErrorHandle);
}

//
// All code below is for debug version only.
//

#ifdef DEBUG

static Int DbgDeep = 0;

static Int DbgTreeMode = 0;

static Char DbgLogFile[MAX_PATH] = {0};

int logfd = 2;

//
// Initialize directory path, where debug log will be created.
//
// path - full path to file, where write log (IN).
//

void DbgInit(const Char *path)
{
  if (DebugLevel == 0 || DbgLogFile[0])
  {
    return;
  }
  
  //
  // If path not specified use default 
  // %ALLUSERSPROFILE%/NoMachine/var/log/nxfs.log".
  //
  
  if (path == NULL)
  {
    StringSizeCopy(DbgLogFile, GetLogPath("nxfs.log"), MAX_PATH);
  }
  
  //
  // Otherside use given path.
  //
  
  else
  {
    StringSizeCopy(DbgLogFile, path, MAX_PATH);
  }
  
  //
  // Open log file for write.
  //
  
  if (DbgLogFile[0])
  {
    #ifdef SEPARATE_LOGS

    Char pidStr[32];
    
    StringPrint(pidStr, ".%d", GetCurrentProcessId());

    StringSizeCat(DbgLogFile, pidStr, MAX_PATH);
    
    #endif
    
    if (DbgLogFile[0])
    {
      logfd = _open(DbgLogFile, _O_APPEND | _O_CREAT | _O_RDWR, _S_IWRITE);
    }                                                     
  }
  
  //
  // First test does log work.
  //
  
  DBG_MSG("Log iniciated propertly.\n");
  DBG_MSG("NXFS-" VERSION);
  DBG_MSG("[Build " __DATE__ " " __TIME__ "]");
}

void DbgStop()
{
  if (logfd < 0)
  {
    return;
  }

  _close(logfd);

  logfd = -1;
}

//
// Debug message for function entry.
//

void DbgEnter(const Char *funcName)
{
  DbgMsg(3, "-> %s()...", funcName);

  DbgDeep += 3;
}

//
// Debug message for function leave.
//

void DbgLeave(const Char *funcName)
{
  DbgDeep -= 3;

  DbgMsg(3, "<- %s()...", funcName);
}

//
// Print debug message.
//

void DbgMsg(int level, const Char *fmt, ...)
{
  if (level > DebugLevel)
  {
    return;
  }
  
  SYSTEMTIME st;

  const Int BUFFER_SIZE = 1024 * 16;
  
  Char msg[BUFFER_SIZE] = {0};

  Char timeMsg[BUFFER_SIZE] = {0};

  char timeStr[16];
  
  va_list ap;
  
  if (logfd < 0)
  {
    return;
  }
  
  GetLocalTime(&st);

  snprintf(timeStr, sizeof(timeStr), "%02d:%02d:%02d %03d",
               st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);
  
  /*
  if (DbgTreeMode)
  {
    for (int i = 0; i < DbgDeep; i++)
    {
      StringSizeCat(msg, " ", BUFFER_SIZE);
    }
  }  
  */
  
  va_start(ap, fmt);
  

  if (DbgTreeMode)
  {
    vsnprintf(msg + DbgDeep, BUFFER_SIZE - DbgDeep, fmt, ap);
    
    StringSizeCat(msg + DbgDeep, "\n", BUFFER_SIZE - DbgDeep);
    
    memset(msg, ' ', DbgDeep);
  }
  else
  {
    vsnprintf(msg, BUFFER_SIZE, fmt, ap);
    
    StringSizeCat(msg, "\n", BUFFER_SIZE);
  }  
    
  StringPrint(timeMsg, "[%lu][%lu] %s %s", GetCurrentProcessId(),
                  GetCurrentThreadId(), timeStr, msg);

  _write(logfd, timeMsg, StringLength(timeMsg));
  
  va_end(ap);
}

//
// Dump raw buffer to log.
//

void DbgDumpBuffer(const void *inpBuf, Int size)
{
  if (DebugLevel == 0)
  {
    return;
  }

  const unsigned Char *buf = (unsigned Char *) inpBuf;
  
  if (buf)
  {
    const Int BUFFER_SIZE = 1024 * 512;

    Char msg[BUFFER_SIZE] = {0};
    
    /*
    if (DbgTreeMode)
    {
      for (int i = 0; i < DbgDeep; i++)
      {
        StringSizeCat(msg, " ", BUFFER_SIZE);
      }
    }  
    */
    
    for (Int i = 0; i < size; i++)
    {
      Char tmp[8] = {0};
      
      if (i > 0 && i % 16 == 0)
      {
        StringSizeCat(msg, "\n", BUFFER_SIZE);
                
        if (DbgTreeMode)
        {
          for (int i = 0; i < DbgDeep; i++)
          {
            StringSizeCat(msg, " ", BUFFER_SIZE);
          }
        }
      }
  
      StringSizePrint(tmp, 8, "%02x ", buf[i]);

      StringSizeCat(msg, tmp, BUFFER_SIZE);
    }
    
    DBG_MSG(msg);
  }
}

void DbgDeleteLog()
{
  if (DebugLevel == 0)
  {
    return;
  }

  if (DbgLogFile[0])
  {
    DeleteFile(DbgLogFile);
  }
}

void DbgSetLevel(Int level)
{
  DebugLevel = level;
}

Int DbgGetLevel()
{
  return DebugLevel;
}

#endif
