/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXSSHFS, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "DokanClient.h"

//
// Dokan library.
//

extern OptionsStruct Options;

static HMODULE DokanDll = NULL;

static DokanMainPtr __DokanMain = NULL;

static DokanRemoveMountPointPtr __DokanRemoveMountPoint = NULL;

HANDLE DokanWorkerThread         = NULL;
HANDLE ReparsePointMonitorThread = NULL;
HANDLE DokanUnmounterThread      = NULL;

//
// Translate dokan error code to error message.
//
// code - number of dokan error code (IN).
//
// RETURNS: Error message.
//

const Char *GetDokanError(Int code)
{
  DBG_ENTER("GetDokanError");
  
  const Char *ret = "Unknown error.";
  
  static DokanErrorCode DokanErrors[] =
  {
    {DOKAN_ERROR,                "Unknown error."},
    {DOKAN_DRIVE_LETTER_ERROR,   "Wrong drive letter."},
    {DOKAN_DRIVER_INSTALL_ERROR, "Could not install driver."},
    {DOKAN_START_ERROR,          "Could not start driver."},
    {DOKAN_MOUNT_ERROR,          "Mount error."},
    {DOKAN_MOUNT_POINT_ERROR,    "Mount point error."},
    {0,                          NULL}
  };

  for (Int i = 0; DokanErrors[i].name_; i++)
  {
    if (DokanErrors[i].code_ == code)
    {
      ret = DokanErrors[i].name_;
    
      break;
    }  
  }
  
  DBG_LEAVE("GetDokanError");
  
  return ret;
}

//
// Workaround for ocassionally dokan driver deadlock,
// which causes non-signaled dokan worker hung-up forever.
// We check for reparse flag from time to time and
// going to quit if it's cleared.
//

DWORD WINAPI ReparsePointMonitor(const wchar_t *mountPoint)
{       
  DBG_ENTER("ReparsePointMonitor");

  Int mountPointLive = 1;
  Int parentLive     = 1;
  
  HANDLE process = NULL;
  
  DWORD attr;
  
  if (Options.parentPid_ > 0)
  {
    process = OpenProcess(SYNCHRONIZE, FALSE, Options.parentPid_);
  }
  
  //
  // Monitor is reparse point flag set
  // and parent process if its PID specified.
  //
  
  while(mountPointLive && parentLive)
  {
    Sleep(5000);
    
    attr = GetFileAttributesW(mountPoint);

    DBG_MSG("ReparsePointMonitor : mount point state : [%x].", attr);
    
    if (attr != 0xffffffff &&
            (attr & FILE_ATTRIBUTE_REPARSE_POINT ||
                attr & IO_REPARSE_TAG_SYMLINK))
    {
      mountPointLive = 1;
    }
    else
    {
      mountPointLive = 0;
    }
    
    if (process)
    {
      if (WaitForSingleObject(process, 0) != WAIT_TIMEOUT)
      {
        DBG_MSG("ReparsePointMonitor : parent process signaled.");
        
        parentLive = 0;
      }
    }
  }

  //
  // Reparse point cleared, send quit signal.
  //
  
  DBG_MSG("Reparse point flag cleared. Going to quit.");

  StopProcessSafely();
    
  //
  // Wait until dokan worker flushes or timeout on deadlock.
  //

  DBG_MSG("Waiting until dokan finished work...");

  if (WaitForSingleObject(DokanWorkerThread, 10000) == WAIT_TIMEOUT)
  {            
    DBG_MSG("Timeout reached. Going to kill dokan worker.");
        
    TerminateThread(DokanWorkerThread, WAIT_TIMEOUT);
  }

  DBG_LEAVE("ReparsePointMonitor");
  
  return 0;
}

//
// Startup thread, which monitor reparse point flag
// on given path.
//
// RETURNS: 0 if OK.
//

Int StartReparsePointMonitor(const wchar_t *mountPoint)
{
  DBG_ENTER("Creating reparse point monitor...");

  Int exitCode = -1;

  LPTHREAD_START_ROUTINE handler = (LPTHREAD_START_ROUTINE) ReparsePointMonitor;
  
  ReparsePointMonitorThread = CreateThread(NULL, 0, handler,
                                               (void *) mountPoint, 0, NULL);
  
  FAIL(ReparsePointMonitorThread == NULL);

  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Error("ERROR: Cannot to start up reparse point monitor.\n"
              "Error code is : %d.\n", GetLastError());
  }
  
  return exitCode;
}

//
// Wait until given path mounted. We use it from another thread
// to signal succesfull mount on stdout.
//
// localRoot - path to monitor (IN).
//

DWORD WINAPI WaitForMount(const wchar_t *mountPoint)
{
  DBG_ENTER("WaitForMount");
  
  int driveMounted = 0;
  
  int triesLeft = 300;
  
  //
  // Wait until drive become availible.
  //
  
  //
  // Mount point is drive letter (X:)
  //
  
  if (wcslen(mountPoint) == 2 && mountPoint[1] == L':')
  {
    while(Not(driveMounted) && triesLeft > 0)
    {
      driveMounted = GetVolumeInformationW(mountPoint, NULL, 0, NULL, 
                                               NULL, NULL, NULL, 0);
 
      Sleep(100);

      triesLeft --;
    }
  }

  //
  // Mount point is ntfs directory.
  //
  
  else
  {
    while(Not(driveMounted) && triesLeft > 0)
    {
      if (PathIsReparsePoint(mountPoint))
      {
        driveMounted = 1;
      }
      else
      {
        triesLeft --;
        
        Sleep(100);
      }
    }
  }

  //
  // Timeout.
  //
  
  if (triesLeft == 0)
  {
    Error("ERROR: Disk is not accessible.\n");
    
    UnmountDrive(mountPoint);
    
    return 1;
  }
  else
  {
    char msg[] = "Mounted.\n";
    
    if (Options.slaveMode_)
    {
      DbgError("Mounted.\n");
    }
    else
    {
      Io::write(0, msg, sizeof(msg));
    }  

    DBG_MSG("Reparse point monted.");    
  }
  
  DBG_LEAVE("WaitForMount");
  
  return 0;
}

//
// Dokan worker thread.
//

DWORD WINAPI DokanWorker(const wchar_t *mountPoint)
{
  DBG_ENTER("DokanWorker");
  
  Int exitCode    = -1;
  Int dokanStatus = DOKAN_ERROR;
  
  Int mountPointLength = 0;
  
  char unmountedMsg[] = "Unmounted.\n";
  
  DOKAN_OPERATIONS dokanOperations = {0};
  DOKAN_OPTIONS dokanOptions       = {0};

  //
  // Set Dokan options.
  //
  
  dokanOptions.Version     = DOKAN_VERSION;
  dokanOptions.ThreadCount = 1;
  dokanOptions.Options     = 0;
  dokanOptions.MountPoint  = mountPoint;
  
  if (DBG_GET_LEVEL() > 0)
  {
    dokanOptions.Options |= DOKAN_OPTION_DEBUG;
  }

  //dokanOptions.Options |= DOKAN_OPTION_NETWORK; DOKAN_OPTION_STDERR | DKAN_OPTION_DEBUG |
  //dokanOptions.Options |= DOKAN_OPTION_REMOVABLE DOKAN_OPTION_STDERR;

  //
  // Set I/O Callbacks.
  //
 
  dokanOperations.CreateFile           = NxCreateFile;
  dokanOperations.OpenDirectory        = NxOpenDirectory;
  dokanOperations.CreateDirectory      = NxCreateDirectory;
  dokanOperations.Cleanup              = NxCleanup;
  dokanOperations.CloseFile            = NxCloseFile;
  dokanOperations.ReadFile             = NxReadFile;
  dokanOperations.WriteFile            = NxWriteFile;
  dokanOperations.FlushFileBuffers     = NxFlushFileBuffers;
  dokanOperations.GetFileInformation   = NxGetFileInformation;
  dokanOperations.FindFiles            = NxFindFiles;
  dokanOperations.FindFilesWithPattern = NULL;
  dokanOperations.SetFileAttributes    = NxSetFileAttributes;
  dokanOperations.SetFileTime          = NxSetFileTime;
  dokanOperations.DeleteFile           = NxDeleteFile;
  dokanOperations.DeleteDirectory      = NxDeleteDirectory;
  dokanOperations.MoveFile             = NxMoveFile;
  dokanOperations.SetEndOfFile         = NxSetEndOfFile;
  dokanOperations.SetAllocationSize    = NxSetAllocationSize;
  dokanOperations.LockFile             = NxLockFile;
  dokanOperations.UnlockFile           = NxUnlockFile;
  dokanOperations.GetFileSecurity      = NxGetFileSecurity;
  dokanOperations.SetFileSecurity      = NxSetFileSecurity;
  dokanOperations.GetDiskFreeSpace     = NxGetDiskFreeSpace;
  dokanOperations.GetVolumeInformation = NxGetVolumeInformation;
  dokanOperations.Unmount              = NxUnmount;

  //
  // Check does proposed mount point availible.
  //
  
  mountPointLength = wcslen(mountPoint);
  
  DBG_MSG("DEBUG: TRYING TO GATHER MORE DATA.\n");

  if (mountPointLength == 1 ||
          mountPointLength == 2 && mountPoint[0] == L':')
  {           
    dokanStatus = DOKAN_DRIVE_LETTER_ERROR;
  
    if(GetVolumeInformationW(mountPoint, NULL, 0, NULL, NULL, NULL, NULL, 0)== 0)
    {
      DBG_MSG("DEBUG:GetVolumeInformation failed.\n");

      goto fail;
    }

  }  
 
  DBG_MSG("DEBUG: Passed the GetVolumeInfomation test.\n");

  //
  // Start Dokan loop.
  //
  
  dokanStatus = __DokanMain(&dokanOptions, &dokanOperations);
  
if(dokanStatus != DOKAN_SUCCESS)
{
  Error("DEBUG: DokanMain failed.\n");

  goto fail;
}
  //
  // Dokan loop finished with success.
  //
  
  if (Options.slaveMode_)
  {
    DbgError("Unmounted.\n");
  }
  else
  {
    Io::write(0, unmountedMsg, sizeof(unmountedMsg));
  }  
 
  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    exitCode = GetLastError();
    
    Error("ERROR: Cannot set up I/O callbacks.\n"
              "Winapi code is: %u.\nLibnxfs code is: %s\n",
                  exitCode, GetDokanError(dokanStatus));
  }
  
  //
  // Remove mount point directory.
  //
  
  if (RemoveDirectoryW(mountPoint) == FALSE)
  {
    DBG_MSG("WARNING: Cannot to remove mount point directory.\n"
                "Error code is : %d.", (Int) GetLastError());
  }

  DBG_LEAVE("DokanWorker");
  
  return exitCode;
}

//
// Load dokan library.
//
// RETURNS: 0 if OK.
//

Int LoadDokanLibrary()
{
  DBG_ENTER("LoadDokanLibrary");
  
  Int exitCode = -1;
  
  #ifdef _WIN64

  DokanDll = LoadLibrary("libnxfs64.dll");

  #else

  DokanDll = LoadLibrary("libnxfs.dll");

  #endif
  
  FAIL(DokanDll == NULL);
  
  __DokanMain = (DokanMainPtr) GetProcAddress(DokanDll, "DokanMain");
  
  FAIL(__DokanMain == NULL);
  
  __DokanRemoveMountPoint = (DokanRemoveMountPointPtr) GetProcAddress(DokanDll, "DokanRemoveMountPoint");
  
  FAIL(__DokanRemoveMountPoint == NULL);
  
  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Error("ERROR: Cannot load libnxfs library. Error code is: %u.\n"
              "Please ensure that this library is properly"
                  " installed on Your system.\n", GetLastError());
  }
  
  DBG_LEAVE("LoadDokanLibrary");
  
  return exitCode;
}

//
// Free dokan library. 
//

void FreeDokanLibrary()
{
  FreeLibrary(DokanDll);
}

//
// Initialize dokan library and set up I/O callbacks by functions
// from CallBacks.cpp. 
//
// mountPoint - local path, where to mount, eg. "X:" (IN).
//

Int SetupDokanCallbacks(const wchar_t *mountPoint)
{
  DBG_ENTER("SetupDokanCallbacks");
  
  Int exitCode = -1;
  
  //
  // Start dokan worker thread.
  //

  DBG_MSG("Creating dokan thread...");

  DokanWorkerThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) DokanWorker,
                                       (void *) mountPoint, 0, NULL);

  FAIL(DokanWorkerThread == NULL);
  
  //
  // Wait until virtual disk will be accesible or timeout.
  //
  
  FAIL(WaitForMount(mountPoint));

  FAIL(StartReparsePointMonitor(mountPoint));

  //
  // Wait until dokan worker finished.
  //

  if (WaitForSingleObject(DokanWorkerThread, INFINITE) == WAIT_OBJECT_0)
  {
    GetExitCodeThread(DokanWorkerThread, (DWORD *) &exitCode);

    DBG_MSG("Dokan worker finished with code [%d].", exitCode);
    
    SetLastError(exitCode);
  }
  
  fail:

  DBG_LEAVE("SetupDokanCallbacks");
  
  return exitCode;
}

//
//
//

DWORD WINAPI DokanUnmounter(const wchar_t *mountPoint)
{
  DBG_ENTER("DokanUnmounterThread");
  
  Int exitCode = -1;
  
  FAIL(__DokanRemoveMountPoint == NULL);
  
  //
  // Worker process (it means nxfs used to mount, which
  // includes main dokan loop).
  //
  
  if (DokanWorkerThread != NULL && PathIsReparsePoint(mountPoint))
  {
    //
    // Send unmount request to dokan driver.
    //
  
    DBG_MSG("Going to umnount [%ls]...\n", mountPoint);

    FAIL(__DokanRemoveMountPoint(mountPoint) == FALSE);
  
    //
    // If it's mounter process wait until dokan flush.
    // If dokan worker doesn't response too long terminate it
    // with ERROR_TIMEOUT code.
    //

    DBG_MSG("Waiting until dokan finished work...");

    if (WaitForSingleObject(DokanWorkerThread, 10000) == WAIT_TIMEOUT)
    {            
      DBG_MSG("Timeout reached. Going to kill dokan worker.");
        
      TerminateThread(DokanWorkerThread, WAIT_TIMEOUT);
    }
  }

  //
  // Unmounter process (nxfs -uz) - wait until reparse 
  // flag cleared.
  //
  
  else
  {
    while(PathIsReparsePoint(mountPoint))
    {
      DBG_MSG("Sending unmount request...");
      
      if (__DokanRemoveMountPoint(mountPoint) == TRUE)
      {
        break;
      }
      else
      {      
        if (GetLastError() == NXFSD_STATUS_NOT_NX)
        {
          Error("ERROR: Not NX mount point or database corrupted.\n");
        
          goto fail;
        }
        else if (GetLastError() == NXFSD_STATUS_SERVICE_NOT_FOUND)
        {
          Error("ERROR: NoMachine File-system Server not found.\n");
        
          goto fail;
        }
      }
      
      //
      // Don't flood mounter service.
      //
      
      Sleep(1000);
    }
  } 

  exitCode = 0;

  //
  // Clean up.
  //
  
  fail:
  

  DBG_LEAVE("DokanUnmounterThread");
  
  return exitCode;
}

//
// Unmount given mount point.
//
// mountPoint - mount point to remove (IN).
//
// RETURNS: 0 if OK.
//

Int UnmountDrive(const wchar_t *mountPoint)
{
  DBG_ENTER("UnmountDrive");
  
  Int exitCode = -1;
  
  DBG_MSG("Creating unmounter thread...");

  LPTHREAD_START_ROUTINE handler = (LPTHREAD_START_ROUTINE) DokanUnmounter;
  
  DokanUnmounterThread = CreateThread(NULL, 0, handler,
                                          (void *) mountPoint, 0, NULL);

  DBG_MSG("Waiting until unmounter finished work...");

  if (WaitForSingleObject(DokanUnmounterThread, 5000) == WAIT_OBJECT_0)
  {
    GetExitCodeThread(DokanUnmounterThread, (DWORD *) &exitCode);

    DBG_MSG("Dokan unmounter finished with code [%d].", exitCode);
    
    SetLastError(exitCode);
  }

  exitCode = 0;
  
  DBG_LEAVE("UnmountDrive");
  
  return exitCode;
}
