/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXSSHFS, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "IOFifo.h"
#include "SftpClient.h"

#undef  DUMP

//
// Defines.
//

#define IOFIFO_INIT_TIMEOUT 10000

//
// Global variables.
//

Int IdCount = 0;

queue<IOFifoEntry *> IOFifo;

HANDLE IOFifoMutex = NULL;

extern SftpClient Sftp;

/* 
  FIXME: Don't use different mutexes for FIFO and Id count.
*/

HANDLE IdCountMutex = NULL;
HANDLE IOFifoThread = NULL;

HANDLE InputHandle  = GetStdHandle(STD_INPUT_HANDLE);
HANDLE OutputHandle = GetStdHandle(STD_OUTPUT_HANDLE);
HANDLE ErrorHandle  = GetStdHandle(STD_ERROR_HANDLE);
HANDLE ReconnHandle = NULL;

Int IOFifoError = 0;

//
// Atomic write i.e. all data written or nothing.
//
// handle      - output handle to write to (IN).
// outBuf      - buffer with data to write (IN).
// bytesToRead - size of outBuf in bytes (IN).
// timeout     - timeout in miliseconds (IN).
//
// RETURNS: 0 if all bytes written.
//

Int AtomicWrite(HANDLE handle, const void *outBuf, Int bytesToWrite, Int timeout)
{
  Int exitCode = -1;
 
  Int written = 0;

  Char *buf = (Char *) outBuf;
  
  fd_set writeSet;
  
  struct timeval tv = {timeout / 1000, (timeout % 1000) * 1000};

  if (handle == NULL)
  {
    handle = OutputHandle;
  }
  
  //
  // Write until all data sent or timeout.
  //
  
  while(bytesToWrite > 0)
  {
    #ifdef EMULATE_SLOW_NETWORK
    Sleep(100);
    #endif
    
    FD_ZERO(&writeSet);
  
    FD_SET((SOCKET) handle, &writeSet);
    
    #ifdef DUMP
    DBG_MSG("AtomicWrite : select...");
    #endif
    
    FAIL(select(1, NULL, &writeSet, NULL, &tv) <= 0);

    if (FD_ISSET((SOCKET) handle, &writeSet))
    {
      #ifdef DUMP
      DBG_MSG("AtomicWrite : Sending [%d] bytes...", bytesToWrite);
      #endif

      written = send((SOCKET) handle, buf, bytesToWrite, 0);
    
      #ifdef DUMP
      DBG_MSG("AtomicWrite : Send() result : [%d].", written);
      #endif
      
      FAIL(written <= 0);
    
      #ifdef DUMP
      DBG_MSG("Written [%d] bytes to [%p].", written, handle);
      #endif

      bytesToWrite -= written;
    
      buf += written;
    }  
  }
  
  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Error("ERROR: Cannot write data to [%p].\n"
             "Error code is : %d.\nWinsock error is : %d.\n",
                 handle, (Int) GetLastError(), (Int) WSAGetLastError());
  }
  
  return exitCode;
}

//
// Atomic read i.e. all data readed or nothing.
//
// handle      - input handle to read from (IN).
// inpBuf      - buffer where to store readed data (IN).
// bytesToRead - number of bytes to read (IN).
// timeout     - timeout in miliseconds (IN).
//
// RETURNS: 0 if all bytes readed.
//

Int AtomicRead(HANDLE handle, void *inpBuf, Int bytesToRead, Int timeout)
{
  Int exitCode = -1;
  
/*
  static Int count = 1;
*/

  fd_set readSet;
    
  Int readed = 0;
  
  Char *buf = (Char *) inpBuf;
  
  struct timeval tv = {timeout / 1000, (timeout % 1000) * 1000};

/*
  count ++;
  
  FAIL(count % 32 == 0);
*/  
  
  if (handle == NULL)
  {
    handle = InputHandle;
  }

  while(bytesToRead > 0)
  {
    #ifdef EMULATE_SLOW_NETWORK
    Sleep(100);
    #endif
    
    FD_ZERO(&readSet);
  
    FD_SET((SOCKET) handle, &readSet);

    #ifdef DUMP
    DBG_MSG("AtomicRead : select...");
    #endif
    
    FAIL(select(1, &readSet, NULL, NULL, &tv) <= 0);
    
    if (FD_ISSET((SOCKET) handle, &readSet))
    {
      #ifdef DUMP
      DBG_MSG("AtomicRead : Receiving [%d] bytes...", bytesToRead);
      #endif 
      
      readed = recv((SOCKET) handle, buf, bytesToRead, 0);
      
      #ifdef DUMP
      DBG_MSG("AtomicRead : Recv() status [%d] bytes...", readed);
      #endif

      FAIL(readed <= 0);

      #ifdef DUMP
      DBG_MSG("Readed [%d] bytes from [%p].", readed, handle);
      #endif

      bytesToRead -= readed;
    
      buf += readed;
    }  
  }

  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Error("ERROR: Cannot read data from [%p].\n"
             "Error code is : %d.\nWinsock error is : %d.\n",
                 handle, (Int) GetLastError(), (Int) WSAGetLastError());
  }
  
  return exitCode;
}

//
// Thread processing incoming IO requests from FIFO.
// To add new request to FIFO caller shoulds use IORequest() 
// function.
//
// initEvent - event to signal creator, that thread ready to work (IN/OUT/OPT).
//

Int IOFifoThreadHandler(HANDLE initEvent)
{
  DBG_ENTER("IOFifoThreadHandler");

  Int exitCode = -1;
  
  Int packetSize = 0;
  
  Int goOn  = 1;
  Int id    = -1;
  Int state = 0;

  IOFifoEntry *iorq = NULL;

  //
  // Tell creator, that thread ready to use from now.
  //
  
  DBG_MSG("IOFifoThreadHandler : ready.");

  if (initEvent != NULL)
  {
    SetEvent(initEvent);
  }
  
  //
  // Fall into main loop.
  //
  
  DBG_MSG("IOFifoThreadHandler : falling to main loop...");

  while(goOn)
  {
    //
    // Go sleep until nothing to do.
    //

    if (IOFifo.empty())
    {
      DBG_MSG("IOFifoThreadHandler : going to sleep...");
      
      SuspendThread(IOFifoThread);

      DBG_MSG("IOFifoThreadHandler : waked up.");
    }
  
    //
    // Process all packets in queue. 
    //
    
    while(Not(IOFifo.empty()))
    {
      //
      // Get next IO request from queue.
      //
      
      IOFifoLock();

      iorq = IOFifo.front();
      
      IOFifo.pop();
      
      IOFifoUnlock();

      id = iorq -> id_;
      
      iorq -> state_ = PENDING_STATE;
      
      DBG_MSG("Popped request no. [%d]...", id);

      //
      // Try process request as long as not finished.
      //
      
      state = PENDING_STATE;
      
      while(state == PENDING_STATE)
      {
        //
        // Write whole packet to server.
        //
   
        packetSize = iorq -> inpData_ -> size_;

        DBG_MSG("Sending [%d] bytes for request no. [%d]...", packetSize, id);
     
        DBG_BUFFER_DUMP(iorq -> inpData_, min(packetSize, MAX_PACKET));
      
        FAILEX(packetSize > MAX_PACKET, "ERROR: Packet too big.\n");

        //
        // Convert packet size to big endian before send.
        // Note that, sftp server compute size excluding
        // first size field.
        //

        iorq -> inpData_ -> size_ = InvertBytes(iorq -> inpData_ -> size_ - 4);

        //
        // Send message to server.
        //

        if (AtomicWrite(OutputHandle, iorq -> inpData_, packetSize))
        {
          break;
        }  
      
        //
        // Sent. Convert packet size back to little endian.
        //
  
        iorq -> inpData_ -> size_ = packetSize;

        //
        // Wait for response from server.
        //

        DBG_MSG("Reading response for request no. [%d]...", id);
      
        //
        // Read packet size.
        //
      
        if (AtomicRead(InputHandle, &packetSize, 4))
        {
          break;
          
        }

        packetSize = InvertBytes(packetSize);      
      
        //
        // If output buffer is allocable allocate needed space.
        //
      
        if (iorq -> dynamic_)
        { 
          //
          // Output buffer doesn't exist yet. Allocate new one.
          //
        
          if (iorq -> outSize_ == 0 || *(iorq -> poutData_) == NULL)
          {
            DBG_MSG("Allocating [%d] bytes for request [%d]...", packetSize + 4, id);
                         
            *(iorq -> poutData_) = (SftpMsg *) malloc(packetSize + 4);                        
          }
          
          //
          // Output buffer exists but it's too small. Reallocate it.
          //
        
          else
          {
            DBG_MSG("Expanding buffer to [%d] bytes for request [%d]...", packetSize + 4, id);
              
            *(iorq -> poutData_) = (SftpMsg *) realloc(*(iorq -> poutData_), packetSize + 4);
          }
        
          FAILEX(iorq -> outData_ == NULL,
                     "ERROR: Cannot allocate output buffer.\n");
                   
          iorq -> outSize_ = packetSize + 4;

          //
          // Read response content.
          //

          (*iorq -> poutData_) -> size_ = packetSize + 4;
      
          if (AtomicRead(InputHandle, &((*iorq -> poutData_) -> type_), packetSize))
          {
            Error("ERROR: Cannot read data for request no. [%d].\n", id);
            
            break;
          }  
        }
      
        //
        // Otherwise, check is given output buffer not too small.
        //
      
        else
        {
          FAILEX(packetSize + 4 > iorq -> outSize_, 
                     "ERROR: Output buffer too small.\n"
                         "Data space needed : [%d] bytes.\n", packetSize + 4);

          iorq -> outData_ -> size_ = packetSize + 4;
      
          if (AtomicRead(InputHandle, &(iorq -> outData_ -> type_), packetSize))
          {
            Error("ERROR: Cannot read data for request no. [%d].\n", id);
            
            break;
          }
        }
      
        //
        // If we reached here, whole request is processed succesfully.
        //

        state = SUCCESS_STATE;
      }
     
      if (state == PENDING_STATE)
      {
        state = ERROR_STATE;
      }
      
      //
      // Reconnect if request failed.
      //
      
      if (state == ERROR_STATE)
      {
        FAIL(IOFifoReconnect());
      }

      //
      // Set request as finished with ERROR or SUCCESS state.
      //

      DBG_MSG("Request no. [%d] finished with code [%x].", iorq -> id_, state);
      
      iorq -> state_ = state;
      
      SetEvent(iorq -> readyEvent_);      
    }
  }
  
  //
  // Error handler. 
  // If we got here we know fatal error occures and we
  // go to unconditional clean up exit.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    //
    // Set FIFO error state.
    //
    
    IOFifoError = 1;
    

    //
    // Clear whole FIFO.
    //
    
    IOFifoLock();

    while(Not(IOFifo.empty()))
    {
      IOFifo.pop();
    }

    IOFifoUnlock();

    //
    // Signal fatal error.
    //
    
    Fatal("ERROR: Cannot process IO request no. [%d].\n"
              "Error code is : %d.\n", id, (Int) GetLastError());
  }
    
  DBG_LEAVE("IOFifoThreadHandler");
  
  return exitCode;
}

//
// Send IORequest to server and wait until response received.
//
// iorq    - Struct describing request (IN/OUT).
// timeout - Timeout limit in ms (IN).
//
// RETURNS: 0 if OK.
//

Int IORequest(IOFifoEntry *iorq, Int timeout)
{
  DBG_ENTER("IORequest");

  Int exitCode = -1;

  Int readyEventAllocated = 0;
  
  //
  // Don't push new requests if FIFO is in error state.
  // Process is going to unmount and exit in this time.
  //
  
  FAILEX(IOFifoError == 1, "ERROR: I/O request [%d] rejected.\n", iorq -> id_);
  
  //
  // If caller doesn't specify own ready event
  // create new one.
  //
  
  if (iorq -> readyEvent_ == NULL)
  {
    DBG_MSG("IORequest : Creating ready event...");
    
    iorq -> readyEvent_ = CreateEvent(NULL, FALSE, FALSE, NULL);
    
    FAILEX(iorq -> readyEvent_ == NULL, 
               "ERROR: Cannot create ready event.\n");
    
    readyEventAllocated = 1;
  }
  
  //
  // Set wait state. It means request waits
  // until worker thread pop it.
  //
  
  iorq -> state_ = WAIT_STATE;

  //
  // Push IO request onto FIFO queue.
  //
  
  DBG_MSG("Pushing request [%d] onto I/O FIFO...", iorq -> id_);
  
  IOFifoLock();

  IOFifo.push(iorq);
  
  IOFifoUnlock();
  
  //
  // Resume worker thread.
  //
  
  DBG_MSG("Waking up I/O FIFO thread...");
    
  while(ResumeThread(IOFifoThread) == 0 && IOFifoError == 0)
  {
    Sleep(1000);
  }

  //
  // Wait until IO request processed or timeout.
  //

  DBG_MSG("Waiting until request [%d] finished...", iorq -> id_);
  
/*
 * FIXME: Timeout moved to woker thread to force
 *        wait until worker thread doesn't stay in reconnect
 *        loop with iorq struct longer.
 *        We need wait, because caller may want free iorq
 *        just after exit from IORequest().
 */

  WaitForSingleObject(iorq -> readyEvent_, INFINITE);
  
  FAILEX(iorq -> state_ != SUCCESS_STATE,
             "Request [%d] finished with error.\n", iorq -> id_);
  
  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Error("ERROR: Cannot send IO request.\n"
              "Error code is : %d.\n", (Int) GetLastError());
  }

  if (readyEventAllocated)
  {
    CloseHandle(iorq -> readyEvent_);
    
    iorq -> readyEvent_ = NULL;
  }
  
  if (iorq -> inpFree_ && iorq -> inpData_)
  {
    free(iorq -> inpData_);
    
    iorq -> inpData_ = NULL;
  }

  DBG_LEAVE("IORequest");
  
  return exitCode;
}

//
// Initialize IO FIFO thread and lock mutex.
// We use FIFO to process IO requests from many threads
// via one network socket.
//
// REURNS: 0 if OK.
//

Int IOFifoInit(HANDLE stdinHandle, HANDLE stdoutHandle, 
                   HANDLE stderrHandle, HANDLE reconnHandle)
{
  Int exitCode = -1;

  WSADATA wsaData;
    
  //
  // Set std handles.
  //
  
  InputHandle  = stdinHandle;
  OutputHandle = stdoutHandle;
  ErrorHandle  = stderrHandle;
  ReconnHandle = reconnHandle;

  //
  // Create event to wait until IO thread became ready to use.
  //
  
  HANDLE threadReady = CreateEvent(NULL, FALSE, FALSE, NULL);
  
  FAILEX(threadReady == NULL, "ERROR: Cannot create ready event.\n");

  //
  // Create IO FIFO thread thread.
  //
  
  if (IOFifoThread == NULL)
  {
    DBG_MSG("Creating IO FIFO thread...");
 
    IOFifoThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) IOFifoThreadHandler,
                                    threadReady, 0, NULL);

    FAILEX(IOFifoThread == NULL, "ERROR: Cannot create IO FIFO thread.\n");
  }
  
  //
  // Init mutex to lock FIFO.
  //
  
  if (IOFifoMutex == NULL)
  {
    DBG_MSG("Creating I/O FIFO mutex...");
  
    IOFifoMutex = CreateMutex(NULL, FALSE, NULL);
  
    FAILEX(IOFifoMutex == NULL, "ERROR: Cannot create IO FIFO mutex.\n");
  }

  //
  // Init mutex to lock IdCount to make sure id
  // uniqually in multithread.
  //
  
  if (IdCountMutex == NULL)
  {
    DBG_MSG("Creating ID count mutex...");
  
    IdCountMutex = CreateMutex(NULL, FALSE, NULL);
  
    FAILEX(IdCountMutex == NULL, "ERROR: Cannot create ID count mutex.\n");
  }
  
  //
  // Wait until thread ready to avoid sending request until it 
  // start processing incoming data.
  //
  
  FAIL(WaitForSingleObject(threadReady, IOFIFO_INIT_TIMEOUT) != WAIT_OBJECT_0);
  
  //
  // Init socket API.
  //
  
  FAILEX(WSAStartup(MAKEWORD(2, 2), &wsaData), 
             "ERROR: Cannot initialize WinSock DLL.");

  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Error("ERROR: Cannot init IO FIFO.\n"
              "Error code is : %d.\n", (Int) GetLastError());
  }
  
  CloseHandle(threadReady);

  return exitCode;
}

void IOFifoCleanup()
{
  TerminateThread(IOFifoThread, 0);
  
  CloseHandle(IOFifoMutex);
  CloseHandle(IdCountMutex);
  CloseHandle(InputHandle);
  CloseHandle(OutputHandle);
}

//
// Generate unique 32-bit id. We use it for sftp packet's
// id in multithread.
//

Int IOFifoGenerateUniqueId()
{
  Int id;
  
  LOCK_OBJECT(IdCount);

  IdCount ++;

  id = IdCount;
  
  UNLOCK_OBJECT(IdCount);
  
  return id;
}

//
// Do reconnect procedure:
//
// - send "Reconnect." message to parrent.
// - parent create new connection to sftp server.
// - wait for new input/output socket pairs from parent.
//
// RETURNS: 0 if OK.
//

Int IOFifoReconnect()
{
  DBG_ENTER("IOFifoReconnect");
  
  Char msg[1024] = "Reconnect.\n";
  
  Int exitCode = -1;

  struct timeval tv = {ATOMIC_TIMEOUT / 1000, 0};
  
  fd_set rset;
  
  //
  // Close old sockets.
  //
  
  DBG_MSG("IOFifoReconnect : Going to shutdown [%d]...\n", InputHandle);
  DBG_MSG("IOFifoReconnect : Going to shutdown [%d]...\n", OutputHandle);
  
  if (InputHandle == OutputHandle)
  {
    DBG_MSG("IOFifoReconnect : Closing input/output handlers : [%d] ", InputHandle);

    CloseHandle(InputHandle);
  }
  else
  {
    DBG_MSG("IOFifoReconnect : Closing input handle : [%d] ", InputHandle);
    DBG_MSG("IOFifoReconnect : Closing output handle : [%d] ", OutputHandle);

    CloseHandle(InputHandle);
    CloseHandle(OutputHandle);
  }

  InputHandle  = NULL;
  OutputHandle = NULL;
  
  FAILEX(ReconnHandle == NULL, "ERROR: Reconnect handle not set.\n");
  
  //
  // Send reconnect request to parent process.
  //

  DBG_MSG("IOFifoReconnect : Sending reconenct request to parent...\n");
  
  FAIL(AtomicWrite(ReconnHandle, msg, strlen(msg)));
  
  //
  // Wait until parent process prepare new connection with
  // sftp server and receive new stdin/stdout handles to it.
  //
  
  FD_ZERO(&rset);
  
  FD_SET((SOCKET) ReconnHandle, &rset);

  DBG_MSG("IOFifoReconnect : Waiting for handles...\n");
    
  FAIL(select(1, &rset, NULL, NULL, &tv) <= 0);
  
  FAIL(Not(FD_ISSET((SOCKET) ReconnHandle, &rset)));
  
  FAIL(recv((SOCKET) ReconnHandle, msg, sizeof(msg), 0) <= 0);
  
  sscanf(msg, "%d %d", (Int *) &InputHandle, (Int *) &OutputHandle);
  
  DBG_MSG("IOFifoReconnect: Readed [%s].\n", msg);
  DBG_MSG("IOFifoReconnect: Reset input handle to  : [%p]", InputHandle);
  DBG_MSG("IOFifoReconnect: Reset output handle to : [%p]", OutputHandle);

  //
  // Reinit sftp protocol (hello packet) and
  // reinit opened handles state.
  //
  
  Sftp.reconnect();
  
  //
  // Error handler.
  //

  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Error("ERROR: Cannot do reconnect procedure.\n"
              "Error code is : %d.\n", (Int) GetLastError());
  }
  
  DBG_LEAVE("IOFifoReconnect");
  
  return exitCode;
}
