/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXSSHFS, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#include <stdlib.h>

#include "Debug.h"
#include "Types.h"
#include "SftpClient.h"
#include "Callbacks.h"
#include "Options.h"

//
// Global variables.
//

OptionsStruct Options;

SftpClient Sftp;

HANDLE CtrlBreakMutex = NULL;

//
// Ctrl-break handler.
//

BOOL WINAPI CtrlHandlerRoutine(DWORD dwCtrlType)
{
  DBG_MSG("CTRL-BREAK received. Going to exit...");

  WaitForSingleObject(CtrlBreakMutex, 0);

  CacheCleanup();

  UtilsCleanup();

  IOFifoCleanup();

  UnmountDrive(Options.localRoot_);

  ReleaseMutex(CtrlBreakMutex);

  return TRUE;
}

//
// Entry point.
//

Int main(Int argc, Char **argv)
{
  Int exitCode = -1;

/*
  FIXME: Causes 0xc000005 on gcc4.5 toolchan.

  srand(time(0));
*/

  //
  // Clear SSH_ASKPASS variable.
  //

  ProcessSetEnvironment("SSH_ASKPASS", "");

  //
  // Set CTRL-BREAK handler.
  //

  SetConsoleCtrlHandler(CtrlHandlerRoutine, TRUE);

  CtrlBreakMutex = CreateMutex(NULL, FALSE, NULL);

  FAIL(CtrlBreakMutex == NULL);

  //
  // Parse command line arguments.
  //

  FAIL(Options.parseArguments(argc, argv));

  FAIL(UtilsInit());

  //
  // Init IO Requests FIFO.
  //

  FAIL(IOFifoInit(Options.stdinHandle_,
                      Options.stdoutHandle_,
                          Options.stderrHandle_,
                              Options.reconnHandle_));

  //
  // Parse options.
  //

  switch(Options.mode_)
  {
    //
    // Mount new drive.
    //

    case MOUNT_MODE:
    {
      Char remoteRoot[MAX_PATH] = {0};

      DBG_MSG("Mode             : [Mount]");
      DBG_MSG("User             : [%s]", Options.user_);
      DBG_MSG("Host             : [%s]", Options.host_);
      DBG_MSG("Port             : [%d]", Options.port_);
      DBG_MSG("Remote path      : [%ls]", Options.remoteRoot_);
      DBG_MSG("Local path       : [%ls]", Options.localRoot_);
      DBG_MSG("Slave mode       : [%d]", Options.slaveMode_);
      DBG_MSG("Timeout          : [%d] ms", Options.timeout_);
      DBG_MSG("Input handle     : [%p]", Options.stdinHandle_);
      DBG_MSG("Output handle    : [%p]", Options.stdoutHandle_);
      DBG_MSG("Error handle     : [%p]", Options.stderrHandle_);
      DBG_MSG("Reconnect handle : [%p]", Options.reconnHandle_);
      DBG_MSG("Parent pid       : [%d]", Options.parentPid_);

      //
      // Init connection to sftp-server.
      //

      DBG_MSG("Connecting to sftp-server...");

      Sftp.setupConnection(Options.user_, Options.host_, Options.port_,
                               Options.password_, Options.keyfile_,
                                   Options.sshArgsCount_, Options.sshArgs_,
                                       Options.timeout_, Options.slaveMode_);

      FAIL(Sftp.connect());

      //
      // Stat remote root ensure is it accesible.
      //

      DBG_MSG("Stating root remote path...");

      GetFullRemotePath(remoteRoot, L"/", MAX_PATH);

      if (Sftp.stat(remoteRoot))
      {
        Error("ERROR: Cannot stat remote root.\n");

        SetLastError(ERROR_BAD_NET_NAME);

        goto fail;
      }

      //
      // Letter mount point:
      // - must NOT exist.
      //

      if (Options.localRoot_[0] == 0 || (Options.localRoot_[1] == L':' &&
              Options.localRoot_[2] == 0))
      {
        DBG_MSG("Letter mount point detected.");

        if (GetVolumeInformationW(Options.localRoot_, NULL,
                                      0, NULL, NULL, NULL, NULL, 0))
        {
          Error("ERROR: Mount point already assigned.\n");

          SetLastError(ERROR_ALREADY_ASSIGNED);

          goto fail;
        }
      }

      //
      // Ntfs reparse point.
      // - cannot be used by another process.
      // - must be empty dir or not exist.
      //

      else
      {
        DBG_MSG("Ntfs reparse point detected.");

        //
        // Create empty mount point if it's not exist yet.
        //

        CreateDirectoryW(Options.localRoot_, NULL);

        //
        // Check is local mouth point not used by another process.
        // If yes, try wait until it will become ready.
        //

        DBG_MSG("Waiting until [%ls] mount point ready...", Options.localRoot_);

        FAIL(WaitUntilPathReady(Options.localRootA_));//Options.localRoot_));

        //
        // Reject if local path is already reparse moint.
        //

        DBG_MSG("Checking mount point attributes...");

        if (PathIsReparsePoint(Options.localRoot_))
        {
          Error("ERROR: Mount point already assigned.\n");

          SetLastError(ERROR_ALREADY_ASSIGNED);

          goto fail;
        }

        //
        // Reject if path is not empty.
        //

        if (Not(PathIsEmpty(Options.localRoot_)))
        {
          Error("ERROR: Mount point is not empty.\n");

          SetLastError(ERROR_DIR_NOT_EMPTY);

          goto fail;
        }
      }

      //
      // Load dokan library.
      //

      DBG_MSG("Loading dokan library...");

      FAIL(LoadDokanLibrary());

      //
      // Map I/O operations to our callbacks.
      //

      DBG_MSG("Setting up I/O callbacks...");

      FAIL(SetupDokanCallbacks(Options.localRoot_));

      break;
    }

    //
    // Unmount one of existing drives.
    //
    // nxsshfs --unmount x:
    // nxsshfs -f x:
    // nxsshfs -uz x:
    //

    case UNMOUNT_MODE:
    {
      DWORD attrib;

      DBG_MSG("Mode        = [Unmount]");
      DBG_MSG("Local path  = [%ls]", Options.localRoot_);

      //
      // Does path exist?
      //

      attrib = GetFileAttributesW(Options.localRoot_);

      if (GetLastError() == ERROR_FILE_NOT_FOUND)
      {
        Error("ERROR: Path not exist.\n");

        SetLastError(ERROR_PATH_NOT_FOUND);

        goto fail;
      }

      //
      // Unmount only if path is reparse point.
      //

      if (PathIsReparsePoint(Options.localRoot_))
      {
        //
        // Load dokan library.
        //

        DBG_MSG("Loading dokan library...");

        FAIL(LoadDokanLibrary());

        //
        // Unmount drive.
        //

        DBG_MSG("Unmounting drive...");

        FAIL(UnmountDrive(Options.localRoot_));
      }
      else
      {
        Error("ERROR: Not a mount point or already unmounted.\n");

        SetLastError(ERROR_DEV_NOT_EXIST);
      }

      break;
    }

    case ERROR_MODE:
    {
      FAIL(1);

      break;
    }
  }

  DBG_MSG("Exititing with code : %d.", exitCode);

  DBG_STOP();

  //
  // Clean up.
  //

  exitCode = 0;

  fail:

  if (exitCode)
  {
    exitCode = GetLastError();
  }

  FreeDokanLibrary();

  // do not exit without waiting for ctrl break thread
  WaitForSingleObject(CtrlBreakMutex, 5000);

  CloseHandle(CtrlBreakMutex);

  return exitCode;
}