/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXSSHFS, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "SftpClient.h"

using std::map;
using std::string;

Int64 SftpResource::localHandleCount_ = 1;


SftpClient::SftpClient()

: failedConnections_(0),
  firstConnection_(1),
  fdin_(-1),
  fdout_(-1),
  fderr_(-1),
  pid_(-1),
  msgCount_(0),
  serverHandleSize_(4)
{
//  DBG_ENTER("SftpClient::SftpClient");
  
//  DBG_LEAVE("SftpClient::SftpClient");
}
  
SftpClient::~SftpClient()
{
//  DBG_ENTER("SftpClient::~SftpClient");

  disconnect();
  
//  DBG_LEAVE("SftpClient::~SftpClient");
}
  
//
// Disconnect existing connection with sftp server.
//

void SftpClient::disconnect()
{
  DBG_ENTER("SftpClient::disconnect()");
  
  //
  // Set all opened remote handle as dropped.
  //

  map<Int64, SftpResource>::iterator it;
  
  for (it = openedResources_.begin(); it != openedResources_.end(); it++)
  {
    (it -> second).dropped_ = 1;
  }
  
  //
  // Close child ssh process.
  //
  
  if (Not(slaveMode_))
  {
    if (fdin_ > 2)
    {
      DBG_MSG("Going to shutdown [%d]...", fdin_);
      
      Io::close(fdin_);
      
      DBG_MSG("Shutted [%d].", fdin_);
    }
  
    if (fderr_ > 2)
    {
      DBG_MSG("Going to shutdown [%d]...", fderr_);
      
      Io::close(fderr_);

      DBG_MSG("Shutted [%d].", fderr_);
    }  
  
    if (fdout_ > 2)
    {
      DBG_MSG("Going to shutdown [%d]...", fdout_);
      
      Io::close(fdout_);

      DBG_MSG("Shutted [%d].", fdout_);
    }
  
    if (pid_ > 0)
    {
      DBG_MSG("Terminating child ssh process...");
      
      TerminateProcess(HandleCast(pid_), 1);
    }
  }

  DBG_LEAVE("SftpClient::disconnect()");
}

Int SftpClient::reconnect()
{
  if (Not(firstConnection_))
  {
    DBG_MSG("SftpClient::reconecting...");

    disconnect();
  
    if (connect() == 0)
    {
      Error("Reconected.\n");

      return 0;
    }
  }

  //
  // Too many connection tries. Exit with error.
  //
  
  disconnect();

  StopProcessSafely();

  return -1;
}

//
// Set up connection parameter. This function MUST be called before
// first connect() call.
//

void SftpClient::setupConnection(const Char *user, const Char *host, Int port,
                                     const Char *password, const Char *keyfile,
                                         Int sshArgsCount, Char **sshArgs,    
                                             Int timeout, Int slaveMode)
{
  user_         = user;
  host_         = host;
  port_         = port;
  sshArgsCount_ = sshArgsCount;
  sshArgs_      = sshArgs;
  password_     = password;
  keyfile_      = keyfile;
  timeout_      = timeout;
  slaveMode_    = slaveMode;
}

//
// Run child ssh-client process with sftp subsystem and initialize
// new sftp connection.
//
// RETURNS: 0 if OK.
//

Int SftpClient::connect()

{
  DBG_ENTER("SftpClient::connect");

  Int exitCode = -1;

  Int retSize;
  
  //
  // Pipes for comunicating with child ssh-client.
  //
  
  Int pipeIn[2]  = {-1, -1};
  Int pipeOut[2] = {-1, -1};
  Int pipeErr[2] = {-1, -1};
  
  //
  // We'll pass stdin/stdout handles to child via these args.
  // This trick allows child to direct use parrent stdin/stdout, but
  // only for passphrase reading (and showing prompt), so we don't
  // need process it manually.  
  //
  
  //Char stdInArg[128]  = {0};
  //Char stdOutArg[128] = {0};

  Char portArg[64] = {0};
  
  //
  // Init packet with sftp version to send to sftp-server.
  //
  
  SftpInitMsg initMsg = 
  {
    sizeof(SftpInitMsg),
    
    SSH2_FXP_INIT,
  
    InvertBytes(SSH2_FILEXFER_VERSION)
  };  
  
  //
  // Buffer for version packet received from sftp-server.
  //
  
  SftpVersionMsg *versionMsg = NULL;
  
  //
  // Parameters to execute child ssh-client.
  //

  Int count = 0;
  
  Char *parameters[64] = {NULL};

  Char tmp[1024 * 4] = {0};

  Char *p = tmp;
  
  //
  // Count connection tries.
  //
  
  failedConnections_ ++;
  
  //
  // Slave mode. Don't create child ssh process.
  // Use stdin/stdout directly to communicate with sftp server.
  //
  
  if (slaveMode_)
  {
    DBG_MSG("SftpClient::connect : Slave mode detected.");
    DBG_MSG("SftpClient::connect : Using std handles...");

    fdout_ = 0;
    fdin_  = 1;
    fderr_ = 2;
  }
  
  //
  // Default mode. Create child ssh process, and use it
  // to communicate with sftp server.
  //
  
  else
  {
    //
    // Create pipe for ssh-client stdin.
    //
  
    DBG_MSG("SftpClient::connect : Creating stdin pipe for ssh-client...");
  
    FAIL(Io::pipe(pipeIn, Io::Bidirectional) < 0);

    fdin_ = Io::duplicate(pipeIn[1]);

    Io::set(fdin_, Io::NoInherit, 1);
  
    Io::set(fdin_, Io::NoDelay, 1);

    DBG_MSG("Going to shutdown [%d].", pipeIn[1]);
    
    Io::close(pipeIn[1]);
    
    DBG_MSG("Shutted [%d].", pipeIn[1]);

    Io::set(pipeIn[0], Io::NoInherit, 0);
  
    Io::set(pipeIn[0], Io::NoDelay, 1);
  
    //
    // Create pipe for ssh-client stdout.
    //

    DBG_MSG("SftpClient::connect : Creating stdout pipe for ssh-client...");
  
    FAIL(Io::pipe(pipeOut, Io::Bidirectional) < 0);

    fdout_ = Io::duplicate(pipeOut[0]);
    
    Io::noBlocking(fdout_, 1);

    Io::set(fdout_, Io::NoInherit, 1);

    Io::set(fdout_, Io::NoDelay, 1);

    DBG_MSG("Going to shutdown [%d].", pipeOut[0]);
    
    Io::close(pipeOut[0]);
    
    DBG_MSG("Shutted [%d].", pipeOut[0]);    

    Io::set(pipeOut[1], Io::NoInherit, 0);

    Io::set(pipeOut[1], Io::NoDelay, 1);

    //
    // Create pipe for ssh-client stderr.
    //

    DBG_MSG("SftpClient::connect : Creating stderr pipe for ssh-client...");

    FAIL(Io::pipe(pipeErr, Io::Bidirectional) < 0);

    fderr_ = Io::duplicate(pipeErr[0]);

    Io::noBlocking(fderr_, 1);

    Io::set(fderr_, Io::NoInherit, 1);

    Io::set(fderr_, Io::NoDelay, 1);

    DBG_MSG("Going to shutdown [%d].", pipeErr[0]);

    Io::close(pipeErr[0]);

    DBG_MSG("Shutted [%d].", pipeErr[0]);    

    Io::set(pipeErr[1], Io::NoInherit, 0);

    Io::set(pipeErr[1], Io::NoDelay, 1);

    //
    // Set up basic ssh parameters.
    //

    StringPrint(portArg, "%d", port_);

    //StringPrint(stdInArg, "-oPassInputHandle=%u", (unsigned Int) GetStdHandle(STD_INPUT_HANDLE));
    //StringPrint(stdOutArg, "-oPassOutputHandle=%u", (unsigned Int) GetStdHandle(STD_OUTPUT_HANDLE));

    parameters[count++] = (char *)".\\";
    parameters[count++] = (char *)"nxssh.exe";
    parameters[count++] = (char *)"-T";

    parameters[count++] = (char *)"-p";
    parameters[count++] = portArg;

    parameters[count++] = (char *)"-s";
    parameters[count++] = (char *)"-l";
    parameters[count++] = (Char *) user_;

    //
    // Public key specified. Try key auth.
    //

    if (keyfile_)
    {
      DBG_MSG("SftpClient::connect : Going to key auth [%s]...", keyfile_);

      parameters[count++] =  (char *)"-i";
      parameters[count++] =  (char *) keyfile_;
      parameters[count++] =  (char *)"-oPubkeyAuthentication=yes";
      parameters[count++] =  (char *)"-oPasswordAuthentication=no";
      parameters[count++] =  (char *)"-oPreferredAuthentications=publickey";
    }

    //
    // No public key specified. Try password auth.
    //

    else
    {
      DBG_MSG("SftpClient::connect : Going to pass auth...");

      parameters[count++] =  (char *)"-oPubkeyAuthentication=no";
      parameters[count++] =  (char *)"-oPasswordAuthentication=yes";
      parameters[count++] =  (char *)"-oPreferredAuthentications=password";
    }

    parameters[count++] =  (char *)"-oStrictHostKeyChecking=no";
    parameters[count++] =  (char *)"-oUserKnownHostsFile=NUL";

    //parameters[9] = "-oServerAliveInterval=5";

    //parameters[8] = stdInArg;
    //parameters[8] = stdOutArg;
    //

    //
    // Add extra optional args if passed.
    //

    for (Int i = 0; i < sshArgsCount_; i++)
    {
      parameters[count++] = sshArgs_[i];
    }

    //
    // Host and subsystem name (sftp in our case)
    // must be the last parameters.
    //

    parameters[count++] = (Char *) host_;
    parameters[count++] = (Char *) "sftp";

    //
    // Execute ssh-client process.
    //

    DBG_MSG("SftpClient::connect : Creating child ssh-client process...");

    pid_ = ProcessCreate("nxssh.exe", count, parameters,
                             pipeIn[0], pipeOut[1], 2, 0, 1);

    FAIL(pid_ < 0);

    //
    // Write password to nxssh stdin if needed.
    //

    if (password_[0])
    {
      DBG_MSG("SftpClient::writing password...");

      for (int i = 0; password_[i]; i++)
      {
        Char c = password_[i] - 1;

        Io::write(fdin_, &c, 1);
      }

      Io::write(fdin_, "\n", sizeof("\n"));
    }

    if (keyfile_ == NULL)
    {
      //
      // Skip all unneeded messages on stdout until 'password: ' prompt found.
      // FIXME: Is there better way flush NXSSH stdout?
      //

      while(strstr(tmp, "password: ") == NULL)
      {
        FAIL(atomicRead(fdout_, p++, 1));
      }

      atomicRead(fdout_, tmp, 2);
    }
  }
  
  //
  // Ssh-client process created. From now we can send raw commands to
  // sftp server throught fdin_, and read server's answer from fdout_.
  // fderr_ should receives errors messages from ssh-client process.
  //

  //
  // Send Init request to negotiate sftp version.
  //

  initMsg.size_ = InvertBytes(sizeof(initMsg) - 4);
  
  FAIL(AtomicWrite(NULL, &initMsg, sizeof(initMsg)));

  FAIL(AtomicRead(NULL, &retSize, sizeof(retSize)));
  
  retSize = InvertBytes(retSize);
  
  versionMsg = (SftpVersionMsg *) malloc(retSize);
  
  FAIL(AtomicRead(NULL, &(versionMsg -> type_), retSize));

  //
  // Check retrieved result from server.
  //
  
  FAIL(versionMsg -> type_ != SSH2_FXP_VERSION);

  DBG_MSG("SftpClient::connect : Server version : %d", 
              InvertBytes(versionMsg -> version_));
  
  failedConnections_ = 0;
  firstConnection_   = 0;
  
  //
  // Init cache.
  //

  FAIL(InitCache());
  
  //
  // Disable server side case sensitive.
  //
  
  // setCaseSensitive(0);
  
  //
  // Clean up.
  //
  
  exitCode = 0;

  fail:
 
  if (exitCode)
  {
    Error("ERROR: Cannot connect to sftp server.\n"
              "Error code is: %u.\n", (int) GetLastError());
    
    disconnect();
  }
 
  if (versionMsg)
  {
    free(versionMsg);
  }
  
  if (Not(slaveMode_))
  {
    //
    // Close child side's handles.
    //
  
    DBG_MSG("Going to shutdown [%d].", pipeIn[0]);
    Io::close(pipeIn[0]);
    DBG_MSG("Shutted [%d].", pipeIn[0]);

    DBG_MSG("Going to shutdown [%d].", pipeOut[1]);
    Io::close(pipeOut[1]);
    DBG_MSG("Shutted [%d].", pipeOut[1]);    
    
    DBG_MSG("Going to shutdown [%d].", pipeErr[1]);
    Io::close(pipeErr[1]);
    DBG_MSG("Shutted [%d].", pipeErr[1]);
  }
  
  DBG_LEAVE("SftpClient::connect");
  
  return exitCode;
}

//
//
//

SftpMsg *SftpClient::allocStringMessage(Int type, Int id, const Char *str, 
                                            const void *data, Int dataSize)
{
  DBG_ENTER("SftpClient::allocStringMessage");
  
  Int exitCode = -1;
  
  SftpStrMsg *msg = NULL;
  
  Int strLen = 0;

  FAIL(str == NULL);

  //
  // Allocate buffer for string message.
  //
  
  DBG_MSG("SftpClient::allocStringMessage : Allocating buffer...");
  
  strLen = StringLength(str);
  
  msg = (SftpStrMsg *) malloc(sizeof(SftpStrMsg) + strLen + 1 + dataSize);
  
  FAIL(msg == NULL);
  
  //
  // Prepare string message packet.
  //
  
  msg -> type_   = type;
  msg -> id_     = id;
  msg -> strLen_ = InvertBytes(strLen);

  StringCopy(msg -> str_, str);

  //
  // Add extra data after string if needed.
  //
  
  if (data != NULL)
  {
    MemoryCopy(msg -> str_ + strLen, data, dataSize);
  }  
  
  msg -> size_ = SFTP_STR_HEAD_SIZE + strLen + dataSize;
 
  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Fatal("ERROR: Cannot alloc string message."
              " Error code is: %d.", (Int) GetLastError());
  }
  
  DBG_LEAVE("SftpClient::allocStringMessage");
  
  return (SftpMsg *) msg;
}

// 
// Check is file exists and optionally retrieve file attributes.
//
// path         - full remote path on server (IN).
// extendedInfo - info about remote file or unused when NULL passed. (OUT).
//
// RETURNS: 0 if OK.
//

Int SftpClient::stat(const Char *path, BY_HANDLE_FILE_INFORMATION *extendedInfo, Int reconn)
{                                          
  DBG_ENTER("SftpClient::stat");

  Int exitCode   = -1;
  Int serverCode = 0;

  SftpMsg *ret = NULL;
  
  Int msgId = IOFifoGenerateUniqueId();
  
  IOFifoEntry iorq = {0};
  
  FAIL(path == NULL);
  
  //
  // Prepare IO request.
  // Send Send SSH2_FXP_STAT message.
  //
  
  DBG_MSG("SftpClient::stat : Sending SSH2_FXP_STAT message...");

  iorq.id_       = msgId;
  iorq.inpData_  = allocStringMessage(SSH2_FXP_STAT_VERSION_0, msgId, path);
  iorq.poutData_ = &ret;
  iorq.dynamic_  = 1;
  iorq.inpFree_  = 1;
  
  FAIL(IORequest(&iorq, timeout_));
  
  //
  // Process received answer.
  //
  
  switch(ret -> type_)
  {
    //
    // SSH2_FXP_ATTR message. We should get resource's
    // attributes here.
    //
    
    case SSH2_FXP_ATTRS:
    {
      DBG_MSG("SftpClient::stat : Received SSH2_FXP_ATTRS.");
      
      SftpAttrsMsg *attrsMsg = (SftpAttrsMsg *) ret;

      FAIL(attrsMsg -> id_ != msgId);

      //
      // If extended request decode reveived attributes.
      //

      if (extendedInfo)
      {
        DBG_MSG("SftpClient::stat : Decoding attributes...");

        Char *p = (Char *) attrsMsg -> data_;

        Int flags   = PopInt32(p);
        Int uid     = 0;
        Int gid     = 0;
        Int perm    = 0;

        unsigned long long Int atime = 0;
        unsigned long long Int mtime = 0;

        //
        //
        //

        extendedInfo -> nNumberOfLinks = 1;
        extendedInfo -> nFileIndexHigh = 0;
        extendedInfo -> nFileIndexLow  = 0;

        //
        // Size.
        //

        if (flags & SSH2_FILEXFER_ATTR_SIZE)
        {
          extendedInfo -> nFileSizeHigh = PopInt32(p);
          extendedInfo -> nFileSizeLow  = PopInt32(p);
        }

        //
        // Guid/Uid flags. We ignore it on Windows.
        //

        if (flags & SSH2_FILEXFER_ATTR_UIDGID)
        {
          uid = PopInt32(p);
          gid = PopInt32(p);
        }

        //
        // Permissions.
        //

        if (flags & SSH2_FILEXFER_ATTR_PERMISSIONS)
        {
          perm = PopInt32(p);

          if (ISDIR(perm))
          {
            extendedInfo -> dwFileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
          }
        }

        //
        // Last acces and last modify times.
        // We need translate time to Winapi comatible.
        //

        if (flags & SSH2_FILEXFER_ATTR_ACMODTIME)
        {
          //
          // Pop acces time and modify times.
          //

          atime = PopInt32(p);
          mtime = PopInt32(p);

          //
          // Convert time from UNIX UTC (from 1970)
          // to FILETIME UTC (from 1601).
          //

          LONGLONG atimell = Int32x32To64(atime, 10000000) + 116444736000000000LL;
          LONGLONG mtimell = Int32x32To64(mtime, 10000000) + 116444736000000000LL;

          extendedInfo -> ftLastAccessTime.dwLowDateTime  = (DWORD) atimell;
          extendedInfo -> ftLastAccessTime.dwHighDateTime = atimell >> 32;

          extendedInfo -> ftLastWriteTime.dwLowDateTime   = (DWORD) mtimell;
          extendedInfo -> ftLastWriteTime.dwHighDateTime  = mtimell >> 32;

          //
          // We don't get creation time from server.
          // We assume creation = last write here.
          //

          extendedInfo -> ftCreationTime.dwLowDateTime  = (DWORD) mtimell;
          extendedInfo -> ftCreationTime.dwHighDateTime = mtimell >> 32;
        }
        
        extendedInfo -> dwFileAttributes |= FILE_ATTRIBUTE_NOT_CONTENT_INDEXED
                                         |  FILE_ATTRIBUTE_OFFLINE;
      }
      
      break;
    }

    //
    // SSH2_FXP_STATUS message. This means error or resource
    // not exists.
    //             
   
    case SSH2_FXP_STATUS:
    {
      DBG_MSG("SftpClient::stat : Received SSH2_FXP_STATUS.");

      SftpStatusMsg *statusMsg = (SftpStatusMsg *) ret;

      FAIL(statusMsg -> id_ != msgId);

      serverCode = InvertBytes(statusMsg -> status_);

      FAIL(serverCode != SSH2_FX_OK);
      
      break;
    }
    
    //
    // Unexpected message type.
    //
    
    default:
    {
      DBG_MSG("SftpClient::stat : ERROR: Unexpected message received [0x%x].",
                  ret -> type_);

      FAIL(1);
    }
  }
  
  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    DBG_MSG("Cannot stat remote file.\n"
                "Error code is: %u.\nServer status is: %u.",
                    GetLastError(), serverCode);
  }
  
  if (ret)
  {
    free(ret);
  }
  
  DBG_LEAVE("SftpClient::stat");
  
  //
  // Dropped connection. Try once.
  //
  
  if (Not(reconn) && (iorq.state_ == ERROR_STATE || serverCode == 4))
  {
    return stat(path, extendedInfo, 1);
  }
  
  return exitCode;
}

//
// Open remote file or directory.
//
// path        - full path to remote directory (IN).
// mode        - sftp access mode (IN).
// localHandle - target local handle. (IN/OPT).
// isDir       - set to 1 if opened handle is directory (OUT/OPT).
//
// RETURNS: Local handle to opened resource or -1 if error.
//

Int64 SftpClient::open(const Char *path, Int mode, Int64 localHandle, Int *isDir)
{
  DBG_ENTER("SftpClient::open");
  
  Int exitCode = -1;

  Int msgId = IOFifoGenerateUniqueId();
    
  SftpMsg *ret = NULL;

  BY_HANDLE_FILE_INFORMATION fileInfo = {0};
  
  Int64 handle = 0;

  Int handleLen    = 0;
  Int serverStatus = 0;
  
  Char *raw1 = NULL;
  Char *raw2 = NULL;
  
  SftpResource sftpRes = {0};
  
  IOFifoEntry iorq = {0};

  //
  // We can't open non existing resources, so we need
  // to try STAT message first to ensure does path exist.
  //
  
  if (Not(mode & SSH2_FXF_CREAT))
  {
    
/*
  FIXME: Use cache if possible.
  
    Int exist = (CacheRemoteStat(path, &fileInfo) == 0);
*/
    Int exist = stat(path, &fileInfo) == 0;
                      
    FAILEX(Not(exist), "SftpClient::open : [%s] not exists.\n", path);                      
  }

  //
  // Target is directory. Send SSH2_FXP_OPENDIR message.
  //
  
  if (fileInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
  {
    DBG_MSG("SftpClient::open : Preparing SSH2_FXP_OPENDIR request...");
    
    iorq.inpData_ = allocStringMessage(SSH2_FXP_OPENDIR, msgId, path);

    sftpRes.isDir_ = 1;
   
    if (isDir)
    {
      *isDir = 1;
    }
  }
    
  //
  // Target is file. Send SSH2_FXP_OPEN message.
  //
    
  else
  {
    if (isDir)
    {
      *isDir = 0;
    }
      
      
    DBG_MSG("SftpClient::open : Preparing SSH2_FXP_OPEN request...");

    Int data[2] = {InvertBytes(mode), 0};
    
    iorq.inpData_ = allocStringMessage(SSH2_FXP_OPEN, msgId, path, data, 8);
  }
  
  //
  // We expect SSH2_FXP_HANDLE message at response.
  //
  
  iorq.id_       = msgId;
  iorq.poutData_ = (SftpMsg **) &ret;
  iorq.inpFree_  = 1;                     
  iorq.dynamic_  = 1;                     

  //
  // Send request to server and wait for response.
  //
   
  FAIL(IORequest(&iorq, timeout_));

  //
  // Parse response.
  //

  switch(ret -> type_)
  {
    //
    // SSH2_FXP_HANDLE with opened server side handle.
    // It means success.
    //
    
    case SSH2_FXP_HANDLE:
    {
      DBG_MSG("SftpClient::open : Received SSH2_FXP_HANDLE.");
      
      SftpHandleMsg *handleMsg = (SftpHandleMsg *) ret;
      
      FAILEX(handleMsg -> id_ != msgId,
                 "ERROR: Unexpected packet id [%d].\n", handleMsg -> id_);

      //
      // We reveive big endian 32 or 64 bit handle here.
      // Convert handle to 64-bit little endian.
      //

      DBG_MSG("SftpClient::open : Reversing bytes in handle...");

      handleLen = InvertBytes(handleMsg -> handleLen_);

      FAIL((Int64) handleLen > sizeof(Int64));

      raw1 = (Char *) &handleMsg -> handle_;
      raw2 = (Char *) &handle;

      for (int i = 0; i < handleLen; i++)
      {
        raw2[handleLen - i - 1] = raw1[i];
      }

      serverHandleSize_ = handleLen;

      //
      // Generate new local handle if needed.
      //

      if (localHandle == -1)
      {
        localHandle = (++ SftpResource::localHandleCount_);
      }

      //
      // Set map for new opened resource.
      //

      sftpRes.remoteHandle_ = handle;
      sftpRes.localHandle_  = localHandle;
      sftpRes.remotePath_   = path;
      sftpRes.mode_         = mode;
      sftpRes.dropped_      = 0;
      
      //
      // Add ISDIR flag to handle value.
      // It is used in remove time to easy determine
      // which function to use - rmdir() or remove().
      //
      
      if (sftpRes.isDir_)
      {
        sftpRes.localHandle_ |= DIR_HANDLE_MASK;
      }

      DBG_MSG("SftpClient::open : Adding new resource [%s][%d] -> [%lld][%lld]...",
                  sftpRes.remotePath_.c_str(), sftpRes.mode_,
                      sftpRes.localHandle_, sftpRes.remoteHandle_);

      openedResources_[sftpRes.localHandle_] = sftpRes;

      DBG_MSG("SftpClient::open : Received %d-bit handle [%d]", 
                  handleLen * 8, handle);

      break;
    }

    //
    // SSH2_FXP_STATUS means errors on server side.
    // Probably access denied.
    //
    
    case SSH2_FXP_STATUS:
    {
      DBG_MSG("SftpClient::open : Received SSH2_FXP_STATUS.");

      SftpStatusMsg *statMsg = (SftpStatusMsg *) ret;

      FAILEX(statMsg -> id_ != msgId,
                 "ERROR: Unexpected packet id [%d].\n", statMsg -> id_);

      serverStatus = InvertBytes(statMsg -> status_);
      
      FAIL(statMsg -> status_ != SSH2_FX_OK);

      break;
    }
    
    //
    // Fatal error. Unexpected packet type.
    //
    
    default:
    {
      FAILEX(1, "ERROR: Unexpected packet type [%d].\n", ret -> type_);
    }
  }
    
  //
  // Cleanup.
  //
  
  exitCode = 0;
  
  fail:

  if (exitCode < 0)
  {
    DBG_MSG("Cannot open remote resource.\n"
                "Error code is: %u.\nServer status is : %d.", 
                    GetLastError(), serverStatus);
  }

  if (ret != NULL)
  {
    free(ret);
  }
  
  DBG_LEAVE("SftpClient::open");

  //
  // Connection dropped. Try again if possible.
  // 
  
  if (iorq.state_ == ERROR_STATE)
  {
    return open(path, mode);
  }

  return exitCode ? -1 : sftpRes.localHandle_;
}

//
// Close remote file or directory.
//
// localHandle - value retrieved from open() function (IN).
//
// RETURNS: 0 if OK.
//

Int SftpClient::close(Int64 localHandle)
{
  DBG_ENTER("SftpClient::close");
  
  Int exitCode   = -1;
  Int serverCode = 0;
  
  Int msgId = IOFifoGenerateUniqueId();
  
  SftpResource sftpRes = {0};
  
  map<Int64, SftpResource>::iterator it;
  
  SftpHandleMsg msg = {0};
  SftpStatusMsg ret = {0};
  
  IOFifoEntry iorq = {0};
   
  msg.size_      = sizeof(SftpHandleMsg);
  msg.type_      = SSH2_FXP_CLOSE;
  msg.id_        = msgId;
  msg.handleLen_ = InvertBytes(serverHandleSize_);

  //
  // Retrieve remote handle from resource map.
  //
  
  it = openedResources_.find(localHandle);
 
  if (it == openedResources_.end())
  {
    SetLastError(ERROR_INVALID_HANDLE);
  
    goto fail;
  }
  else
  {
    sftpRes = it -> second;
  }
  
  if (Not(sftpRes.dropped_))
  {
    //
    // Send 32 bit handle.
    //
  
    if (serverHandleSize_ == 4)
    {
      DBG_MSG("Sending 32-bit handle...");
    
      msg.handle_[0] = InvertBytes(sftpRes.remoteHandle_);
    }

    //
    // Send 64 bit handle.
    //
  
    else
    {
      DBG_MSG("Sending 64-bit handle...");
     
      Int64 *raw = (Int64 *) msg.handle_;

      *raw = InvertBytes64(sftpRes.remoteHandle_);
    }

    //
    // Send SSH2_FXP_CLOSE message.
    // Expect SSH2_FXP_STATUS message.
    //

    DBG_MSG("SftpClient::close : Sending SSH2_FXP_CLOSE message...");
  
    iorq.id_      = msgId;
    iorq.inpData_ = (SftpMsg *) &msg;
    iorq.outData_ = (SftpMsg *) &ret;
    iorq.outSize_ = sizeof(ret);
  
    FAIL(IORequest(&iorq, timeout_));

    //
    // Check received message.
    //
    
    FAIL(ret.type_ != SSH2_FXP_STATUS);
  
    FAIL(ret.id_ != msgId);

    serverCode = InvertBytes(ret.status_);

    FAIL(serverCode != SSH2_FX_OK);
  }

  //
  // Remove item from resource map.
  //
  
  openedResources_.erase(it);
  
  //
  // Cleanup.
  //
  
  exitCode = 0;

  fail:

  if (exitCode)
  {
    DBG_MSG("ERROR: Cannot close remote resource."
                " Error code is: %u. Server status is: %u", 
                    GetLastError(), serverCode);
  }
  
  DBG_LEAVE("SftpClient::close");
  
  return exitCode;

  return 0;
}

//
// Read directory.
//
// handle  - value retrieved from open() function (IN).
// addData - callback function called for every listed file (IN).
//
// RETURNS: 0 if OK.
//

Int SftpClient::readDir(WIN32_FIND_DATAW **buf, Int *count, Int64 localHandle, Int reconn)
{
  DBG_ENTER("SftpClient::readDir");
  
  Int exitCode = -1;
  
  Int msgId = -1;
  
  Int goOn = 1;
  
  Int serverStatus = 0;

  Int capacity = 0;
  
  Int iter = 0;
  
  SftpHandleMsg msg = {0};
  
  SftpNameMsg *ret = NULL;
  
  SftpResource sftpRes = {0};
  
  IOFifoEntry iorq = {0};
  
  //
  // Check args.
  //
    
  FAIL(buf == NULL);
  
  *buf = NULL;
  
  FAIL(count == NULL);
  
  //
  // Get remote handle.
  //
  
  FAIL(getSftpResource(&sftpRes, localHandle));
  
  //
  // Send requests for data until EOF or error.
  //
  
  *count = 0;
  
  while(goOn)
  {
    iter ++;
    
    //
    // Send SSH2_FXP_READDIR message.
    //

    msgId = IOFifoGenerateUniqueId();
    
    msg.size_      = sizeof(SftpHandleMsg);
    msg.type_      = SSH2_FXP_READDIR;
    msg.id_        = msgId;
    msg.handleLen_ = InvertBytes(serverHandleSize_);

    //
    // Send 32 bit handle.
    //
  
    if (serverHandleSize_ == 4)
    {
      DBG_MSG("Sending 32-bit handle...");
    
      msg.handle_[0] = InvertBytes(sftpRes.remoteHandle_);
    }

    //
    // Send 64 bit handle.
    //
  
    else
    {
      DBG_MSG("Sending 64-bit handle...");
    
      Int64 *raw = (Int64 *) msg.handle_;

      *raw = InvertBytes64(sftpRes.remoteHandle_);
    }
    
    //
    // Send request to server.
    //
    
    iorq.id_       = msgId;
    iorq.inpData_  = (SftpMsg *) &msg;
    iorq.poutData_ = (SftpMsg **) &ret;
    iorq.dynamic_  = 1;
  
    FAIL(IORequest(&iorq, timeout_));
    
    FAILEX(ret -> id_ != msgId, "ERROR: Unexpected packet id.\n");

    //
    // Interprate received message.
    //

    switch(ret -> type_)
    {
      //
      // Status message. This means EOF or error.
      //

      case SSH2_FXP_STATUS:
      {
        DBG_MSG("SftpClient::readDir : SSH2_FXP_STATUS received.");

        SftpStatusMsg *stat = (SftpStatusMsg *) ret;
        
        serverStatus = InvertBytes(stat -> status_);

        //
        // If we get EOF on first iteration silently
        // reopen handle on server.
        // This trick allow us to use lazy handles
        // on directories too, because we know we
        // can list files always even after EOF reached.
        //
        
        if (serverStatus == SSH2_FX_EOF && iter == 1)
        {
          DBG_MSG("SftpClient::readDir : EOF on first packet. Reopening...");
          
          FAIL(close(localHandle));
          
          FAIL(open(sftpRes.remotePath_.c_str(), SSH2_FXF_READ, localHandle) < 0);
        
          FAIL(getSftpResource(&sftpRes, localHandle));
        }                        

        //
        // Normally (not first iteration)
        // we stop on EOF.
        //
        
        else
        {
          goOn = 0;

          FAIL(serverStatus != SSH2_FX_EOF);
        }  

        break;
      }

      //
      // SSH2_FXP_NAME. Next portion of data.
      //

      case SSH2_FXP_NAME:
      {
        DBG_MSG("SftpClient::readDir : SSH2_FXP_NAME received.");
        
        Char *p = (Char *) &ret -> data_;

        Int lastCount = *count;

        *count = lastCount + InvertBytes(ret -> count_);
       
        DBG_MSG("SftpClient::readDir : lastCount = [%d], count = [%d]", lastCount, *count);

        //
        // Alloc/Realloc output buffer.
        //
        
        if (*count > capacity)
        {
          Int oldCapacity = capacity;

          capacity = (*count) * 2;
          
          Int delta = capacity - oldCapacity;
          
          
          DBG_MSG("SftpClient::readDir : Resizing buffer from %d to %d...",
                      oldCapacity, capacity);
          
          *buf = (WIN32_FIND_DATAW *) realloc(*buf, capacity * sizeof(WIN32_FIND_DATAW));

          FAILEX(*buf == NULL, "ERROR: Out of memory.\n");
          
          ZeroMemory(*buf + oldCapacity, delta * sizeof(WIN32_FIND_DATAW));
        }  
        
        //
        // Iterate through every elements.
        //
                
        for (int i = lastCount; i < *count; i++)
        {
          Char name[MAX_PATH]     = {0};
          Char longName[MAX_PATH] = {0};

          Int nameLen     = 0;
          Int longNameLen = 0;

          //
          // Get element's name.
          //

          PopString(name, p, &nameLen, MAX_PATH);

          PopString(longName, p, &longNameLen, MAX_PATH);

          //
          // We expecting UTF8 from network.
          // Convert UTF8 to UTF16.
          //

          MultiByteToWideChar(CP_UTF8, 0, name, -1,
                                  (*buf)[i].cFileName, 
                                      sizeof((*buf)[i].cFileName));

          //
          // Get element's attributes.
          //

          Int flags = PopInt32(p);
          Int uid   = 0;
          Int gid   = 0;
          Int perm  = 0;
          
          unsigned long long Int atime = 0;
          unsigned long long Int mtime = 0;
          
          //
          // Size.
          //
          
          if (flags & SSH2_FILEXFER_ATTR_SIZE)
          {
            (*buf)[i].nFileSizeHigh = PopInt32(p);
            (*buf)[i].nFileSizeLow  = PopInt32(p);
          }

          //
          // Guid/Uid flags. We ignore it on Windows.
          //
          
          if (flags & SSH2_FILEXFER_ATTR_UIDGID)
          {
            uid = PopInt32(p);
            gid = PopInt32(p);
          }

          //
          // Permissions.
          //
          
          if (flags & SSH2_FILEXFER_ATTR_PERMISSIONS)
          {
            perm = PopInt32(p);
            
            if (ISDIR(perm))
            {
              (*buf)[i].dwFileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
            }
          }

          //
          // Skip adding the attribute FILE_ATTRIBUTE_OFFLINE.
          //

          /*

          //
          // Disable indexing service.
          //

          (*buf)[i].dwFileAttributes |= FILE_ATTRIBUTE_NOT_CONTENT_INDEXED 
                                     |  FILE_ATTRIBUTE_OFFLINE;
          */

          //
          // Last access and last modify times.
          // We need translate time to Winapi comatible.
          //
                        
          if (flags & SSH2_FILEXFER_ATTR_ACMODTIME)
          {
            //
            // Pop acces time and modify times.
            //
            
            atime = PopInt32(p);
            mtime = PopInt32(p);
            
            //
            // Convert time from UNIX UTC (from 1970)
            // to FILETIME UTC (from 1601).
            //
            
            LONGLONG atimell = Int32x32To64(atime, 10000000) + 116444736000000000LL;
            LONGLONG mtimell = Int32x32To64(mtime, 10000000) + 116444736000000000LL;
            
            (*buf)[i].ftLastAccessTime.dwLowDateTime  = (DWORD) atimell;
            (*buf)[i].ftLastAccessTime.dwHighDateTime = atimell >> 32;

            (*buf)[i].ftLastWriteTime.dwLowDateTime   = (DWORD) mtimell;
            (*buf)[i].ftLastWriteTime.dwHighDateTime  = mtimell >> 32;

            //
            // We don't get creation time from server.
            // We assume creation = last write here.
            //
            
            (*buf)[i].ftCreationTime.dwLowDateTime  = (DWORD) mtimell;
            (*buf)[i].ftCreationTime.dwHighDateTime = mtimell >> 32;
          }

          //
          // Skip other, additional attributes.
          //
          
          if (flags & SSH2_FILEXFER_ATTR_EXTENDED)
          {
            Int n = PopInt32(p);


            for (Int i = 0; i < n; i++)
            {
              SkipString(p);
              SkipString(p);
            }
          }
        }

        break;
      }

      //
      // Unexpected packet type. Break.
      //

      default:
      {
        Fatal("SftpClient::readDir : ERROR: Unexpected packet received"
                  " [%d].", ret -> type_);

        FAIL(1);
      }
    }
  }

  DBG_MSG("SftpClient::readDir : %d elements listed.", *count);
  
  //
  // Cleanup.
  //
  
  exitCode = 0;
  
  fail:

  if (exitCode)
  {
    DBG_MSG("ERROR: Cannot read remote directory."
                " Server status is: %d.", serverStatus);

    //
    // Free output buffer.
    //
    
    if (buf && *buf)
    {
      free(*buf);
      
      *buf = NULL;
    }
  }
  
  if (ret)
  {
    free(ret);
  }
  
  DBG_LEAVE("SftpClient::readDir");
  
  //
  // Droped connection try once again.
  //
  
  if (Not(reconn) && (exitCode && GetLastError() == WAIT_TIMEOUT && failedConnections_ < 3))
  {
    return readDir(buf, count, localHandle, 1);
  }
  
  return exitCode;
}

//
// Read data from remote file.
//

Int SftpClient::read(Int64 localHandle, Char *buffer,
                         Int64 offset, Int bytesToRead, Int reconn)
{
  DBG_ENTER("SftpClient::read");
  
  Int exitCode     = -1;
  Int serverStatus = 0;
  
  Int goOn = 1;
  
  SftpReadMsg msg = {0};

  SftpMsg *ret = NULL;
  
  Int msgId = -1;
  
  Int readed = 0;
  
  double refTime = GetTimeMs();

  SftpResource sftpRes = {0};
  
  IOFifoEntry iorq = {0};
  
  //
  //
  //
  
  //Char *originalBuffer = buffer;

  //Int64 originalOffset = offset;
  
  //Int originalBytesToRead = bytesToRead;
  
  //
  // Get remote handle.
  //
  
  FAIL(getSftpResource(&sftpRes, localHandle));

  DBG_MSG("SftpClient::read(localHandle = [%lld], remoteHandle = [%lld], "
              "buffer = [%p], offset = [%lld], bytesToRead = [%d])...", 
                  localHandle, sftpRes.remoteHandle_, buffer, offset, 
                      bytesToRead);
  
  FAIL(sftpRes.isDir_);
  
  //
  // Read packets unitil:
  //
  // - EOF.
  // - error.
  // - read time can exceed system I/O timeout 
  //   (for big caller buffer or slow network).
  //
  
  while(goOn)
  {
    //
    // Prepare next SSH2_FXP_READ message.
    //
  
    msgId = IOFifoGenerateUniqueId();
    
    msg.size_        = sizeof(SftpReadMsg);
    msg.type_        = SSH2_FXP_READ;
    msg.id_          = msgId;
    msg.handleLen_   = InvertBytes(sizeof(msg.handle_));
    msg.handle_      = InvertBytes(sftpRes.remoteHandle_);
    msg.offsetLow_   = InvertBytes(LOW_DWORD(offset));
    msg.offsetHigh_  = InvertBytes(HIGH_DWORD(offset));
    msg.bytesToRead_ = InvertBytes(bytesToRead - readed);

    //
    // Send SSH2_FXP_READ message.
    // Expect SSH2_FXP_DATA or SSH2_FXP_STATUS.
    //
  
    DBG_MSG("SftpClient::read : Sending SSH2_FXP_READ message...");
    
    iorq.id_       = msgId;
    iorq.inpData_  = (SftpMsg *) &msg;
    iorq.poutData_ = &ret;
    iorq.dynamic_  = 1;
  
    FAIL(IORequest(&iorq, timeout_));
    
    //
    // Parse response from server.
    //
    
    switch(ret -> type_)
    {
      //
      // Next portion of readed data.
      //
      
      case SSH2_FXP_DATA:
      {
        DBG_MSG("SftpClient::read : SSH2_FXP_DATA received...");
        
        Int len = 0;
        
        SftpStrMsg *data = (SftpStrMsg *) (ret);

        //
        // Put reveived data to output buffer.
        //
        
        len = InvertBytes(data -> strLen_);
        
        DBG_MSG("SftpClient::read : Received %d bytes...", len);
        
        memcpy(buffer, data -> str_, len);
        
        readed += len;
        
        buffer += len;
        
        offset += len;
        
        reconn = 0;
        
        //
        // If we readed all expected data stop listening.
        //
        
        if (readed == bytesToRead)
        {
          DBG_MSG("SftpClient::read : Reading completed.");
          
          goOn = 0;
        }
        
        break;
      }
      
      //
      // Status message means EOF or error.
      //
      
      case SSH2_FXP_STATUS:
      {
        DBG_MSG("SSH2_FXP_STATUS received...");
        
        SftpStatusMsg *status = (SftpStatusMsg *) ret;
        
        serverStatus = InvertBytes(status -> status_);
        
        FAIL(serverStatus != SSH2_FX_EOF);
        
        goOn = 0;
        
        break;
      }
      
      //
      // Unexpected packet's type. 
      //
      
      default:
      {
        DBG_MSG("SftpClient::read : ERROR: Unexpected packet received [%u].",
                    ret -> type_);
                    
        FAIL(1);                    
      }
    }

    //
    // Check how much time we spend in read loop.
    // If we stucked here too long return what we readed 
    // (readed < byesToRead) to avoid reach of system I/O timeout.
    //
    
    if (GetTimeMs() - refTime > MAX_TIME_PER_CALLBACK)
    {
      DBG_MSG("SftpClient :: read : Callback's time limit reached.");
      
      goOn = 0;
    }
  }

  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    DBG_MSG("ERROR: Cannot read from remote file. Error code is: %u."
                " Server status: is %u.", GetLastError(), serverStatus);
  }
  
  if (ret)
  {
    free(ret);
  }
  
  DBG_LEAVE("SftpClient::read");
  
  //
  // Dropped connection. Try once again if possible.
  //
  
  if (Not(reconn) && (iorq.state_ == ERROR_STATE || serverStatus == 4))
  {
    Int nextReaded = read(localHandle, buffer, offset, bytesToRead - readed, 1);
    
    return (nextReaded == -1) ? -1 : readed + nextReaded;
  }

  return exitCode ? -1 : readed;
}
  
//
// Write data to remote file.
//

Int SftpClient::write(Int64 localHandle, Char *buffer, 
                          long long Int offset, Int bytesToWrite, Int reconn)
{
  DBG_ENTER("SftpClient::write");

  Int exitCode     = -1;
  Int serverStatus = 0;
  
  SftpWriteMsg *msg = NULL;

  SftpStatusMsg ret = {0};
  
  SftpResource sftpRes = {0};
  
  IOFifoEntry iorq = {0};
  
  Int msgId = -1;

  Int written = 0;
  
  Int pieceSize = 0;
  
  Int goOn = 1;
  
  double refTime = GetTimeMs();

  #define MAX_PIECE_SIZE (32768)
  
  //
  // Get remote handle.
  //
  
  FAIL(getSftpResource(&sftpRes, localHandle));
  
  //
  // Allocate buffor for request.
  //
  
  msg = (SftpWriteMsg *) malloc(sizeof(SftpWriteMsg) + MAX_PIECE_SIZE + 4);

  FAILEX(msg == NULL, "SfptClient::write : ERROR: Cannot allocate input buffer.\n");
  
  //
  // Write packets until all data sent or error.
  //
  
  while(goOn)
  {
    //
    // Compute size of current piece.
    //
    
    pieceSize = MIN(bytesToWrite - written, MAX_PIECE_SIZE);    
    
    //
    // Prepare next SSH2_FXP_WRITE message.
    //
  
    msgId = IOFifoGenerateUniqueId();
    
    msg -> size_         = sizeof(SftpWriteMsg) - 1 + pieceSize;
    msg -> type_         = SSH2_FXP_WRITE;
    msg -> id_           = msgId;
    msg -> handleLen_    = InvertBytes(sizeof(msg -> handle_));
    msg -> handle_       = InvertBytes(sftpRes.remoteHandle_);
    msg -> offsetLow_    = InvertBytes(LOW_DWORD(offset));
    msg -> offsetHigh_   = InvertBytes(HIGH_DWORD(offset));
    msg -> bytesToWrite_ = InvertBytes(pieceSize);
    
/*
  FIXME: Avoid unnecesery copy of file data buffer.
*/  

    MemoryCopy(&msg -> data_, buffer, pieceSize);
    
    //
    // Send request onto server and wait for response.
    //

    iorq.id_      = msgId;
    iorq.inpData_ = (SftpMsg *) msg;
    iorq.outData_ = (SftpMsg *) &ret;
    iorq.outSize_ = sizeof(ret);
    
    FAIL(IORequest(&iorq, timeout_));
        
    //
    // Parse response from server.
    // We expect SSH2_FXP_STATUS here.
    //

    FAIL(ret.type_ != SSH2_FXP_STATUS);
    
    FAIL(ret.id_ != msgId);

    serverStatus = InvertBytes(ret.status_);
        
    FAIL(serverStatus != SSH2_FX_OK);
        
    //
    // Move to next piece to send.
    //

    written += pieceSize;
        
    buffer += pieceSize;

    offset += pieceSize;
    
    reconn = 0;
    
    //
    // Check do we still have some to write.
    //
    
    if (written >= bytesToWrite)
    {
      goOn = 0;
    }
    
    //
    // Check how much time we spend in write loop.
    // If we stucked here too long return what we readed 
    // (written < byesToWrite) to avoid reach of system I/O timeout.
    //
    
    if (GetTimeMs() - refTime > MAX_TIME_PER_CALLBACK)
    {
      DBG_MSG("SftpClient :: write : Callback's time limit reached.");
      
      goOn = 0;
    }
  }

  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    DBG_MSG("ERROR: Cannot write to remote file. Error code is: %u."
                " Server status is: %u.", GetLastError(), serverStatus);
  }
  
  if (msg != NULL)
  {
    free(msg);
  }
  
  DBG_LEAVE("SftpClient::write");
  
  //
  // Dropped connection. Try once again if possible.
  //
  
  if (Not(reconn) && (iorq.state_ == ERROR_STATE || serverStatus == 4))
  {
    Int nextWritten = write(localHandle, buffer, offset, bytesToWrite - written, 1);
    
    return (nextWritten == -1) ? -1 : written + nextWritten;
  }

  return exitCode ? -1 : written;
}

//
// Create remote directory.
//

Int SftpClient::mkdir(const Char *path, Int reconn)
{
  DBG_ENTER("SftpClient::mkdir");
  
  Int exitCode = -1;
  
  Int msgId = IOFifoGenerateUniqueId();
  
  Int serverStatus = 0;
  
  SftpMkdirMsg msg  = {0};
  SftpStatusMsg ret = {0};
  
  IOFifoEntry iorq = {0};
    
  //
  // Prepare SSH2_FXP_MKDIR message.
  //
  
  msg.size_    = sizeof(SftpMkdirMsg);
  msg.type_    = SSH2_FXP_MKDIR;
  msg.id_      = msgId;
  msg.pathLen_ = InvertBytes(MAX_PATH);
  
  StringSizeCopy(msg.path_, path, MAX_PATH);
  
  //
  // Send SSH2_FXP_MKDIR message.
  // Expect SSH2_FXP_STATUS message.
  //
  
  DBG_MSG("SftpClient::mkdir : Sending SSH2_FXP_MKDIR message...");
    
  iorq.id_      = msgId;
  iorq.inpData_ = (SftpMsg *) &msg;
  iorq.outData_ = (SftpMsg *) &ret;
  iorq.outSize_ = sizeof(ret);
  
  FAIL(IORequest(&iorq, timeout_));

  //
  // Parse response from server.
  //
  
  FAILEX(ret.type_ != SSH2_FXP_STATUS,
             "ERROR: Unexpected packet's type [0x%x].\n", ret.type_);
  
  FAILEX(ret.id_ != msgId, "ERROR: Unexpected packet's id.\n");
  
  serverStatus = InvertBytes(ret.status_);
  
  FAIL(serverStatus != SSH2_FX_OK);
  
  //
  // Clean up.
  //

  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    DBG_MSG("ERROR: Cannot create remote directory. Error code is: %u."
                " Server status is: %u.", GetLastError(), serverStatus);
  }
  
  DBG_LEAVE("SftpClient::mkdir");
 
  //
  // Dropped connection. Try once again if possible.
  //

  if (iorq.state_ == ERROR_STATE)
  {
    return mkdir(path);
  }

  return exitCode;
}

//
// Remove remote file or directory.
//

Int SftpClient::remove(const Char *path, Int isDir, Int reconn)
{
  DBG_ENTER("SftpClient::remove");
  
  Int exitCode = -1;
  
  Int msgId = IOFifoGenerateUniqueId();
  
  Int serverStatus = 0;
  
  SftpStatusMsg ret = {0};
  
  SftpNxMsg msg(msgId, "nxremove", path);
  
  IOFifoEntry iorq = {0};
    
  //
  // Prepare SSH2_FXP_RMDIR or SSH2_FXP_REMOVE request.
  //
  
  iorq.id_      = msgId;
  iorq.inpData_ = (SftpMsg *) &msg;
  iorq.outData_ = (SftpMsg *) &ret;
  iorq.outSize_ = sizeof(ret);
  
  DBG_MSG("SftpClient::remove : Sending 'nxremove' message...");

  //
  // Send request to server and wait for response.
  //
  
  FAIL(IORequest(&iorq, timeout_));

  //
  // Parse response from server.
  //

  FAIL(ret.type_ != SSH2_FXP_STATUS);
  
  FAIL(ret.id_ != msgId);
  
  serverStatus = InvertBytes(ret.status_);
  
  FAIL(serverStatus != SSH2_FX_OK);
  
  //
  // Clean up.
  //

  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    DBG_MSG("ERROR: Cannot remove remote resource. Error code is: %u."
                " Server status is: %u.", GetLastError(), serverStatus);

    exitCode = -translateError(serverStatus);
  }
  
  DBG_LEAVE("SftpClient::remove");
 
  //
  // Dropped connection. Try once again if possible.
  //
  
  if (Not(reconn) && (iorq.state_ == ERROR_STATE || serverStatus == 4))
  {
    return remove(path, isDir, 1);
  }

  return exitCode;
}

//
// Rename remote resource.
//

Int SftpClient::rename(const Char *path1, const Char *path2, Int reconn)
{
  DBG_ENTER("SftpClient::rename");

  Int exitCode = -1;
  
  Int msgId = IOFifoGenerateUniqueId();
  
  Int serverStatus = 0;
  
  SftpRenameMsg msg = {0};
  SftpStatusMsg ret = {0};
  
  IOFifoEntry iorq = {0};
    
  //
  // Prepare SSH2_FXP_MKDIR message.
  //
  
  msg.size_     = sizeof(SftpRenameMsg);
  msg.type_     = SSH2_FXP_RENAME;
  msg.id_       = msgId;
  msg.path1Len_ = InvertBytes(MAX_PATH);
  msg.path2Len_ = InvertBytes(MAX_PATH);
  
  StringSizeCopy(msg.path1_, path1, MAX_PATH);
  StringSizeCopy(msg.path2_, path2, MAX_PATH);
  
  //
  // Send SSH2_FXP_RENAME message.
  //
  
  DBG_MSG("SftpClient::rename : Sending SSH2_FXP_RENAME message...");

  iorq.id_      = msgId;
  iorq.inpData_ = (SftpMsg *) &msg;
  iorq.outData_ = (SftpMsg *) &ret;
  iorq.outSize_ = sizeof(ret);

  FAIL(IORequest(&iorq, timeout_));

  //
  // Parse response from server.
  //
  
  FAIL(ret.type_ != SSH2_FXP_STATUS);
  
  FAIL(ret.id_ != msgId);
  
  serverStatus = InvertBytes(ret.status_);
  
  FAIL(serverStatus != SSH2_FX_OK);
  
  //
  // Clean up.
  //

  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    DBG_MSG("ERROR: Cannot rename remote directory. Error code is: %u."
                " Server status is: %u.", GetLastError(), serverStatus);
  }
  
  DBG_LEAVE("SftpClient::rename");
 
  //
  // Dropped connection. Try once again if possible.
  //
  
  if (Not(reconn) && (iorq.state_ == ERROR_STATE || serverStatus == 4))
  {
    return rename(path1, path2, 1);
  }

  return exitCode;
}

//
// Read data from given fd using non-block reading with timeout.
//
// fd   - file descriptior to read (IN).
// buf  - destination buffer (OUT).
// size - number of bytes to read (IN).
//
// RETURNS: 0 if OK (all expected bytes readed).
//

Int SftpClient::atomicRead(Int fd, Char *buf, Int size)
{
  Int exitCode = -1;
  
  Int readed      = 0;
  Int totalReaded = 0;

  OVERLAPPED ov = {0};  

  FAILEX(size > MAX_PACKET, "ERROR: Packet too big.\n")
  
  //
  // Slave mode uses stdin.
  //

  if (slaveMode_)
  {
    DBG_MSG("SftpClient::atomicRead: Reading from stdin...");

    FAIL(AtomicRead(buf, size, 
                        GetStdHandle(STD_INPUT_HANDLE), 
                            timeout_ * 1000));
  }
  
  //
  // Default mode.
  //
  
  else
  {
    Io::Set setReadFd;

    Io::Time selectTime;  

    //
    // While not readed all expected data.
    //
   
    while(totalReaded < size)
    {
      //
      // Prepare select for non-block read.
      //
    
      Io::zero(&setReadFd);

      Io::set(fd, &setReadFd);
    
      selectTime.tv_sec  = timeout_;
      selectTime.tv_usec = 0;
    
      //
      // Wait for reply.
      //
     
      FAIL(Io::select(fd + 1, &setReadFd, NULL, NULL, &selectTime) < 0);
     
      if (Not(Io::isSet(fd, &setReadFd)))
      {
        Error("ERROR: Server does not reply.\n");
        
        reconnect();
      
        SetLastError(WAIT_TIMEOUT);
      
        return -1;
      }
    
      //
      // Read next portion of data.
      //
    
      readed = Io::read(fd, buf, size - totalReaded);

      FAIL(readed < 0);
    
      //
      // If successfull readed, move pointers and counters.
      //

      buf += readed;
     
      totalReaded += readed;
    }
  }
  
  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Fatal("ERROR: Cannot read packet from server.\n"
                 "Error code is: %u.\n", (Int) GetLastError());
  }

  CloseHandle(ov.hEvent);
  
  return exitCode;
}

//
// Retrieve info about opened sftp resource. If connection dropped
// resource is reopened if needed.
//
// sftpRes     - buffer to fill with resource info (OUT)
// localHandle - local handle to resource returnet by open function (IN)
//
// RETURNS: 0 if OK.
//

Int SftpClient::getSftpResource(SftpResource *sftpRes, Int64 localHandle)
{
  Int exitCode = -1;
  
  map<Int64, SftpResource>::iterator it = openedResources_.find(localHandle);
  
  //
  // Resource not locally opened.
  //
  
  if (it == openedResources_.end())
  {
    SetLastError(ERROR_INVALID_HANDLE);
    
    FAIL(1);
  }

  //
  // Resource locally found. Check what on remote side.
  //
  
  else 
  {
    SftpResource &res = it -> second;
    
    *sftpRes = it -> second;

    //
    // Remote handle dropped. Try reopen.
    //
    
    DBG_MSG("Reusing remote handle [%s][%lld][%lld]...", 
                res.remotePath_.c_str(), localHandle,
                    res.remoteHandle_);
    
    if (res.dropped_)
    {
      DBG_MSG("Remote handle [%s][%lld][...] dropped. Reopening...",
                  res.remotePath_.c_str(), localHandle);
      
      FAIL(open(res.remotePath_.c_str(), res.mode_, localHandle) < 0);
      
      sftpRes -> remoteHandle_ = res.remoteHandle_;
    }
  }
  
  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Error("ERROR: Cannot retrieve remote handle for [%lld].\n"
                 "Error code is : %d.\n", localHandle, (Int) GetLastError());
  }
  
  return exitCode;
}

//
// Retrieve info about remote disk.
//
// WARNING: Server MUST support "statvfs@openssh.com" extension.
//
// stvfs - buffer to store info (see Sftp.h) (OUT)
// path  - remote path to check (IN)
//
// RETURNS: 0 if OK.
//

Int SftpClient::statvfs(SftpStatvfsReplyMsg *stvfs, const Char *remotePath, Int reconn)
{
  DBG_ENTER("SftpClient::getFreeSpace");
  
  Int exitCode = -1;
  
  Int msgId = IOFifoGenerateUniqueId();
  
  SftpStatvfsMsg msg(msgId, remotePath);
  
  SftpMsg *ret = NULL;
  
  IOFifoEntry iorq = {0};
  
  Int serverCode = 0;

  //
  // Send statvfs@openssh.com message.
  // Expect SSH2_FXP_EXTENDED_REPLY message.
  //
  
  DBG_MSG("SftpClient::getFreeSpace : Sending statvfs@openssh.com...");

  iorq.id_       = msgId;
  iorq.inpData_  = (SftpMsg *) &msg;
  iorq.poutData_ = &ret;
  iorq.dynamic_  = 1;
  
  FAIL(IORequest(&iorq, timeout_));

  //
  // Process response.
  //
  
  switch(ret -> type_)
  {      
    case SSH2_FXP_EXTENDED_REPLY:
    {
      DBG_MSG("SftpClient::getFreeSpace : Received SSH2_FXP_EXTENDED_REPLY.");

      memcpy(stvfs, (SftpStatvfsReplyMsg *) ret, sizeof(SftpStatvfsReplyMsg));
      
      FAIL(stvfs -> id_ != msgId); 

      stvfs -> bsize_   = InvertBytes64(stvfs -> bsize_);
      stvfs -> frsize_  = InvertBytes64(stvfs -> frsize_);
      stvfs -> blocks_  = InvertBytes64(stvfs -> blocks_);
      stvfs -> bfree_   = InvertBytes64(stvfs -> bfree_);
      stvfs -> bavail_  = InvertBytes64(stvfs -> bavail_);
      stvfs -> files_   = InvertBytes64(stvfs -> files_);
      stvfs -> ffree_   = InvertBytes64(stvfs -> ffree_);
      stvfs -> favail_  = InvertBytes64(stvfs -> favail_);
      stvfs -> fsid_    = InvertBytes64(stvfs -> fsid_);
      stvfs -> flags_   = InvertBytes64(stvfs -> flags_);
      stvfs -> namemax_ = InvertBytes64(stvfs -> namemax_);

      DBG_MSG("bsize   : [%llu]", stvfs -> bsize_);
      DBG_MSG("frsize  : [%llu]", stvfs -> frsize_);
      DBG_MSG("blocks  : [%llu]", stvfs -> blocks_);
      DBG_MSG("bfree   : [%llu]", stvfs -> bfree_);
      DBG_MSG("bavail  : [%llu]", stvfs -> bavail_);
      DBG_MSG("files   : [%llu]", stvfs -> files_);
      DBG_MSG("ffree   : [%llu]", stvfs -> ffree_);
      DBG_MSG("favail  : [%llu]", stvfs -> favail_);
      DBG_MSG("fsid    : [%llu]", stvfs -> fsid_);
      DBG_MSG("flags   : [%llu]", stvfs -> flags_);
      DBG_MSG("namemax : [%llu]", stvfs -> namemax_);
      
      break;
    }

    case SSH2_FXP_STATUS:
    {
      DBG_MSG("SftpClient::getFreeSpace : Received SSH2_FXP_STATUS.");

      SftpStatusMsg *statusMsg = (SftpStatusMsg *) ret;

      
      FAIL(statusMsg -> id_ != msgId);

      serverCode = InvertBytes(statusMsg -> status_);

      FAIL(serverCode != SSH2_FX_OK);
      
      break;
    }
    
    default:
    {
      Error("ERROR: Unexpected packet [0x%x] received.\n", ret -> type_);

      FAIL(1);
      
      break;
    }
  }
                             
  //
  // Error handler.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    DBG_MSG("ERROR: Cannot retrieve free space on remote disk.\n"
                "Server code is : %d.", serverCode);
  }

  if (ret)
  {
    free(ret);
  }
  
  //
  // Dropped connection. Try once again if possible.
  //

  if (Not(reconn) && (iorq.state_ == ERROR_STATE || serverCode == 4))
  {
    return statvfs(stvfs, remotePath, 1);
  }

  DBG_LEAVE("SftpClient::getFreeSpace");

  return exitCode;
};

//
// Translate sftp server code to WINAPI error code.
//
// serverStatus - sftp server status (IN).
//
// RETURNS: Related WINAPI error code.
//

Int SftpClient::translateError(Int serverStatus)
{
  struct
  {
    Int serverStatus_;
    Int winapiStatus_;
  }
  tran[] =
  {
/*
    {SSH2_FX_OK,                ERROR_SUCCESS},
*/    
    {SSH2_FX_EOF,               ERROR_HANDLE_EOF},
    {SSH2_FX_NO_SUCH_FILE,      ERROR_FILE_NOT_FOUND},
    {SSH2_FX_PERMISSION_DENIED, ERROR_ACCESS_DENIED},
    {SSH2_FX_FAILURE,           ERROR_GEN_FAILURE},
    {SSH2_FX_BAD_MESSAGE,       ERROR_GEN_FAILURE},
    {SSH2_FX_NO_CONNECTION,     WSAENETDOWN},
    {SSH2_FX_CONNECTION_LOST,   WSAECONNRESET},
    {SSH2_FX_OP_UNSUPPORTED,    ERROR_INVALID_FUNCTION},
    {SSH2_FX_MAX,               ERROR_OUTOFMEMORY},
    {-1,                        -1}
  };
  
  for (Int i = 0; tran[i].serverStatus_ != -1; i++)
  {
    if (serverStatus == tran[i].serverStatus_)
    {
      return tran[i].winapiStatus_;
    }
  }
  
  return ERROR_GEN_FAILURE;
}

//
// Send nxsetcase request to turn on/off case sensitive 
// on server side.
//
// value - 1 to enable, 0 to disable (IN).
//
// RETURNS: 0 if OK.
//

Int SftpClient::setCaseSensitive(Int value)
{
  Int exitCode = -1;
  
  const Char *valueStr = value ? "on" : "off";
  
  Int msgId = IOFifoGenerateUniqueId();
  
  Int serverStatus = 0;
  
  SftpStatusMsg ret = {0};

  SftpNxMsg msg(msgId, "nxsetcase", valueStr);

  IOFifoEntry iorq = {0};
    
  //
  // Prepare SSH2_FXP_RMDIR or SSH2_FXP_REMOVE request.
  //
  
  iorq.id_      = msgId;
  iorq.inpData_ = (SftpMsg *) &msg;
  iorq.outData_ = (SftpMsg *) &ret;
  iorq.outSize_ = sizeof(ret);
  
  DBG_MSG("SftpClient::setCaseSensitive : Sending 'nxsetcase' message...");

  //
  // Send request to server and wait for response.
  //
  
  FAIL(IORequest(&iorq, timeout_));

  //
  // Parse response from server.
  //

  FAIL(ret.type_ != SSH2_FXP_STATUS);
  
  FAIL(ret.id_ != msgId);
  
  serverStatus = InvertBytes(ret.status_);
  
  FAIL(serverStatus != SSH2_FX_OK);
  
  DBG_MSG("Case sensitive is %s.", valueStr);
  
  //
  // Clean up.
  //
  
  fail:

  if (exitCode)
  {
    DBG_MSG("ERROR: Cannot set case sensitive. Error code is: %u."
                " Server status is: %u.", GetLastError(), serverStatus);

    exitCode = -translateError(serverStatus);
  }
  
  return exitCode;
}  
