/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXSSHFS, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef SftpClient_H
#define SftpClient_H

#include <string>
#include <map>

#include "NX/Io.h"
#include "Debug.h"
#include "Utils.h"
#include "String.h"
#include "Sftp.h"
#include "Cache.h"
#include "IOFifo.h"

#define ISDIR(mode) (((mode) & (_S_IFMT)) == (_S_IFDIR))

#define DIR_HANDLE_MASK (1 << 63)

//
// Limit maximum time spent in ReadFile/WriteFile callback
// to avoid reach of system IO timeout, when caller buffer
// is 'very big' or network time 'very slow'. 
// The goal is stop read/write loop before system time out
// and return partially finished read/write operation
// (it. means return success, but with 
// bytesProcessed < bytesToProcess).
//

#define MAX_TIME_PER_CALLBACK 5000.0

typedef void (*AddFindDataCallback) (WIN32_FIND_DATAW *element, void *arg);

struct SftpResource
{
  Int64 remoteHandle_;
  Int64 localHandle_;
  
  string remotePath_;
  
  Int mode_;
  Int dropped_;
  Int isDir_;

  static Int64 localHandleCount_;
};


class SftpClient
{
  private:
  
  Int failedConnections_;
  Int firstConnection_;
  
  const Char *password_;
  const Char *user_;
  const Char *keyfile_;
  const Char *host_;
  const Char *remoteRoot_;

  Char **sshArgs_;
  
  Int sshArgsCount_;

  Int port_;
  Int timeout_;
  Int slaveMode_;

  //
  // File descriptiors to comunicate with child ssh-client process.
  //
  
  Int fdin_;
  Int fdout_;
  Int fderr_;
  
  //
  // Pid of child ssh-client process.
  //
  
  Int pid_;
  
  //
  // Counter of sent messages for creating unique message's id.
  //
  
  Int msgCount_;
  
  //
  // Size of handle expected by server in bytes.
  //

  Int serverHandleSize_;
  
  Int atomicRead(Int fd, Char *buf, Int size);
  
  inline void setTimeout(int timeout)
  {
    timeout_ = timeout;
  }
  
  //
  // Map of current opened resources.
  // Local handle -> {SftpResource}.
  //
  
  map<Int64, SftpResource> openedResources_;
  
  Int getSftpResource(SftpResource *sftpRes, Int64 localHandle);
  
  public:
  
  SftpClient();

  ~SftpClient();
  

  void setupConnection(const Char *user, const Char *host, Int port,
                           const Char *password, const Char *keyfile,
                               Int sshArgsCount, Char **sshArgs = NULL,
                                   Int timeout = 30, Int slaveMode = 0);

  Int connect();
  Int reconnect();

  void disconnect();  

  SftpMsg *allocStringMessage(Int type, Int id, const Char *str,
                                  const void *data = NULL, Int dataSize = 0);
                                  
  Int translateError(Int);
  
  //
  // Wrappers of raw sftp commands.
  //
  
  Int stat(const Char *path, BY_HANDLE_FILE_INFORMATION *extendInfo = NULL, Int reconn = 0);

  Int64 open(const Char *path, Int mode = SSH2_FXP_OPEN,
                 Int64 localHandle = -1, Int *isDir = NULL);
  
  Int close(Int64 handle);
  
  Int readDir(WIN32_FIND_DATAW **buf, Int *count, Int64 handle, Int reconn = 0);

  Int read(Int64 handle, Char *buffer, long long Int offset, Int size, Int reconn = 0);
  
  Int write(Int64 handle, Char *buffer, long long Int offset, Int size, Int reconn = 0);
  
  Int mkdir(const Char *path, Int reconn = 0);

  Int remove(const Char *path, Int isDir = 0, Int reconn = 0);
  
  Int rmdir(const Char *path, Int reconn = 0)
  {
    return remove(path, 1, reconn);
  }
  
  Int rename(const Char *path1, const Char *path2, Int reconn = 0);
  
  Int statvfs(SftpStatvfsReplyMsg *stat, const Char *remotePath, Int reconn = 0);
  
  Int setCaseSensitive(Int value);
};

#endif /* SftpClient_H */
