/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXSSHFS, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Utils.h"

//
// Invert bytes in DWORD. We need this, becouse sftp server needs
// big endian values.
//

Int32 InvertBytes(Int32 x)
{
  DWORD y;

  Char *px = (Char *) &x;
  Char *py = (Char *) &y;
  
  py[0] = px[3];
  py[1] = px[2];
  py[2] = px[1];
  py[3] = px[0];
  
  return y;
}

//
// Invert bytes in QWORD. We need this, becouse sftp server needs
// big endian values.
//

Int64 InvertBytes64(Int64 x)
{
  Int64 y;

  Char *px = (Char *) &x;
  Char *py = (Char *) &y;
  
  py[0] = px[7];
  py[1] = px[6];
  py[2] = px[5];
  py[3] = px[4];
  py[4] = px[3];
  py[5] = px[2];
  py[6] = px[1];
  py[7] = px[0];
  
  return y;
}

DWORD PopInt32(Char *&p)
{
  DWORD *x = (DWORD *) (p);
  
  p += sizeof(DWORD);
  
  return InvertBytes(*x);
}

void PopString(Char *outString, Char *&p, Int *readed, Int capacity)
{
  Int len = PopInt32(p);
  
  if (outString && p)
  {
    Int copyLen = len;

    if (copyLen >= capacity)
    {
      copyLen = capacity - 1;
    }

    memcpy(outString, p, copyLen);
    
    outString[copyLen] = 0;
    
    *readed = copyLen;
  }

  p += len;
}

void SkipString(Char *&p)
{
  Int len = PopInt32(p);
  
  p += len;
}

//
// Transform relative (to where nxsshfs stored) path into full path.
// E.g. it transforms "/somedir" into "<where is nxdevice.exe>/somedir".
// We use it becouse daemon (service) doesn't start in path, where nxdevice.exe
// stored until we don't set current directory manually.
//
// normPath - full, generated path (OUT).
// pathSize - size of path buffer in bytes (IN).
// relative - relative path postfix to add to base path (IN).
//
// RETURNS: 0 if OK.
//

Int GetRelativePath(Char *normPath, Int pathSize, Char *relative)
{
  Int exitCode = -1;

  Int lastSlash = 0;

  Char *path = (Char *) malloc(pathSize);

  FAIL(path == NULL);

  //
  // Get full path to nxdevice.exe binary.
  //

  FAIL(GetModuleFileName(NULL, path, pathSize) == FALSE);

  //
  // Remove filename from path.
  //

  for (Int i = 0; path[i] && i < pathSize; i++)
  {
    if (path[i] == '\\' || path[i] == '/')
    {
      lastSlash = i;
    }
  }

  path[lastSlash] = 0;

  //
  // Add relative postfix if any.
  //

  if (relative)
  {
    StringSizeCat(path, "\\", pathSize);
    StringSizeCat(path, relative, pathSize);
  }

  //
  // Normalize path for output (e.g. remove '..' entries).
  //

  for (int i = 0; path[i]; i++)
  {
    if (path[i] == '/')
    {
      path[i] = '\\';
    }
  }

  PathCanonicalize(normPath, path);

  exitCode = 0;

  //
  // Error handler.
  //

  fail:

  if (exitCode)
  {
    Int err = GetLastError();

    Error("ERROR: Cannot retrieve path of binary.\n"
                 "Error code is : %d.\n", err);

    if (err != 0)
    {
      exitCode = err;
    }
  }

  return exitCode;
}

//
// Wait until given directory become ready to use (i.e. not blocked 
// by any process).
//
// path    - full path to directory to check (IN)
// timeout - maximum time to wait in ms (IN)
//
// RETURNS: 0 if directory became ready until timeout,
//         -1 if errorm,
//          number of opened handles if still busy.
//

Int WaitUntilPathReady(const Char *path, Int timeout)
{
  Int exitCode = -1;
  Int result   = -1;
  
  Char response[1024 * 16];

  FAIL(DeviceDaemonSend(&result, response, sizeof(response), 
                            "\"--ispathbusy\" \"--path\" \"%s\" \"--timeout\" \"%d\"", 
                                path, timeout));

  if (result)
  {
    Error(response);
  }

  exitCode = result;
  
  fail:
  
  return exitCode;
}

//
// Normalize slashes in path to avoid mixing '/' and '\\' style.
//

void NormalizePath(wchar_t *path, Int pathSize, wchar_t slash)
{
  DBG_MSG("Normalizing path [%ls]...", path);
  
  //
  // Remove extra slash at the end.
  //

  Int pathLen = wcslen(path);

  while((pathLen > 0) && ((path[pathLen - 1] == L'\\') || (path[pathLen - 1] == L'/')))
  {
    path[pathLen - 1] = 0;

    pathLen --;
  }

  //
  // Normalize slashes.
  //

  for (Int i = 0; path[i]; i++)
  {
    if (path[i] == L'/' || path[i] == L'\\')
    {
      path[i] = slash;
    }
  }
}

//
// Check does given path is a reparse point.
//

Int PathIsReparsePoint(const wchar_t *path)
{
  DWORD pathAttributes = GetFileAttributesW(path);
  
  DBG_MSG("PathIsReparsePoint : attributes for [%ls] is [%x].", path, pathAttributes);
  
  if ((pathAttributes & FILE_ATTRIBUTE_REPARSE_POINT) ||
          (pathAttributes & IO_REPARSE_TAG_SYMLINK))
  {
    return 1;
  }
  
  return 0;
}

//
// Check does fiven path is empty.
//

Int PathIsEmpty(const wchar_t *path)
{
  wchar_t mask[MAX_PATH];
  
  int count = 0;
  
  HANDLE hFind = INVALID_HANDLE_VALUE;
  
  WIN32_FIND_DATAW ffd = {0};
  
  _snwprintf(mask, MAX_PATH, L"%ls/*", path);
    
  hFind = FindFirstFileW(mask, &ffd);

  while (FindNextFileW(hFind, &ffd)) 
  {
    count++;
  }  
  
  FindClose(hFind);
  
  return (count < 2);
}

//
// Retrieve parent directory from full path.
//
// parent     - retrieved parent directory (OUT).
// parentSize - size of paernt buffer in bytes (IN).
// path       - full path to split (IN).
//
// RETURNS: 0 if OK.
//

Int GetParentPath(wchar_t *parent, Int parentSize, const wchar_t *path)
{
  Int lastSlash = 0;
  
  for (Int i = 0; path[i]; i++)
  {
    if (path[i] == L'/' || path[i] == L'\\')
    {
      lastSlash = i;
    }
  }
  
  if (lastSlash >= parentSize)
  {
    lastSlash = parentSize - 2;
  }
  
  //
  // Root path : '\\abc' |-> '\\'.
  // Save last slash.
  //
  
  if (lastSlash == 0)
  {
    parent[0] = L'\\';
    parent[1] = 0;
  }
  
  //
  // Non-root path : '\\xyz\\abc' |-> '\\xyz'
  // Remove last slash.
  //
  
  else
  {
    memcpy(parent, path, lastSlash * sizeof(wchar_t));

    parent[lastSlash] = 0;
  }

  DBG_MSG("Retrieved parent [%ls] for [%ls].", parent, path);
  
  return 0;
}

//
// Send formatted messaege to daemon service.
//
// daemonResult - exit code returned from daemon (OUT)
// response     - buffer, where to store daemon's response (OUT).
//                Can be NULL if response not needed.
//
// responseSize - size of response buffer in bytes (IN).
// fmt...       - printf like arguments to generate formatted string (IN).
//
// RETURNS: 0 if command send to daemon and answer from daemon received.
//          Note, that command may still fail on daemon side.
//          Check daemonResult arg for get daemon result.
//

Int DeviceDaemonSend(Int *daemonResult, Char *response,
                         Int responseSize, const char *fmt, ...)
{
  DBG_ENTER("DaemonSend");

  Int exitCode = -1;

  unsigned Int msgSize = 1024 * 16;

  Char msg[msgSize];

  DWORD written = 0;
  DWORD readed  = 0;
  DWORD totalReaded = 0;

  HANDLE pipe = NULL;

  Int endReached = 0;

  OVERLAPPED ov = {0};

  //
  // Wait until communication pipe availible.
  //

  DBG_MSG("Waiting for communication pipe availible...");

  FAIL(WaitNamedPipe(NXDEVICED_PIPE, 20000) == FALSE);

  //
  // Open communication pipe.
  //

  DBG_MSG("Opening communication pipe...");

  pipe = CreateFile(NXDEVICED_PIPE, GENERIC_READ | GENERIC_WRITE,
                        0, NULL, OPEN_EXISTING, 0, NULL);

  FAIL(pipe == INVALID_HANDLE_VALUE);

  //
  // Create formatted string with message to send.
  //

  va_list ap;

  va_start(ap, fmt);

  vsnprintf(msg, msgSize, fmt, ap);

  va_end(ap);

  //
  // Add "__END__" tag at the end of message.
  //

  StringSizeCat(msg, "__END__", msgSize);

  //
  // Write formatted message to communication pipe
  // (i.e. to service daemon).
  //

  DBG_MSG("Sending message to daemon...");

  ResetEvent(ov.hEvent);

  WriteFile(pipe, msg, strlen(msg), NULL, &ov);

  FAIL(GetOverlappedResult(pipe, &ov, &written, TRUE) == FALSE);

  //
  // Read response from daemon if needed.
  //

  if (response && responseSize > 0)
  {
    //
    // Create event for overlapped I/O.
    //

    DBG_MSG("Setting up overlapped mode...");

    ov.hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);

    FAIL(ov.hEvent == INVALID_HANDLE_VALUE);

    //
    // Read response until __END__ tag readed or error.
    //

    DBG_MSG("Reading response from daemon...");

    response[0] = 0;

    endReached = 0;

    while(Not(endReached))
    {
      ResetEvent(ov.hEvent);

      FAIL(totalReaded >= (DWORD) responseSize);

      if (ReadFile(pipe, response + totalReaded,
                       responseSize - totalReaded - 1,
                           NULL, &ov) == FALSE)
      {
        FAIL(GetLastError() != ERROR_IO_PENDING);
      }

      FAIL(GetOverlappedResult(pipe, &ov, &readed, TRUE) == FALSE);

      FAIL(readed <= 0);

      //
      // Daemon successfull executed.
      //

      if (StringFind(response + totalReaded, "__OK__"))
      {
        totalReaded += readed - StringLength("__OK__");

        endReached = 1;

        if (daemonResult)
        {
          *daemonResult = 0;
        }
      }

      //
      // Daemon executed command but result is FAIL.
      //

      else if (StringFind(response + totalReaded, "__FAIL__"))
      {
        totalReaded += readed - StringLength("__FAIL__");

        endReached = 1;

        if (daemonResult)
        {
          *daemonResult = -1;
        }
      }

      //
      // Still more data to send. Wait for finish.
      //

      else
      {
        totalReaded += readed;
      }

      response[totalReaded] = 0;
    }
  }

  DBG_MSG("Daemon response [%s]", response);

  exitCode = 0;

  //
  // Clean up.
  //

  fail:

  if (exitCode)
  {
    Error("ERROR: Cannot send message to daemon.\n"
              "Error code is : %d.\n", (Int) GetLastError());
  }

  DisconnectNamedPipe(pipe);

  CloseHandle(pipe);

  DBG_LEAVE("DaemonSend");

  return exitCode;
}

//
// Workaround to implement atomic	 read with timeout without
// overlapped mode. This function is synchronous read worker,
// which is executed in background thread.
//
// RETURNS: 0 if OK.
//

Int AtomicReadWorker(AtomicReadInfo *ari)
{
  DWORD readed      = 0;
  DWORD bytesToRead = ari -> bufSize;
  
  Char *buf = ari -> buf;
  
  while(bytesToRead > 0)
  {
    FAIL(ReadFile(ari -> handle, buf, bytesToRead, &readed, NULL) == FALSE);
  
    DBG_MSG("Readed [%d] bytes from [%p].", readed, ari -> handle);
  
    bytesToRead -= readed;
    buf         += readed;
  }
  
  fail:
  
  if (bytesToRead == 0)
  {
    return 0;
  }
  else
  {
    DWORD err = GetLastError();
    
    if (err)
    {
      return err;
    }
    else
    {
      return -1;
    }  
  }
}

//
// Atomic read for non-overlapped mode fd.
//
// buf     - buffer where to store readed data (OUT).
// bufSize - bytes to read (IN).
// handle  - input handle (IN).
// timeout - timeout in ms (IN).
//
// RETURNS: 0 if All data readed.
//

Int AtomicRead(Char *buf, Int bufSize, HANDLE handle, DWORD timeout)
{
  Int exitCode = -1;
  
  AtomicReadInfo ari;

  HANDLE worker = NULL;

  DWORD workerResult = -1;

  ari.handle  = handle;
  ari.buf     = buf;
  ari.bufSize = bufSize;

  worker = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) AtomicReadWorker,
                            &ari, 0, NULL);
  
  FAILEX(worker == NULL, "ERROR: Cannot create read worker.\n");
  
  FAILEX(WaitForSingleObject(worker, timeout) != WAIT_OBJECT_0,
             "ERROR: Cannot wait for read worker.\n");
  
  FAIL(GetExitCodeThread(worker, &workerResult) == FALSE);

  FAILEX(workerResult != 0, 
             "ERROR: Read thread failed with code [%d].\n", workerResult);
  
  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Error("ERROR: Cannot read [%d] bytes from [%p].\n"
              "Error code is : [%d].\nWorker result is : [%d].\n", 
                  bufSize, handle, GetLastError());
  }
  
  if (worker != NULL)
  {
    TerminateThread(worker, -1);
  }  
  
  return exitCode;
}     

//
// Returns current time in ms.
//

double GetTimeMs()
{
  struct timeval tv;
  
  gettimeofday(&tv, NULL);

  return tv.tv_sec * 1000.0 + tv.tv_usec / 1000.0;
}

//
// Set security string for given named object (e.g. file).
//
// objName - object name (e.g. path to file). (IN)
// rights  - security string with access list to set (IN).
// inherit - New ACL inherits rights from parent if 1 (IN).
//
// RETURNS: 0 if OK.
//

Int SetObjectRights(const wchar_t *objName, const Char *rights, Int inherit)
{
  Int exitCode = -1;
  
  ACL *acl = NULL;

  SECURITY_DESCRIPTOR sd = {0};

  PSECURITY_DESCRIPTOR psd = &sd;
  
  DWORD flags;
  
  BOOL aclExists  = FALSE;
  BOOL aclDefault = FALSE;
 
  //
  // Initialize empty security descriptor.
  //
  
  FAILEX(InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION) == FALSE,
             "ERROR: Cannot initialize security descriptor.\n");
  
  //
  // Set NULL access list for security descriptor.
  //

  FAILEX(SetSecurityDescriptorDacl(&sd, TRUE, NULL, FALSE) == FALSE,
             "ERROR: Cannot set DACL for security descriptor.\n");

  //
  // Init security descriptor from security string.
  //
  
  FAILEX(ConvertStringSecurityDescriptorToSecurityDescriptor(rights, SDDL_REVISION_1,
                                                               &psd, NULL) == FALSE,
             "ERROR: Cannot process security string.\n");

  //
  // Retrieve access list from craeted security descriptor.
  //
  
  FAILEX(GetSecurityDescriptorDacl(psd, &aclExists, &acl, &aclDefault) == FALSE,
             "ERROR: Cannot retrive acl from security descriptor.\n");

  //
  // Set new access list to named object.
  //

  flags = DACL_SECURITY_INFORMATION;
  
  if (Not(inherit))
  {
    flags |= PROTECTED_DACL_SECURITY_INFORMATION;
  }

  FAILEX(SetNamedSecurityInfoW((wchar_t *) objName, SE_FILE_OBJECT, flags,
                                   NULL, NULL, acl, NULL),
                                       "ERROR: Cannot set access list.\n")

  //
  // Clean up.
  //
  
  exitCode = 0;
  
  fail:
  
  if (exitCode)
  {
    Error("ERROR: Cannot set security string [%s] for [%ls].\n"
              "Error code is : %d.\n", rights, objName, GetLastError());
  }
  
  if (psd)
  {
    LocalFree(psd);
  }
  
  return exitCode;
}

//
// Transform relative (to where nxdevice stored) path into full path.
// E.g. it transforms "/somedir" into "<where is nxdevice.exe>/somedir".
// We use it becouse daemon (service) doesn't start in path, where nxdevice.exe
// stored until we don't set current directory manually.
//
// WARNING: If passed path is already absolute it returns the same string.
//
// normPath - full, generated path (OUT).
// pathSize - size of path buffer in bytes (IN).
// relative - relative path postfix to add to base path (IN).
// quiet    - do not write logs if set to 1 (IN).
//
// RETURNS: 0 if OK.
//

Int ExpandRelativePath(Char *normPath, Int pathSize, const Char *relative, Int quiet)
{
  if (Not(quiet))
  {
    DBG_ENTER("ExpandRelativePath");
  }  
  
  Int exitCode = -1;

  Int lastSlash = 0;

  Char *path = (Char *) malloc(pathSize);

  FAILEX(path == NULL, "ERROR: Out of memory.\n");
  
  //
  // Check is it already absolute?
  //
  
  if (relative && relative[0] && relative[1] == ':')
  {
    if (Not(quiet))
    {
      DBG_MSG("Path [%s] is already absolute.\n", relative);
    }  
    
    StringSizeCopy(path, relative, pathSize);
  }
  else
  {
    if (Not(quiet))
    {
      DBG_MSG("Expanding path [%s]...\n", relative);
    }  
    
    //
    // Get full path to nxdevice.exe binary.
    //

    FAIL(GetModuleFileName(NULL, path, pathSize) == FALSE);
  
    //
    // Remove filename from path.
    //

    for (Int i = 0; path[i] && i < pathSize; i++)
    {
      if (path[i] == '\\' || path[i] == '/')
      {
        lastSlash = i;
      }
    }

    path[lastSlash] = 0;

    //
    // Add relative postfix if any.
    //

    if (relative)
    {
      StringSizeCat(path, "\\", pathSize);
      StringSizeCat(path, relative, pathSize);
    }
  }
    
  //
  // Normalize path for output (e.g. remove '..' entries).
  //

  for (int i = 0; path[i]; i++)
  {
    if (path[i] == '/')
    {
      path[i] = '\\';
    }
  }

  PathCanonicalize(normPath, path);
    
  exitCode = 0;
  
  if (Not(quiet))
  {
    DBG_MSG("Path [%s] expanded to [%s].\n", relative, normPath);
  }  

  //
  // Error handler.
  //

  fail:

  if (exitCode)
  {
    Int err = GetLastError();

    Error("ERROR: Cannot retrieve path of binary.\n"
              "Error code is : %d.\n", err);

    if (err != 0)
    {
      exitCode = err;
    }
  }
  
  if (path != NULL)
  {
    free(path);
  }

  if (Not(quiet))
  {
    DBG_LEAVE("ExpandRelativePath");
  }  

  return exitCode;
}

//
// Get current running windows version.
//

WindowsVersion GetWindowsVersion()
{
  OSVERSIONINFOEX ver;

  ver.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);

  GetVersionEx((OSVERSIONINFO *) &ver);

  switch (ver.dwMajorVersion)
  {
    case 5:
    {
      return windows_xp;
    }
    case 6:
    {
      return (ver.dwMinorVersion >= 1) ? windows_7 : windows_vista;
    }
  }

  return windows_unknown;
}

//
// Retrieve directory where log file is stored.
//

const Char *GetLogPath(const Char *postfix)
{
  static Char root[MAX_PATH] = {0};
  static Char path[MAX_PATH] = {0};
  
  if (path[0] == 0)
  {
    if (getenv("ALLUSERSPROFILE"))
    {
      snprintf(root, MAX_PATH, "%s\\NoMachine\\var\\log", getenv("ALLUSERSPROFILE"));
    }
    else
    {
      if (GetWindowsVersion() == windows_xp)
      {
        snprintf(root, MAX_PATH, "C:\\Documents And Settings\\All Users\\NoMachine\\var\\log");
      }
      else
      {  
        snprintf(root, MAX_PATH, "C:\\ProgramData\\NoMachine\\var\\log");
      }  
    }
    
    if (PathFileExistsA(root))
    {
      strncpy(path, root, MAX_PATH);
      strncat(path, "\\", MAX_PATH);
      strncat(path, postfix, MAX_PATH);
    }
    else
    {
      Char relative[MAX_PATH];
      
      strncpy(relative, "..\\var\\log\\", MAX_PATH);
      strncat(relative, postfix, MAX_PATH);
      
      ExpandRelativePath(path, MAX_PATH, relative, 1);
    }
  }  
  
  return path;
}

HANDLE ShutdownMutex = NULL;

Int UtilsInit()
{
  Int exitCode = 0;

  ShutdownMutex = CreateMutex(NULL, FALSE, NULL);

  if (ShutdownMutex == NULL)
  {
    exitCode = -1;
  }

  return exitCode;
}

void UtilsCleanup()
{
  DBG_ENTER("UtilsCleanup");

  CloseHandle(ShutdownMutex);

  DBG_LEAVE("UtilsCleanup");
}

void StopProcessSafely()
{
  // aquire and never release ownership of mutex, since console event
  // should be generated only once
  if (WaitForSingleObject(ShutdownMutex, 0) == WAIT_OBJECT_0)
  {
    GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT, 0);
  }
}
